unit ut_packages;
{
   UT Packages Delphi unit v1.1
   (C) 2001 Antonio Cordero Balcazar

   This unit allows you to read Unreal engine package files and their
   internal objects and extract information from them.

   Thanks to Epic Games and to everyone who helped.

   Legal: this source code is released as freeware. You can freely use and
          modify this source code, but you cannot distribute modified copies
          of it. Any changes and ideas are welcome.

   History:

   1.1 - Fixed UV mapping in 3DStudio exporter.
         Added MirrorX effect in 3DStudio exporter.
         Changed PrepareExporter function to support multiple frames.
         Added Unreal3D exporter class.
         Implemented Unreal3D exporter for SkeletalMeshes.
         Renames screwed object names.
         Renamed some properties of TUTExportTableObjectData and
           TUTImportTableObjectData to avoid name clashes.
         Added support for user-definable Native Functions array.
   1.0 - First public version.
         Support for following classes:
         Class (null class name),Field,Const,Enum,Struct,Function,State,
         Property,ByteProperty,IntProperty,BoolProperty,FloatProperty,
         ObjectProperty,ClassProperty,NameProperty,StructProperty,StrProperty,
         StringProperty,ArrayProperty,FixedArrayProperty,MapProperty,
         Palette,Sound,Music,TextBuffer,Font,Polys,
         Texture,WaterTexture,WaveTexture,WetTexture,IceTexture,FireTexture,
         ScriptedTexture,
         Primitive,Mesh,LodMesh,SkeletalMesh
}

// TODO : Add Property/Object/Package creation (another class)
// TODO : Complete support for other serialized property types.
// TODO : Complete support for compressed mipmaps in TUTObjectClassTexture
// TODO : Add class "Animation"

interface

uses windows, sysutils, classes, graphics;

const
  // Package flags
  PKG_AllowDownload = $0001;            // Allow downloading package.
  PKG_ClientOptional = $0002;           // Purely optional for clients.
  PKG_ServerSideOnly = $0004;           // Only needed on the server side.
  PKG_BrokenLinks = $0008;              // Loaded from linker with broken import links.
  PKG_Unsecure = $0010;                 // Not trusted.
  PKG_Need = $8000;                     // Client needs to download this package.

  // Object flags
  RF_Transactional = $00000001;         // Object is transactional.
  RF_Unreachable = $00000002;           // Object is not reachable on the object graph.
  RF_Public = $00000004;                // Object is visible outside its package.
  RF_TagImp = $00000008;                // Temporary import tag in load/save.
  RF_TagExp = $00000010;                // Temporary export tag in load/save.
  RF_SourceModified = $00000020;        // Modified relative to source files.
  RF_TagGarbage = $00000040;            // Check during garbage collection.
  RF_Unk_00000080 = $00000080;
  RF_Unk_00000100 = $00000100;
  RF_NeedLoad = $00000200;              // During load, indicates object needs loading.
  RF_HighlightedName = $00000400;       // A hardcoded name which should be syntax-highlighted.
  RF_InSingularFunc = $00000800;        // In a singular function.
  RF_Suppress = $00001000;              // Suppressed log name.
  RF_InEndState = $00002000;            // Within an EndState call.
  RF_Transient = $00004000;             // Don't save object.
  RF_PreLoading = $00008000;            // Data is being preloaded from file.
  RF_LoadForClient = $00010000;         // In-file load for client.
  RF_LoadForServer = $00020000;         // In-file load for client.
  RF_LoadForEdit = $00040000;           // In-file load for client.
  RF_Standalone = $00080000;            // Keep object around for editing even if unreferenced.
  RF_NotForClient = $00100000;          // Don't load this object for the game client.
  RF_NotForServer = $00200000;          // Don't load this object for the game server.
  RF_NotForEdit = $00400000;            // Don't load this object for the editor.
  RF_Destroyed = $00800000;             // Object Destroy has already been called.
  RF_NeedPostLoad = $01000000;          // Object needs to be postloaded.
  RF_HasStack = $02000000;              // Has execution stack.
  RF_Native = $04000000;                // Native (UClass only).
  RF_Marked = $08000000;                // Marked (for debugging).
  RF_ErrorShutdown = $10000000;         // ShutdownAfterError called.
  RF_DebugPostLoad = $20000000;         // For debugging Serialize calls.
  RF_DebugSerialize = $40000000;        // For debugging Serialize calls.
  RF_DebugDestroy = $80000000;          // For debugging Destroy calls.
  // The following flags have duplicated values but they seem to be used only
  // internally and so they will not appear in packages.
  //RF_EliminateObject  = 0x00000400,   // NULL out references to this during garbage collecion.
  //RF_RemappedName     = 0x00000800,   // Name is remapped.
  //RF_StateChanged     = 0x00001000,   // Object did a state change.

  // Property types
  otNone = 0;
  otByte = 1;
  otInt = 2;
  otBool = 3;
  otFloat = 4;
  otObject = 5;
  otName = 6;
  otString = 7;                         // old type
  otClass = 8;                          // not implemented
  otArray = 9;                          // not implemented
  otStruct = 10;
  otVector = 11;                        // not implemented => seen as struct...
  otRotator = 12;                       // not implemented => seen as struct...
  otStr = 13;
  otMap = 14;                           // not implemented
  otFixedArray = 15;                    // not implemented
  // Extended value types
  otExtendedValue = $00000100;
  otBuffer = otExtendedValue or 0;
  otWord = otExtendedValue or 1;

type
  TUTPackage = class;
  TUTPropertyType = cardinal;

  // TUTProperty
  TUTProperty = class
  private
    FIsInitialized: boolean;
    FOwner: TUTPackage;
    FName: string;
    FArrayIndex: integer;
    FPropertyType: TUTPropertyType;
    FValue: array of byte;
    FTypeName: string;
    function GetValueCount: integer;
    function GetFirstValue: variant;
    function GetTypeName: string;
    function GetDescription: string;
  public
    procedure SetProperty(Owner: TUTPackage; n: string; i: integer; t: TUTPropertyType; var value; valuesize: integer; typename: string = '');
    property Name: string read FName;
    property ArrayIndex: integer read FArrayIndex;
    property PropertyType: TUTPropertyType read FPropertyType;
    property Description: string read GetDescription;
    property SpecificTypeName: string read GetTypeName;
    property ValueCount: integer read GetValueCount;
    procedure GetValue(i: integer; var valuename: string; var value: variant; var valuedescription: string; var valuetype: TUTPropertyType);
    function GetValueTypeName(t: TUTPropertyType): string;
    property Value: variant read GetFirstValue;
  end;

  TUTPropertyClass = class of TUTProperty;

  // TUTPropertyList
  TUTPropertyList = class
  private
    FProperties: tlist;
    function GetProperty(i: integer): TUTProperty;
    function GetPropertyCount: integer;
    function NewProperty: TUTProperty;
    function GetPropertyByName(name: string): TUTProperty;
    function GetPropertyListDescriptions: string;
    function GetPropertyByNameValue(name: string): variant;
    function GetPropertyValue(i: integer): variant;
  public
    constructor Create;
    destructor Destroy; override;
    property New: TUTProperty read NewProperty;
    property Count: integer read GetPropertyCount;
    property PropertyByPosition[i: integer]: TUTProperty read GetProperty;
    property PropertyByName[name: string]: TUTProperty read GetPropertyByName;
    property PropertyByNameValue[name: string]: variant read GetPropertyByNameValue; default;
    property PropertyByPositionValue[i: integer]: variant read GetPropertyValue;
    procedure Clear;
    property Descriptions: string read GetPropertyListDescriptions;
    procedure FixArrayIndices;
  end;

  // TUTObject : generic class for UT exported objects
  // Status: completed
  TUTObject = class
  private
    FStartInPackage: integer;
    Fowner: TUTPackage;
    Fexportedindex: integer;
    function GetClassName: string;
    function GetObjectname: string;
    function GetPackageName: string;
    function GetSuperName: string;
    function GetClassIndex: integer;
    function GetObjectIndex: integer;
    function GetPackageIndex: integer;
    function GetSuperIndex: integer;
    function GetSerialOffset: integer;
    function GetSerialSize: integer;
    function GetFlags: longword;
    function GetProperties: TUTPropertyList;
    procedure check_initialized;
    function GetFullName: string;
  protected
    FProperties: TUTPropertyList;
    procedure InitializeObject; virtual;
    procedure InterpretObject; virtual;
    procedure ReadProperties; virtual;
  public
    Buffer: TMemoryStream;
    constructor create(owner: TUTPackage; exportedindex: integer); virtual;
    destructor destroy; override;
    procedure ReadObject(interpret: boolean = true);
    procedure ReleaseObject; virtual;
    property Properties: TUTPropertyList read GetProperties;
    property UTObjectIndex: integer read GetObjectIndex;
    property UTClassIndex: integer read GetClassIndex;
    property UTPackageIndex: integer read GetPackageIndex;
    property UTSuperIndex: integer read GetSuperIndex;
    property UTSerialOffset: integer read GetSerialOffset;
    property UTSerialSize: integer read GetSerialSize;
    property UTFlags: longword read GetFlags;
    property UTObjectName: string read GetObjectname;
    property UTClassName: string read GetClassName;
    property UTPackageName: string read GetPackageName;
    property UTSuperName: string read GetSuperName;
    property UTFullName: string read GetFullName;
  end;

  TUTObjectClass = class of TUTObject;

  // TUTObject_NoHeader (generic class for objects that do not have a header)
  // Status: completed
  TUTObject_NoHeader = class(TUTObject)
  protected
    procedure InterpretObject; override;
  end;

  // TUTObjectClassField
  // Status: completed
  TUTObjectClassField = class(TUTObject)
  private
    FSuperField, FNext: integer;
    function GetNext: integer;
    function GetSuperField: integer;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    property SuperField: integer read GetSuperField;
    property Next: integer read GetNext;
    procedure ReleaseObject; override;
  end;

  // TUTObjectClassEnum
  // Status: completed
  TUTObjectClassEnum = class(TUTObjectClassField)
  private
    FValues: array of integer;
    function GetCount: integer;
    function GetValue(i: integer): integer;
    function GetValueName(i: integer): string;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    property Count: integer read GetCount;
    property EnumValue[i: integer]: integer read GetValue;
    property EnumName[i: integer]: string read GetValueName;
    procedure ReleaseObject; override;
    function GetDeclaration: string;
  end;

  // TUTObjectClassConst
  // Status: completed
  TUTObjectClassConst = class(TUTObjectClassField)
  private
    FValue: string;
    function GetValue: string;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    property Value: string read GetValue;
    procedure ReleaseObject; override;
    function GetDeclaration: string;
  end;

  // TUTObjectClassProperty
  // Status: completed
  TUTObjectClassProperty = class(TUTObjectClassField)
  private
    FArrayDimension: integer;
    FElementSize: integer;
    FPropertyFlags: longword;
    FCategory: string;
    FRepOffset: word;
    function GetArrayDimension: integer;
    function GetCategory: string;
    function GetElementSize: integer;
    function GetPropertyFlags: longword;
    function GetRepOffset: word;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    property ArrayDimension: integer read GetArrayDimension;
    property ElementSize: integer read GetElementSize;
    property PropertyFlags: longword read GetPropertyFlags;
    property Category: string read GetCategory;
    property ReplicationOffset: word read GetRepOffset;
    function TypeName: string; virtual;
    function GetFlags(cn: string): string;
    function GetDeclaration(context, cn: string): string;
    procedure ReleaseObject; override;
  end;

  // TUTObjectClassByteProperty
  // Status: completed
  TUTObjectClassByteProperty = class(TUTObjectClassProperty)
  private
    FEnum: byte;
    function GetEnum: byte;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    property Enum: byte read GetEnum;
    function TypeName: string; override;
    procedure ReleaseObject; override;
  end;

  // TUTObjectClassIntProperty
  // Status: completed
  TUTObjectClassIntProperty = class(TUTObjectClassProperty)
  private
  protected
  public
    function TypeName: string; override;
  end;

  // TUTObjectClassBoolProperty
  // Status: completed
  TUTObjectClassBoolProperty = class(TUTObjectClassProperty)
  private
  protected
  public
    function TypeName: string; override;
  end;

  // TUTObjectClassFloatProperty
  // Status: completed
  TUTObjectClassFloatProperty = class(TUTObjectClassProperty)
  private
  protected
  public
    function TypeName: string; override;
  end;

  // TUTObjectClassNameProperty
  // Status: completed
  TUTObjectClassNameProperty = class(TUTObjectClassProperty)
  private
  protected
  public
    function TypeName: string; override;
  end;

  // TUTObjectClassStrProperty
  // Status: completed
  TUTObjectClassStrProperty = class(TUTObjectClassProperty)
  private
  protected
  public
    function TypeName: string; override;
  end;

  // TUTObjectClassStringProperty
  // Status: completed
  TUTObjectClassStringProperty = class(TUTObjectClassProperty)
  private
  protected
  public
    function TypeName: string; override;
  end;

  // TUTObjectClassObjectProperty
  // Status: completed
  TUTObjectClassObjectProperty = class(TUTObjectClassProperty)
  private
    FObject: integer;
    function GetObject: integer;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    property UTObjectType: integer read GetObject;
    function TypeName: string; override;
    procedure ReleaseObject; override;
  end;

  // TUTObjectClassClassProperty
  // Status: completed
  TUTObjectClassClassProperty = class(TUTObjectClassObjectProperty)
  private
    FClass: integer;
    function GetClass: integer;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    property UTClassType: integer read GetClass;
    function TypeName: string; override;
    procedure ReleaseObject; override;
  end;

  // TUTObjectClassFixedArrayProperty
  // Status: complete
  TUTObjectClassFixedArrayProperty = class(TUTObjectClassProperty)
  private
    FInner: integer;
    FCount: integer;
    function GetCount: integer;
    function GetInner: integer;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    property ElementCount: integer read GetCount;
    property InnerProperty: integer read GetInner;
    function TypeName: string; override;
    procedure ReleaseObject; override;
  end;

  // TUTObjectClassArrayProperty
  // Status: completed
  TUTObjectClassArrayProperty = class(TUTObjectClassProperty)
  private
    FInner: integer;
    function GetInner: integer;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    property InnerProperty: integer read GetInner;
    function TypeName: string; override;
    procedure ReleaseObject; override;
  end;

  // TUTObjectClassMapProperty
  // Status: completed
  TUTObjectClassMapProperty = class(TUTObjectClassProperty)
  private
    FKey: integer;
    FValue: integer;
    function GetKey: integer;
    function GetValue: integer;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    property KeyProperty: integer read GetKey;
    property ValueProperty: integer read GetValue;
    function TypeName: string; override;
    procedure ReleaseObject; override;
  end;

  // TUTObjectClassStructProperty
  // Status: completed
  TUTObjectClassStructProperty = class(TUTObjectClassProperty)
  private
    FStruct: integer;
    function GetStruct: integer;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    property Struct: integer read GetStruct;
    function TypeName: string; override;
    procedure ReleaseObject; override;
  end;

  TUTLabelEntry = record
    Name: integer;
    iCode: integer;
  end;

  // TUTObjectClassStruct
  // Status: completed
  TUTObjectClassStruct = class(TUTObjectClassField)
  private
    FScriptText: integer;
    FChildren: integer;
    FFriendlyName: string;
    FLine: integer;
    FTextPos: integer;
    FScriptSize: integer;
    function GetChildren: integer;
    function GetFriendlyName: string;
    function GetLine: integer;
    function GetScriptSize: integer;
    function GetScriptText: integer;
    function GetTextPos: integer;
    function ReadStatement: string;
  protected
    FScriptStart: integer;
    jumplist, nest: tlist;
    endnestlist: tstringlist;
    labellist, indent_chars: string;
    need_semicolon, context_change: boolean;
    position_icode: integer;
    indent_level: integer;
    FLabelTable: array of TUTLabelEntry;
    procedure InitializeObject; override;
    procedure InterpretObject; override;
    function ReadStatements(beautify: boolean = true): string;
    procedure SkipStatements;
  public
    property ScriptText: integer read GetScriptText;
    property FirstChild: integer read GetChildren;
    property FriendlyName: string read GetFriendlyName;
    property Line: integer read GetLine;
    property TextPos: integer read GetTextPos;
    property ScriptSize: integer read GetScriptSize;
    procedure ReleaseObject; override;
    function GetDeclaration: string;
    function ReadToken(OuterOperatorPrecedence: byte = 255): string;
  end;

  // TUTObjectClassState
  // Status: completed
  TUTObjectClassState = class(TUTObjectClassStruct)
  private
    FProbeMask: int64;
    FIgnoreMask: int64;
    FLabelTableOffset: word;
    FStateFlags: longword;
    function GetIgnoreMask: int64;
    function GetLabelTableOffset: word;
    function GetProbeMask: int64;
    function GetStateFlags: longword;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    procedure ReleaseObject; override;
    property ProbeMask: int64 read GetProbeMask;
    property IgnoreMask: int64 read GetIgnoreMask;
    property LabelTableOffset: word read GetLabelTableOffset;
    property StateFlags: longword read GetStateFlags;
    function Decompile(beautify: boolean = true): string;
  end;

  TUTDependency = record
    _Class: integer;
    Deep: integer;
    ScriptTextCRC: integer;
  end;

  // TUTObjectClassClass
  // Status: completed
  TUTObjectClassClass = class(TUTObjectClassState)
  private
    FClassFlags: longword;
    FClassGuid: TGuid;
    FDependencies: array of TUTDependency;
    FPackageImports: array of integer;
    FClassWithin: integer;
    FClassConfigName: integer;
    function GetDependencies(i: integer): TUTDependency;
    function GetPackageImports(i: integer): integer;
    function GetClassConfigName: integer;
    function GetClassFlags: longword;
    function GetClassGuid: TGuid;
    function GetClassWithin: integer;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    procedure ReleaseObject; override;
    function Decompile(beautify: boolean = true): string;
    property ClassFlags: longword read GetClassFlags;
    property ClassGuid: TGuid read GetClassGuid;
    function DependencyCount: integer;
    property Dependencies[i: integer]: TUTDependency read GetDependencies;
    function PackageImportsCount: integer;
    property PackageImports[i: integer]: integer read GetPackageImports;
    property ClassWithin: integer read GetClassWithin;
    property ClassConfigName: integer read GetClassConfigName;
  end;

  // TUTObjectClassPalette
  // Status: completed
  TUTObjectClassPalette = class(TUTObject_NoHeader)
  private
    FColorCount: integer;
    FColors: array of TRGBQuad;
    function GetColor(n: integer): TColor;
    function GetNewPalette: HPalette;
    function GetColorCount: integer;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    property ColorCount: integer read GetColorCount;
    property Color[n: integer]: TColor read GetColor;
    property GetPalette: HPalette read GetNewPalette;
    procedure ReleaseObject; override;
  end;

  // TUTObjectClassSound
  // Status: completed
  TUTObjectClassSound = class(TUTObject_NoHeader)
  private
    FFormat: string;
    FData: string;
    function GetData: string;
    function GetFormat: string;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    property Format: string read GetFormat;
    property Data: string read GetData;
    procedure SaveToFile(filename: string);
    procedure ReleaseObject; override;
  end;

  // TUTObjectClassMusic
  // Status: completed
  TUTObjectClassMusic = class(TUTObject_NoHeader)
  private
    FFormat: string;
    FData: string;
    function GetData: string;
    function GetFormat: string;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    property Format: string read GetFormat;
    property Data: string read GetData;
    procedure SaveToFile(filename: string);
    procedure ReleaseObject; override;
  end;

  // TUTObjectClassTextBuffer
  // Status: completed
  TUTObjectClassTextBuffer = class(TUTObject_NoHeader)
  private
    FData: string;
    function GetData: string;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    property Data: string read GetData;
    procedure SaveToFile(filename: string);
    procedure ReleaseObject; override;
  end;

  TUT_FontCharacter = record
    Texture: integer;
    X, Y, W, H: integer;
  end;

  // TUTObjectClassFont
  // Status: completed
  TUTObjectClassFont = class(TUTObject_NoHeader)
  private
    FCharacters: array of TUT_FontCharacter;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    procedure GetCharacterInfo(i: integer; var texture, x, y, w, h: integer);
    procedure ReleaseObject; override;
  end;

  TUT_Struct_Vector = packed record
    X, Y, Z: single;
  end;
  TUT_Struct_Rotator = packed record
    Yaw, Roll, Pitch: integer;
  end;

  TUT_Struct_Polygon = packed record
    Base, Normal, TextureU, TextureV: TUT_Struct_Vector;
    Vertex: array of TUT_Struct_Vector;
    PolyFlags, Actor, Texture, ItemName, iLink, iBrushPoly, pan_u, pan_v: integer;
  end;

  // TUTObjectClassPolys
  // Status: completed
  TUTObjectClassPolys = class(TUTObject_NoHeader)
  private
    FPolygons: array of TUT_Struct_Polygon;
    function GetPolygonCount: integer;
    function GetPolygon(n: integer): TUT_Struct_Polygon;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    property PolygonCount: integer read GetPolygonCount;
    property Polygon[n: integer]: TUT_Struct_Polygon read GetPolygon;
    procedure ReleaseObject; override;
  end;

  // TUTObjectClassTexture
  // Status: near completed
  //   Compressed MipMaps with formats TEXF_RGBA7, TEXF_RGB16, TEXF_DXT1,
  //   TEXF_RGB8 and TEXF_RGBA8 are not supported yet.
  TUTObjectClassTexture = class(TUTObject_NoHeader)
  private
    FMipMaps: array of TBitmap;
    FCompMipMaps: array of TBitmap;
    function GetMipMapCount: integer;
    function GetMipMap(i: integer): TBitmap;
    function GetCompMipMapCount: integer;
    function GetCompMipMap(i: integer): TBitmap;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    procedure ReleaseObject; override;
    property MipMapCount: integer read GetMipMapCount;
    property MipMap[i: integer]: TBitmap read GetMipMap;
    property CompMipMapCount: integer read GetCompMipMapCount;
    property CompMipMap[i: integer]: TBitmap read GetCompMipMap;
    procedure SaveMipMapToFile(mipmap: integer; filename: string); virtual;
    procedure SaveCompMipMapToFile(mipmap: integer; filename: string); virtual;
  end;

  TUT_Struct_Spark = packed record
    SparkType, Heat: byte;
    X, Y: byte;
    X_Speed, Y_Speed: byte;
    Age, ExpTime: byte;
  end;

  // TUTObjectClassFireTexture
  // Status: completed.
  TUTObjectClassFireTexture = class(TUTObjectClassTexture)
  private
    FSparks: array of TUT_Struct_Spark;
    function GetSpark(i: integer): TUT_Struct_Spark;
    function GetSparkCount: integer;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    procedure ReleaseObject; override;
    property SparkCount: integer read GetSparkCount;
    property Spark[i: integer]: TUT_Struct_Spark read GetSpark;
  end;

  TUT_Struct_BoundingBox = packed record
    Min, Max: TUT_Struct_Vector;
    Valid: byte;
  end;
  TUT_Struct_BoundingSphere = packed record
    Center: TUT_Struct_Vector;
    Radius: single;
  end;

  // TUTObjectClassPrimitive
  // Status: completed
  TUTObjectClassPrimitive = class(TUTObject_NoHeader)
  protected
    FPrimitiveBoundingBox: TUT_Struct_BoundingBox;
    FPrimitiveBoundingSphere: TUT_Struct_BoundingSphere;
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    procedure ReleaseObject; override;
    property PrimitiveBoundingBox: TUT_Struct_BoundingBox read FPrimitiveBoundingBox;
    property PrimitiveBoundingSphere: TUT_Struct_BoundingSphere read FPrimitiveBoundingSphere;
  end;

  // TUT_MeshExporter
  // Helper class. Requires unique (Vertex,UV) pairs (duplicated vertex or UV when needed).
  TIntegerArray = array of integer;
  TUT_MeshExporter_Material = record
    Name: string;
    DiffuseColor: array[0..2] of byte;
  end;
  TUT_MeshExporter_Vertex = record
    X, Y, Z: single;
    U, V: byte;
  end;
  TUT_MeshExporter_Face = record
    VertexIndex1, VertexIndex2, VertexIndex3: integer;
    MaterialIndex: integer;
    Flags: integer;                     // face material flags (if any)
  end;
  TUT_MeshExporter = class
  public
    Vertices: array of TUT_MeshExporter_Vertex;
    Faces: array of TUT_MeshExporter_Face;
    Materials: array of TUT_MeshExporter_Material;
    AnimationFrames: integer;
  end;
  TUT_3DStudioExporter_Smoothing = (exp3ds_smooth_None, exp3ds_smooth_One, exp3ds_smooth_exp3ds_smooth_ByMaterial);
  TUT_3DStudioExporter = class(TUT_MeshExporter)
  public
    Smoothing: TUT_3DStudioExporter_Smoothing;
    MirrorX: boolean;
    procedure Save(filename: string);
  end;
  TUT_Unreal3DExporter = class(TUT_MeshExporter)
  public
    CoordsDivisor: single;
    procedure Save(filename: string);
  end;

  TUT_Struct_Vert = packed record
    X, Y, Z: single;
  end;
  TUT_Struct_Tri = packed record
    VertexIndex1, VertexIndex2, VertexIndex3: integer;
    U1, V1, U2, V2, U3, V3: byte;
    Flags, TextureIndex: integer;
  end;
  TUT_Struct_Texture = packed record
    Value: integer;
  end;
  TUT_Struct_AnimSeqFunction = packed record
    Time: single;
    _Function: integer;
  end;
  TUT_Struct_AnimSeq = packed record
    Name, Group: integer;
    StartFrame, NumFrames: integer;
    Rate: single;
    Functions: array of TUT_Struct_AnimSeqFunction;
  end;
  TUT_Struct_Connects = packed record
    NumVertTriangles, TriangleListOffset: integer;
  end;

  // TUTObjectClassMesh
  // Status: completed
  TUTObjectClassMesh = class(TUTObjectClassPrimitive)
  private
    function GetAnimSeq(i: integer): TUT_Struct_AnimSeq;
    function GetAnimSeqCount: integer;
    function GetAnimFrames: integer;
    function GetBoundingBox(i: integer): TUT_Struct_BoundingBox;
    function GetBoundingBoxCount: integer;
    function GetBoundingSphere(i: integer): TUT_Struct_BoundingSphere;
    function GetBoundingSphereCount: integer;
    function GetConnect(i: integer): TUT_Struct_Connects;
    function GetConnectsCount: integer;
    function GetTexture(i: integer): TUT_Struct_Texture;
    function GetTextureLOD(i: integer): single;
    function GetTextureLODCount: integer;
    function GetTexturesCount: integer;
    function GetTri(i: integer): TUT_Struct_Tri;
    function GetTrisCount: integer;
    function GetVert(i: integer): TUT_Struct_Vert;
    function GetVertLink(i: integer): integer;
    function GetVertLinksCount: integer;
    function GetVertsCount: integer;
  protected
    FVerts: array of TUT_Struct_Vert;
    FTris: array of TUT_Struct_Tri;
    FTextures: array of TUT_Struct_Texture;
    FAnimSeqs: array of TUT_Struct_AnimSeq;
    FConnects: array of TUT_Struct_Connects;
    FBoundingBox: TUT_Struct_BoundingBox;
    FBoundingSphere: TUT_Struct_BoundingSphere;
    FVertLinks: array of integer;
    FBoundingBoxes: array of TUT_Struct_BoundingBox;
    FBoundingSpheres: array of TUT_Struct_BoundingSphere;
    FScale, FOrigin: TUT_Struct_Vector;
    FRotOrigin: TUT_Struct_Rotator;
    FFrameVerts, FAnimFrames: integer;
    FANDFlags, FORFlags, FCurPoly, FCurVertex: integer;
    FTextureLOD: array of single;
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    procedure ReleaseObject; override;
    property VertCount: integer read GetVertsCount;
    property Verts[i: integer]: TUT_Struct_Vert read GetVert;
    property TrisCount: integer read GetTrisCount;
    property Tris[i: integer]: TUT_Struct_Tri read GetTri;
    property TextureCount: integer read GetTexturesCount;
    property Textures[i: integer]: TUT_Struct_Texture read GetTexture;
    property AnimSeqCount: integer read GetAnimSeqCount;
    property AnimSeqs[i: integer]: TUT_Struct_AnimSeq read GetAnimSeq;
    property ConnectCount: integer read GetConnectsCount;
    property Connects[i: integer]: TUT_Struct_Connects read GetConnect;
    property BoundingBox: TUT_Struct_BoundingBox read FBoundingBox;
    property BoundingSphere: TUT_Struct_BoundingSphere read FBoundingSphere;
    property VertLinksCount: integer read GetVertLinksCount;
    property VertLinks[i: integer]: integer read GetVertLink;
    property BoundingBoxCount: integer read GetBoundingBoxCount;
    property BoundingBoxes[i: integer]: TUT_Struct_BoundingBox read GetBoundingBox;
    property BoundingSphereCount: integer read GetBoundingSphereCount;
    property BoundingSpheres[i: integer]: TUT_Struct_BoundingSphere read GetBoundingSphere;
    property Scale: TUT_Struct_Vector read FScale;
    property Origin: TUT_Struct_Vector read FOrigin;
    property RotOrigin: TUT_Struct_Rotator read FRotOrigin;
    property FrameVerts: integer read FFrameVerts;
    property AnimFrames: integer read GetAnimFrames;
    property ANDFlags: integer read FANDFlags;
    property ORFlags: integer read FORFlags;
    property CurPoly: integer read FCurPoly;
    property CurVertex: integer read FCurVertex;
    property TextureLODCount: integer read GetTextureLODCount;
    property TextureLOD[i: integer]: single read GetTextureLOD;
    procedure Save_Unreal3D(filename: string); virtual;
    procedure Save_UnrealUC(filename: string); virtual;
    procedure PrepareExporter(exporter: TUT_MeshExporter; frames: TIntegerArray);
    procedure Save_3DS(filename: string; frame: integer = 0;
      smoothing: TUT_3DStudioExporter_Smoothing = exp3ds_smooth_None; MirrorX: boolean = false); virtual;
  end;

  TUT_Struct_Wedge = packed record
    VertexIndex: integer;
    U, V: byte;
  end;
  TUT_Struct_Face = packed record
    WedgeIndex1, WedgeIndex2, WedgeIndex3, MatIndex: integer;
  end;
  TUT_Struct_Material = packed record
    Flags, TextureIndex: integer;
  end;

  // TUTObjectClassLodMesh
  // Status: completed
  TUTObjectClassLodMesh = class(TUTObjectClassMesh)
  private
    function GetCollapsePointThus(i: integer): word;
    function GetCollapsePointThusCount: integer;
    function GetCollapseWedgeThus(i: integer): word;
    function GetCollapseWedgeThusCount: integer;
    function GetFace(i: integer): TUT_Struct_Face;
    function GetFaceCount: integer;
    function GetFaceLevel(i: integer): word;
    function GetFaceLevelCount: integer;
    function GetMaterial(i: integer): TUT_Struct_Material;
    function GetMaterialCount: integer;
    function GetRemapAnimVerts(i: integer): word;
    function GetRemapAnimVertsCount: integer;
    function GetSpecialFace(i: integer): TUT_Struct_Face;
    function GetSpecialFaceCount: integer;
    function GetWedge(i: integer): TUT_Struct_Wedge;
    function GetWedgeCount: integer;
  protected
    FWedges: array of TUT_Struct_Wedge;
    FFaces: array of TUT_Struct_Face;
    FMaterials: array of TUT_Struct_Material;
    FCollapsePointThus: array of word;
    FFaceLevel: array of word;
    FCollapseWedgeThus: array of word;
    FSpecialFaces: array of TUT_Struct_Face;
    FRemapAnimVerts: array of word;
    FModelVerts, FSpecialVerts, FOldFrameVerts: integer;
    FMeshScaleMax, FLODHysteresis, FLODStrength, FLODMinVerts, FLODMorph, FLODZDisplace: single;
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    procedure ReleaseObject; override;
    property WedgeCount: integer read GetWedgeCount;
    property Wedges[i: integer]: TUT_Struct_Wedge read GetWedge;
    property FaceCount: integer read GetFaceCount;
    property Faces[i: integer]: TUT_Struct_Face read GetFace;
    property MaterialCount: integer read GetMaterialCount;
    property Materials[i: integer]: TUT_Struct_Material read GetMaterial;
    property CollapsePointThusCount: integer read GetCollapsePointThusCount;
    property CollapsePointThus[i: integer]: word read GetCollapsePointThus;
    property FaceLevelCount: integer read GetFaceLevelCount;
    property FaceLevel[i: integer]: word read GetFaceLevel;
    property CollapseWedgeThusCount: integer read GetCollapseWedgeThusCount;
    property CollapseWedgeThus[i: integer]: word read GetCollapseWedgeThus;
    property SpecialFaceCount: integer read GetSpecialFaceCount;
    property SpecialFaces[i: integer]: TUT_Struct_Face read GetSpecialFace;
    property RemapAnimVertsCount: integer read GetRemapAnimVertsCount;
    property RemapAnimVerts[i: integer]: word read GetRemapAnimVerts;
    property ModelVerts: integer read FModelVerts;
    property SpecialVerts: integer read FSpecialVerts;
    property OldFrameVerts: integer read FOldFrameVerts;
    property MeshScaleMax: single read FMeshScaleMax;
    property LODHysteresis: single read FLODHysteresis;
    property LODStrength: single read FLODStrength;
    property LODMinVerts: single read FLODMinVerts;
    property LODMorph: single read FLODMorph;
    property LODZDisplace: single read FLODZDisplace;
    procedure Save_Unreal3D(filename: string); override;
    procedure Save_UnrealUC(filename: string); override;
    procedure PrepareExporter(exporter: TUT_MeshExporter; frames: TIntegerArray);
    procedure Save_3DS(filename: string; frame: integer = 0;
      smoothing: TUT_3DStudioExporter_Smoothing = exp3ds_smooth_None; MirrorX: boolean = false); override;
  end;

  TUT_Struct_MeshFloatUV = packed record
    U, V: single;
  end;

  TUT_Struct_MeshExtWedge = packed record
    iVertex: word;
    Flags: word;
    TexUV: TUT_Struct_MeshFloatUV;
  end;

  TUT_Struct_Quat = packed record
    X, Y, Z, W: single;
  end;

  TUT_Struct_JointPos = packed record
    Orientation: TUT_Struct_Quat;
    Position: TUT_Struct_Vector;
    Length: single;
    XSize: single;
    YSize: single;
    ZSize: single;
  end;

  TUT_Struct_MeshBone = packed record
    Name: integer;
    Flags: longword;
    BonePos: TUT_Struct_JointPos;
    NumChildren: integer;
    ParentIndex: integer;
    //Depth:integer; // not serialized?
  end;

  TUT_Struct_BoneInfIndex = packed record
    WeightIndex: word;
    Number: word;
    DetailA: word;
    DetailB: word;
  end;

  TUT_Struct_BoneInfluence = packed record
    PointIndex: word;
    BoneWeight: word;
  end;

  TUT_Struct_Coords = packed record
    Origin: TUT_Struct_Vector;
    XAxis: TUT_Struct_Vector;
    YAxis: TUT_Struct_Vector;
    ZAXis: TUT_Struct_Vector;
  end;

  // TUTObjectClassSkeletalMesh
  // Status: completed
  TUTObjectClassSkeletalMesh = class(TUTObjectClassLodMesh)
  private
    function GetBoneWeight(i: integer): TUT_Struct_BoneInfluence;
    function GetBoneWeightCount: integer;
    function GetBoneWeightIdx(i: integer): TUT_Struct_BoneInfIndex;
    function GetBoneWeightIdxCount: integer;
    function GetExtWedge(i: integer): TUT_Struct_MeshExtWedge;
    function GetExtWedgeCount: integer;
    function GetLocalPoint(i: integer): TUT_Struct_Vector;
    function GetLocalPointCount: integer;
    function GetPoint(i: integer): TUT_Struct_Vector;
    function GetPointCount: integer;
    function GetRefSkeleton(i: integer): TUT_Struct_MeshBone;
    function GetRefSkeletonCount: integer;
  protected
    FExtWedges: array of TUT_Struct_MeshExtWedge;
    FPoints: array of TUT_Struct_Vector;
    FRefSkeleton: array of TUT_Struct_MeshBone;
    FBoneWeightIdx: array of TUT_Struct_BoneInfIndex;
    FBoneWeights: array of TUT_Struct_BoneInfluence;
    FLocalPoints: array of TUT_Struct_Vector;
    FSkeletalDepth: integer;
    FDefaultAnimation: integer;
    FWeaponBoneIndex: integer;
    FWeaponAdjust: TUT_Struct_Coords;
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    procedure ReleaseObject; override;
    property ExtWedgeCount: integer read GetExtWedgeCount;
    property ExtWedges[i: integer]: TUT_Struct_MeshExtWedge read GetExtWedge;
    property PointCount: integer read GetPointCount;
    property Points[i: integer]: TUT_Struct_Vector read GetPoint;
    property RefSkeletonCount: integer read GetRefSkeletonCount;
    property RefSkeleton[i: integer]: TUT_Struct_MeshBone read GetRefSkeleton;
    property BoneWeightIdxCount: integer read GetBoneWeightIdxCount;
    property BoneWeightIdx[i: integer]: TUT_Struct_BoneInfIndex read GetBoneWeightIdx;
    property BoneWeightCount: integer read GetBoneWeightCount;
    property BoneWeights[i: integer]: TUT_Struct_BoneInfluence read GetBoneWeight;
    property LocalPointCount: integer read GetLocalPointCount;
    property LocalPoints[i: integer]: TUT_Struct_Vector read GetLocalPoint;
    property SkeletalDepth: integer read FSkeletalDepth;
    property DefaultAnimation: integer read FDefaultAnimation;
    property WeaponBoneIndex: integer read FWeaponBoneIndex;
    property WeaponAdjust: TUT_Struct_Coords read FWeaponAdjust;
    procedure Save_Unreal3D(filename: string); override;
    procedure Save_UnrealUC(filename: string); override;
    procedure PrepareExporter(exporter: TUT_MeshExporter; frames: TIntegerArray);
    procedure Save_3DS(filename: string; frame: integer = 0;
      smoothing: TUT_3DStudioExporter_Smoothing = exp3ds_smooth_None; MirrorX: boolean = false); override;
  end;

  // TUTObjectClassFunction
  // Status: completed
  TUTObjectClassFunction = class(TUTObjectClassStruct)
  private
    FiNative: integer;
    FRepOffset: integer;
    FOperatorPrecedence: integer;
    FFunctionFlags: longword;
    function GetFunctionFlags: longword;
    function GetiNative: integer;
    function GetOperatorPrecedence: integer;
    function GetRepOffset: integer;
  protected
    procedure InitializeObject; override;
    procedure InterpretObject; override;
  public
    procedure ReleaseObject; override;
    function Decompile(beautify: boolean = true): string;
    property NativeIndex: integer read GetiNative;
    property ReplicationOffset: integer read GetRepOffset;
    property OperatorPrecedence: integer read GetOperatorPrecedence;
    property FunctionFlags: longword read GetFunctionFlags;
  end;

  TUTImportTableObjectData = class
  private
    FObjectIndex: integer;
    FClassPackageIndex: integer;
    FClassIndex: integer;
    FPackageIndex: integer;
    FPackageName: string;
    FClassName: string;
    FClassPackageName: string;
    FObjectName: string;
    procedure SetObjectIndex(const Value: integer);
    procedure SetClassIndex(const Value: integer);
    procedure SetClassName(const Value: string);
    procedure SetClassPackageIndex(const Value: integer);
    procedure SetClassPackageName(const Value: string);
    procedure SetObjectName(const Value: string);
    procedure SetPackageIndex(const Value: integer);
    procedure SetPackageName(const Value: string);
  public
    // Read data
    property UTObjectIndex: integer read FObjectIndex write SetObjectIndex;
    property UTClassPackageIndex: integer read FClassPackageIndex write SetClassPackageIndex;
    property UTClassIndex: integer read FClassIndex write SetClassIndex;
    property UTPackageIndex: integer read FPackageIndex write SetPackageIndex;
    // Generated data
    property UTObjectName: string read FObjectName write SetObjectName;
    property UTClassPackageName: string read FClassPackageName write SetClassPackageName;
    property UTClassName: string read FClassName write SetClassName;
    property UTPackageName: string read FPackageName write SetPackageName;
  end;

  TUTExportTableObjectData = class
  private
    FSerialOffset: integer;
    FClassIndex: integer;
    FObjectIndex: integer;
    FSerialSize: integer;
    FPackageIndex: integer;
    FSuperIndex: integer;
    FFlags: integer;
    FPackageName: string;
    FClassName: string;
    FObjectName: string;
    FSuperName: string;
    FUTObject: TUTObject;
    procedure SetClassIndex(const Value: integer);
    procedure SetClassName(const Value: string);
    procedure SetFlags(const Value: integer);
    procedure SetObjectIndex(const Value: integer);
    procedure SetObjectName(const Value: string);
    procedure SetPackageIndex(const Value: integer);
    procedure SetPackageName(const Value: string);
    procedure SetSerialOffset(const Value: integer);
    procedure SetSerialSize(const Value: integer);
    procedure SetSuperIndex(const Value: integer);
    procedure SetSuperName(const Value: string);
    procedure SetUTObject(const Value: TUTObject);
  public
    // Read data
    property UTObjectIndex: integer read FObjectIndex write SetObjectIndex;
    property UTClassIndex: integer read FClassIndex write SetClassIndex;
    property UTPackageIndex: integer read FPackageIndex write SetPackageIndex;
    property UTSuperIndex: integer read FSuperIndex write SetSuperIndex;
    property Flags: integer read FFlags write SetFlags;
    property SerialSize: integer read FSerialSize write SetSerialSize;
    property SerialOffset: integer read FSerialOffset write SetSerialOffset;
    // Generated data
    property UTObjectName: string read FObjectName write SetObjectName;
    property UTClassName: string read FClassName write SetClassName;
    property UTPackageName: string read FPackageName write SetPackageName;
    property UTSuperName: string read FSuperName write SetSuperName;
    property UTObject: TUTObject read FUTObject write SetUTObject;
  end;

  TUTPackageObjectLocations = (utolNames, utolExports, utolImports);
  TUTPackageFindWhat = (utfwName, utfwPackage, utfwClass);
  TUTPackageFindWhatSet = set of TUTPackageFindWhat;
  TUT_OnProgressEvent = procedure(Sender: TObject; position: integer) of object;

  TUTPackage_GenerationInfo = record
    ExportCount: integer;
    NameCount: integer;
  end;

  // TUTPackage
  // Status: completed
  TUTPackage = class
  private
    FOnProgress: TUT_OnProgressEvent;
    Fstr: TStream;
    FPackage: string;
    FVersion: word;
    FLicenseeMode: word;
    FFlags: longword;
    FReadingPackageCount: integer;
    FNameTableList: tstringlist;
    FImportTableList: tlist;
    FExportTableList: tlist;
    FHeritageTableList: array of TGUID;
    FGenerationInfo: array of TUTPackage_GenerationInfo;
    procedure Process;
    procedure DoOnProgress(position, maxposition: integer);
    function GetExport(i: integer): TUTExportTableObjectData;
    function GetHeritage(i: integer): TGUID;
    function GetImport(i: integer): TUTImportTableObjectData;
    function GetName(i: integer): string;
    function GetNameFlags(i: integer): integer;
    function GetExportCount: integer;
    function GetHeritageCount: integer;
    function GetImportCount: integer;
    function GetNameCount: integer;
    function GetPackagePosition: integer;
    function GetStream: TStream;
    function GetExportIndex(objectname, classname: string): integer;
    function GetNameIndex(objectname: string): integer;
    function GetGeneration(i: integer): TUTPackage_GenerationInfo;
    function GetGenerationCount: integer;
  protected
    function IndentText(indent, txt: string): string;
  public
    property stream: TStream read GetStream;
    constructor Create(package: string = '');
    procedure Initialize(package: string);
    destructor Destroy; override;
    property OnProgress: TUT_OnProgressEvent read FOnProgress write FOnProgress;
    property Version: word read FVersion;
    property LicenseeMode: word read FLicenseeMode;
    property Flags: longword read FFlags;
    property NameCount: integer read GetNameCount;
    property Names[i: integer]: string read GetName;
    property NameFlags[i: integer]: integer read GetNameFlags;
    function GetObjectFlagsText(const e: cardinal): string;
    property ExportedCount: integer read GetExportCount;
    property Exported[i: integer]: TUTExportTableObjectData read GetExport;
    property ImportedCount: integer read GetImportCount;
    property Imported[i: integer]: TUTImportTableObjectData read GetImport;
    property HeritageCount: integer read GetHeritageCount;
    property Heritages[i: integer]: TGUID read GetHeritage;
    property GenerationCount: integer read GetGenerationCount;
    property Generations[i: integer]: TUTPackage_GenerationInfo read GetGeneration;
    property ExportIndex[objectname, classname: string]: integer read GetExportIndex;
    property NameIndex[objectname: string]: integer read GetNameIndex;
    function FindObject(where: TUTPackageObjectLocations; what: TUTPackageFindWhatSet; packagename, objectname, classname: string; start: integer = 0): integer;
    procedure ReadAllObjects;
    function EncodeIndex(i: integer): string;
    procedure StartReadingPackage;
    procedure EndReadingPackage;
    property Position: integer read GetPackagePosition;
    procedure Seek(p: integer);
    function ReadProperty(prop: TUTProperty; stream: tstream): boolean;
    procedure SaveDataBlock(filename: string; position, size: integer);
    function GetObjectPath(limit, index: integer): string;
    //function GetObjectPath_Simple(const index: integer): string;
    procedure read_buffer(var buffer; const size: integer; stream: tstream);
    function read_asciiz(stream: tstream): string;
    function read_bool(stream: tstream): boolean;
    function read_byte(stream: tstream): byte;
    function read_float(stream: tstream): single;
    function read_idx(stream: tstream): integer;
    function read_int(stream: tstream): integer;
    function read_qword(stream: tstream): int64;
    function read_guid(stream: tstream): TGuid;
    function read_sizedascii(stream: tstream): string;
    function read_doublesizedasciiz(stream: tstream): string;
    function read_sizedasciiz(stream: tstream): string;
    function read_word(stream: tstream): word;
    function Read_Name(stream: tstream): string;
  end;

  // Use this procedure to add classes to the package processor.
  // You can override existing registered classes.
procedure RegisterUTObjectClass(classname: string; classclass: TUTObjectClass);
procedure AddUTClassEquivalence(classname, equivalentclass: string);
procedure ClearUTClassEquivalences;

const
  PF_Masked = $00000002;                // Poly should be drawn masked.
  PF_Translucent = $00000004;           // Poly is transparent.
  PF_NotSolid = $00000008;              // Poly is not solid, doesn't block (Weapon)
  PF_Environment = $00000010;           // Poly should be drawn environment mapped.
  PF_Modulated = $00000040;             // Modulation transparency.
  PF_TwoSided = $00000100;              // Poly is visible from both sides.
  PF_NoSmooth = $00000800;              // Don't smooth textures.
  PF_Flat = $00004000;                  // Flat surface.
  PF_Unlit = $00400000;                 // Unlit.

  M3DMAGIC = $4D4D;
  M3D_VERSION = $0002;
  MDATA = $3D3D;
  MESH_VERSION = $3D3E;
  MAT_ENTRY = $AFFF;
  MAT_NAME = $A000;
  MAT_DIFFUSE = $A020;
  COLOR_24 = $0011;
  NAMED_OBJECT = $4000;
  N_TRI_OBJECT = $4100;
  POINT_ARRAY = $4110;
  TEX_VERTS = $4140;
  FACE_ARRAY = $4120;
  MSH_MAT_GROUP = $4130;
  SMOOTH_GROUP = $4150;
  KFDATA = $B000;
  KFHDR = $B00A;
  KFSEG = $B008;
  KFCURTIME = $B009;
  OBJECT_NODE_TAG = $B002;
  NODE_ID = $B030;
  NODE_HDR = $B010;
  PIVOT = $B013;
  POS_TRACK_TAG = $B020;
  ROT_TRACK_TAG = $B021;
  SCL_TRACK_TAG = $B022;
  HIDE_TRACK_TAG = $B029;
  FaceCAVisable3DS = $0001;
  FaceBCVisable3DS = $0002;
  FaceABVisable3DS = $0004;

  EX_LocalVariable = $00;
  EX_InstanceVariable = $01;
  EX_DefaultVariable = $02;
  // =$03
  EX_Return = $04;
  EX_Switch = $05;
  EX_Jump = $06;
  EX_JumpIfNot = $07;
  EX_Stop = $08;
  EX_Assert = $09;
  EX_Case = $0A;
  EX_Nothing = $0B;
  EX_LabelTable = $0C;
  EX_GotoLabel = $0D;
  EX_EatString = $0E;
  EX_Let = $0F;
  EX_DynArrayElement = $10;
  EX_New = $11;
  EX_ClassContext = $12;
  EX_Metacast = $13;
  EX_LetBool = $14;
  // =$15
  EX_EndFunctionParms = $16;
  EX_Self = $17;
  EX_Skip = $18;
  EX_Context = $19;
  EX_ArrayElement = $1A;
  EX_VirtualFunction = $1B;
  EX_FinalFunction = $1C;
  EX_IntConst = $1D;
  EX_FloatConst = $1E;
  EX_StringConst = $1F;
  EX_ObjectConst = $20;
  EX_NameConst = $21;
  EX_RotationConst = $22;
  EX_VectorConst = $23;
  EX_ByteConst = $24;
  EX_IntZero = $25;
  EX_IntOne = $26;
  EX_True = $27;
  EX_False = $28;
  EX_NativeParm = $29;
  EX_NoObject = $2A;
  // =$2B
  EX_IntConstByte = $2C;
  EX_BoolVariable = $2D;
  EX_DynamicCast = $2E;
  EX_Iterator = $2F;
  EX_IteratorPop = $30;
  EX_IteratorNext = $31;
  EX_StructCmpEq = $32;
  EX_StructCmpNe = $33;
  EX_UnicodeStringConst = $34;
  // =$35
  EX_StructMember = $36;
  // =$37
  EX_GlobalFunction = $38;
  EX_RotatorToVector = $39;
  EX_ByteToInt = $3A;
  EX_ByteToBool = $3B;
  EX_ByteToFloat = $3C;
  EX_IntToByte = $3D;
  EX_IntToBool = $3E;
  EX_IntToFloat = $3F;
  EX_BoolToByte = $40;
  EX_BoolToInt = $41;
  EX_BoolToFloat = $42;
  EX_FloatToByte = $43;
  EX_FloatToInt = $44;
  EX_FloatToBool = $45;
  EX_StringToName = $46;                // not defined in UT source, but used in unrealscript
  EX_ObjectToBool = $47;
  EX_NameToBool = $48;
  EX_StringToByte = $49;
  EX_StringToInt = $4A;
  EX_StringToBool = $4B;
  EX_StringToFloat = $4C;
  EX_StringToVector = $4D;
  EX_StringToRotator = $4E;
  EX_VectorToBool = $4F;
  EX_VectorToRotator = $50;
  EX_RotatorToBool = $51;
  EX_ByteToString = $52;
  EX_IntToString = $53;
  EX_BoolToString = $54;
  EX_FloatToString = $55;
  EX_ObjectToString = $56;
  EX_NameToString = $57;
  EX_VectorToString = $58;
  EX_RotatorToString = $59;

  EX_ExtendedNative = $60;
  EX_FirstNative = $70;

  NEST_Foreach = 1;
  NEST_Switch = 2;
  NEST_If = 3;
  NEST_Else = 4;

  nffFunction = 0;
  nffPreOperator = 1;
  nffPostOperator = 2;
  nffOperator = 3;

  // When changing this array decrease the OperatorPrecedence for "&&", "||", "^^"
  // operators to a value smaller than that of any other operator so that they will
  // have parenthesis around their operands when that operands are other operators.
type
  TNativeFunction = record
    Index: integer;
    Format: byte;
    OperatorPrecedence: byte;
    Name: string;
  end;
  TNativeFunctions=array of TNativeFunction;

var
   NativeFunctions:TNativeFunctions;

const
  NativeFunctions_UT: array[0..231] of TNativeFunction = (
    {0000}(Index: 112; Format: nffOperator; OperatorPrecedence: 40; Name: '$'), // Core.u
    {0001}(Index: 113; Format: nffFunction; OperatorPrecedence: 0; Name: 'GotoState'), // Core.u
    {0002}(Index: 114; Format: nffOperator; OperatorPrecedence: 24; Name: '=='), // Core.u
    {0003}(Index: 115; Format: nffOperator; OperatorPrecedence: 24; Name: '<'), // Core.u
    {0004}(Index: 116; Format: nffOperator; OperatorPrecedence: 24; Name: '>'), // Core.u
    {0005}(Index: 117; Format: nffFunction; OperatorPrecedence: 0; Name: 'Enable'), // Core.u
    {0006}(Index: 118; Format: nffFunction; OperatorPrecedence: 0; Name: 'Disable'), // Core.u
    {0007}(Index: 119; Format: nffOperator; OperatorPrecedence: 26; Name: '!='), // Core.u
    {0008}(Index: 120; Format: nffOperator; OperatorPrecedence: 24; Name: '<='), // Core.u
    {0009}(Index: 121; Format: nffOperator; OperatorPrecedence: 24; Name: '>='), // Core.u
    {0010}(Index: 122; Format: nffOperator; OperatorPrecedence: 24; Name: '=='), // Core.u
    {0011}(Index: 123; Format: nffOperator; OperatorPrecedence: 26; Name: '!='), // Core.u
    {0012}(Index: 124; Format: nffOperator; OperatorPrecedence: 24; Name: '~='), // Core.u
    {0013}(Index: 125; Format: nffFunction; OperatorPrecedence: 0; Name: 'Len'), // Core.u
    {0014}(Index: 126; Format: nffFunction; OperatorPrecedence: 0; Name: 'InStr'), // Core.u
    {0015}(Index: 127; Format: nffFunction; OperatorPrecedence: 0; Name: 'Mid'), // Core.u
    {0016}(Index: 128; Format: nffFunction; OperatorPrecedence: 0; Name: 'Left'), // Core.u
    {0017}(Index: 129; Format: nffPreOperator; OperatorPrecedence: 0; Name: '!'), // Core.u
    {0018}(Index: 130; Format: nffOperator; OperatorPrecedence: 10 {30}; Name: '&&'), // Core.u
    {0019}(Index: 131; Format: nffOperator; OperatorPrecedence: 10 {30}; Name: '^^'), // Core.u
    {0020}(Index: 132; Format: nffOperator; OperatorPrecedence: 11 {32}; Name: '||'), // Core.u
    {0021}(Index: 133; Format: nffOperator; OperatorPrecedence: 34; Name: '*='), // Core.u
    {0022}(Index: 134; Format: nffOperator; OperatorPrecedence: 34; Name: '/='), // Core.u
    {0023}(Index: 135; Format: nffOperator; OperatorPrecedence: 34; Name: '+='), // Core.u
    {0024}(Index: 136; Format: nffOperator; OperatorPrecedence: 34; Name: '-='), // Core.u
    {0025}(Index: 137; Format: nffPreOperator; OperatorPrecedence: 0; Name: '++'), // Core.u
    {0026}(Index: 138; Format: nffPreOperator; OperatorPrecedence: 0; Name: '--'), // Core.u
    {0027}(Index: 139; Format: nffPostOperator; OperatorPrecedence: 0; Name: '++'), // Core.u
    {0028}(Index: 140; Format: nffPostOperator; OperatorPrecedence: 0; Name: '--'), // Core.u
    {0029}(Index: 141; Format: nffPreOperator; OperatorPrecedence: 0; Name: '~'), // Core.u
    {0030}(Index: 142; Format: nffOperator; OperatorPrecedence: 24; Name: '=='), // Core.u
    {0031}(Index: 143; Format: nffPreOperator; OperatorPrecedence: 0; Name: '-'), // Core.u
    {0032}(Index: 144; Format: nffOperator; OperatorPrecedence: 16; Name: '*'), // Core.u
    {0033}(Index: 145; Format: nffOperator; OperatorPrecedence: 16; Name: '/'), // Core.u
    {0034}(Index: 146; Format: nffOperator; OperatorPrecedence: 20; Name: '+'), // Core.u
    {0035}(Index: 147; Format: nffOperator; OperatorPrecedence: 20; Name: '-'), // Core.u
    {0036}(Index: 148; Format: nffOperator; OperatorPrecedence: 22; Name: '<<'), // Core.u
    {0037}(Index: 149; Format: nffOperator; OperatorPrecedence: 22; Name: '>>'), // Core.u
    {0038}(Index: 150; Format: nffOperator; OperatorPrecedence: 24; Name: '<'), // Core.u
    {0039}(Index: 151; Format: nffOperator; OperatorPrecedence: 24; Name: '>'), // Core.u
    {0040}(Index: 152; Format: nffOperator; OperatorPrecedence: 24; Name: '<='), // Core.u
    {0041}(Index: 153; Format: nffOperator; OperatorPrecedence: 24; Name: '>='), // Core.u
    {0042}(Index: 154; Format: nffOperator; OperatorPrecedence: 24; Name: '=='), // Core.u
    {0043}(Index: 155; Format: nffOperator; OperatorPrecedence: 26; Name: '!='), // Core.u
    {0044}(Index: 156; Format: nffOperator; OperatorPrecedence: 28; Name: '&'), // Core.u
    {0045}(Index: 157; Format: nffOperator; OperatorPrecedence: 28; Name: '^'), // Core.u
    {0046}(Index: 158; Format: nffOperator; OperatorPrecedence: 28; Name: '|'), // Core.u
    {0047}(Index: 159; Format: nffOperator; OperatorPrecedence: 34; Name: '*='), // Core.u
    {0048}(Index: 160; Format: nffOperator; OperatorPrecedence: 34; Name: '/='), // Core.u
    {0049}(Index: 161; Format: nffOperator; OperatorPrecedence: 34; Name: '+='), // Core.u
    {0050}(Index: 162; Format: nffOperator; OperatorPrecedence: 34; Name: '-='), // Core.u
    {0051}(Index: 163; Format: nffPreOperator; OperatorPrecedence: 0; Name: '++'), // Core.u
    {0052}(Index: 164; Format: nffPreOperator; OperatorPrecedence: 0; Name: '--'), // Core.u
    {0053}(Index: 165; Format: nffPostOperator; OperatorPrecedence: 0; Name: '++'), // Core.u
    {0054}(Index: 166; Format: nffPostOperator; OperatorPrecedence: 0; Name: '--'), // Core.u
    {0055}(Index: 167; Format: nffFunction; OperatorPrecedence: 0; Name: 'Rand'), // Core.u
    {0056}(Index: 168; Format: nffOperator; OperatorPrecedence: 40; Name: '@'), // Core.u
    {0057}(Index: 169; Format: nffPreOperator; OperatorPrecedence: 0; Name: '-'), // Core.u
    {0058}(Index: 170; Format: nffOperator; OperatorPrecedence: 12; Name: '**'), // Core.u
    {0059}(Index: 171; Format: nffOperator; OperatorPrecedence: 16; Name: '*'), // Core.u
    {0060}(Index: 172; Format: nffOperator; OperatorPrecedence: 16; Name: '/'), // Core.u
    {0061}(Index: 173; Format: nffOperator; OperatorPrecedence: 18; Name: '%'), // Core.u
    {0062}(Index: 174; Format: nffOperator; OperatorPrecedence: 20; Name: '+'), // Core.u
    {0063}(Index: 175; Format: nffOperator; OperatorPrecedence: 20; Name: '-'), // Core.u
    {0064}(Index: 176; Format: nffOperator; OperatorPrecedence: 24; Name: '<'), // Core.u
    {0065}(Index: 177; Format: nffOperator; OperatorPrecedence: 24; Name: '>'), // Core.u
    {0066}(Index: 178; Format: nffOperator; OperatorPrecedence: 24; Name: '<='), // Core.u
    {0067}(Index: 179; Format: nffOperator; OperatorPrecedence: 24; Name: '>='), // Core.u
    {0068}(Index: 180; Format: nffOperator; OperatorPrecedence: 24; Name: '=='), // Core.u
    {0069}(Index: 181; Format: nffOperator; OperatorPrecedence: 26; Name: '!='), // Core.u
    {0070}(Index: 182; Format: nffOperator; OperatorPrecedence: 34; Name: '*='), // Core.u
    {0071}(Index: 183; Format: nffOperator; OperatorPrecedence: 34; Name: '/='), // Core.u
    {0072}(Index: 184; Format: nffOperator; OperatorPrecedence: 34; Name: '+='), // Core.u
    {0073}(Index: 185; Format: nffOperator; OperatorPrecedence: 34; Name: '-='), // Core.u
    {0074}(Index: 186; Format: nffFunction; OperatorPrecedence: 0; Name: 'Abs'), // Core.u
    {0075}(Index: 187; Format: nffFunction; OperatorPrecedence: 0; Name: 'Sin'), // Core.u
    {0076}(Index: 188; Format: nffFunction; OperatorPrecedence: 0; Name: 'Cos'), // Core.u
    {0077}(Index: 189; Format: nffFunction; OperatorPrecedence: 0; Name: 'Tan'), // Core.u
    {0078}(Index: 190; Format: nffFunction; OperatorPrecedence: 0; Name: 'Atan'), // Core.u
    {0079}(Index: 191; Format: nffFunction; OperatorPrecedence: 0; Name: 'Exp'), // Core.u
    {0080}(Index: 192; Format: nffFunction; OperatorPrecedence: 0; Name: 'Loge'), // Core.u
    {0081}(Index: 193; Format: nffFunction; OperatorPrecedence: 0; Name: 'Sqrt'), // Core.u
    {0082}(Index: 194; Format: nffFunction; OperatorPrecedence: 0; Name: 'Square'), // Core.u
    {0083}(Index: 195; Format: nffFunction; OperatorPrecedence: 0; Name: 'FRand'), // Core.u
    {0084}(Index: 196; Format: nffOperator; OperatorPrecedence: 22; Name: '>>>'), // Core.u
    {0085}(Index: 203; Format: nffOperator; OperatorPrecedence: 26; Name: '!='), // Core.u
    {0086}(Index: 210; Format: nffOperator; OperatorPrecedence: 24; Name: '~='), // Core.u
    {0087}(Index: 211; Format: nffPreOperator; OperatorPrecedence: 0; Name: '-'), // Core.u
    {0088}(Index: 212; Format: nffOperator; OperatorPrecedence: 16; Name: '*'), // Core.u
    {0089}(Index: 213; Format: nffOperator; OperatorPrecedence: 16; Name: '*'), // Core.u
    {0090}(Index: 214; Format: nffOperator; OperatorPrecedence: 16; Name: '/'), // Core.u
    {0091}(Index: 215; Format: nffOperator; OperatorPrecedence: 20; Name: '+'), // Core.u
    {0092}(Index: 216; Format: nffOperator; OperatorPrecedence: 20; Name: '-'), // Core.u
    {0093}(Index: 217; Format: nffOperator; OperatorPrecedence: 24; Name: '=='), // Core.u
    {0094}(Index: 218; Format: nffOperator; OperatorPrecedence: 26; Name: '!='), // Core.u
    {0095}(Index: 219; Format: nffOperator; OperatorPrecedence: 16; Name: 'Dot'), // Core.u
    {0096}(Index: 220; Format: nffOperator; OperatorPrecedence: 16; Name: 'Cross'), // Core.u
    {0097}(Index: 221; Format: nffOperator; OperatorPrecedence: 34; Name: '*='), // Core.u
    {0098}(Index: 222; Format: nffOperator; OperatorPrecedence: 34; Name: '/='), // Core.u
    {0099}(Index: 223; Format: nffOperator; OperatorPrecedence: 34; Name: '+='), // Core.u
    {0100}(Index: 224; Format: nffOperator; OperatorPrecedence: 34; Name: '-='), // Core.u
    {0101}(Index: 225; Format: nffFunction; OperatorPrecedence: 0; Name: 'VSize'), // Core.u
    {0102}(Index: 226; Format: nffFunction; OperatorPrecedence: 0; Name: 'Normal'), // Core.u
    {0103}(Index: 227; Format: nffFunction; OperatorPrecedence: 0; Name: 'Invert'), // Core.u
    {0104}(Index: 229; Format: nffFunction; OperatorPrecedence: 0; Name: 'GetAxes'), // Core.u
    {0105}(Index: 230; Format: nffFunction; OperatorPrecedence: 0; Name: 'GetUnAxes'), // Core.u
    {0106}(Index: 231; Format: nffFunction; OperatorPrecedence: 0; Name: 'Log'), // Core.u
    {0107}(Index: 232; Format: nffFunction; OperatorPrecedence: 0; Name: 'Warn'), // Core.u
    {0108}(Index: 233; Format: nffFunction; OperatorPrecedence: 0; Name: 'Error'), // Engine.u
    {0109}(Index: 234; Format: nffFunction; OperatorPrecedence: 0; Name: 'Right'), // Core.u
    {0110}(Index: 235; Format: nffFunction; OperatorPrecedence: 0; Name: 'Caps'), // Core.u
    {0111}(Index: 236; Format: nffFunction; OperatorPrecedence: 0; Name: 'Chr'), // Core.u
    {0112}(Index: 237; Format: nffFunction; OperatorPrecedence: 0; Name: 'Asc'), // Core.u
    {0113}(Index: 242; Format: nffOperator; OperatorPrecedence: 24; Name: '=='), // Core.u
    {0114}(Index: 243; Format: nffOperator; OperatorPrecedence: 26; Name: '!='), // Core.u
    {0115}(Index: 244; Format: nffFunction; OperatorPrecedence: 0; Name: 'FMin'), // Core.u
    {0116}(Index: 245; Format: nffFunction; OperatorPrecedence: 0; Name: 'FMax'), // Core.u
    {0117}(Index: 246; Format: nffFunction; OperatorPrecedence: 0; Name: 'FClamp'), // Core.u
    {0118}(Index: 247; Format: nffFunction; OperatorPrecedence: 0; Name: 'Lerp'), // Core.u
    {0119}(Index: 248; Format: nffFunction; OperatorPrecedence: 0; Name: 'Smerp'), // Core.u
    {0120}(Index: 249; Format: nffFunction; OperatorPrecedence: 0; Name: 'Min'), // Core.u
    {0121}(Index: 250; Format: nffFunction; OperatorPrecedence: 0; Name: 'Max'), // Core.u
    {0122}(Index: 251; Format: nffFunction; OperatorPrecedence: 0; Name: 'Clamp'), // Core.u
    {0123}(Index: 252; Format: nffFunction; OperatorPrecedence: 0; Name: 'VRand'), // Core.u
    {0124}(Index: 254; Format: nffOperator; OperatorPrecedence: 24; Name: '=='), // Core.u
    {0125}(Index: 255; Format: nffOperator; OperatorPrecedence: 26; Name: '!='), // Core.u
    {0126}(Index: 256; Format: nffFunction; OperatorPrecedence: 0; Name: 'Sleep'), // Engine.u
    {0127}(Index: 258; Format: nffFunction; OperatorPrecedence: 0; Name: 'ClassIsChildOf'), // Core.u
    {0128}(Index: 259; Format: nffFunction; OperatorPrecedence: 0; Name: 'PlayAnim'), // Engine.u
    {0129}(Index: 260; Format: nffFunction; OperatorPrecedence: 0; Name: 'LoopAnim'), // Engine.u
    {0130}(Index: 261; Format: nffFunction; OperatorPrecedence: 0; Name: 'FinishAnim'), // Engine.u
    {0131}(Index: 262; Format: nffFunction; OperatorPrecedence: 0; Name: 'SetCollision'), // Engine.u
    {0132}(Index: 263; Format: nffFunction; OperatorPrecedence: 0; Name: 'HasAnim'), // Engine.u
    {0133}(Index: 264; Format: nffFunction; OperatorPrecedence: 0; Name: 'PlaySound'), // Engine.u
    {0134}(Index: 266; Format: nffFunction; OperatorPrecedence: 0; Name: 'Move'), // Engine.u
    {0135}(Index: 267; Format: nffFunction; OperatorPrecedence: 0; Name: 'SetLocation'), // Engine.u
    {0136}(Index: 272; Format: nffFunction; OperatorPrecedence: 0; Name: 'SetOwner'), // Engine.u
    {0137}(Index: 275; Format: nffOperator; OperatorPrecedence: 22; Name: '<<'), // Core.u
    {0138}(Index: 276; Format: nffOperator; OperatorPrecedence: 22; Name: '>>'), // Core.u
    {0139}(Index: 277; Format: nffFunction; OperatorPrecedence: 0; Name: 'Trace'), // Engine.u
    {0140}(Index: 278; Format: nffFunction; OperatorPrecedence: 0; Name: 'Spawn'), // Engine.u
    {0141}(Index: 279; Format: nffFunction; OperatorPrecedence: 0; Name: 'Destroy'), // Engine.u
    {0142}(Index: 280; Format: nffFunction; OperatorPrecedence: 0; Name: 'SetTimer'), // Engine.u
    {0143}(Index: 281; Format: nffFunction; OperatorPrecedence: 0; Name: 'IsInState'), // Core.u
    {0144}(Index: 282; Format: nffFunction; OperatorPrecedence: 0; Name: 'IsAnimating'), // Engine.u
    {0145}(Index: 283; Format: nffFunction; OperatorPrecedence: 0; Name: 'SetCollisionSize'), // Engine.u
    {0146}(Index: 284; Format: nffFunction; OperatorPrecedence: 0; Name: 'GetStateName'), // Core.u
    {0147}(Index: 287; Format: nffOperator; OperatorPrecedence: 16; Name: '*'), // Core.u
    {0148}(Index: 288; Format: nffOperator; OperatorPrecedence: 16; Name: '*'), // Core.u
    {0149}(Index: 289; Format: nffOperator; OperatorPrecedence: 16; Name: '/'), // Core.u
    {0150}(Index: 290; Format: nffOperator; OperatorPrecedence: 34; Name: '*='), // Core.u
    {0151}(Index: 291; Format: nffOperator; OperatorPrecedence: 34; Name: '/='), // Core.u
    {0152}(Index: 293; Format: nffFunction; OperatorPrecedence: 0; Name: 'GetAnimGroup'), // Engine.u
    {0153}(Index: 294; Format: nffFunction; OperatorPrecedence: 0; Name: 'TweenAnim'), // Engine.u
    {0154}(Index: 296; Format: nffOperator; OperatorPrecedence: 16; Name: '*'), // Core.u
    {0155}(Index: 297; Format: nffOperator; OperatorPrecedence: 34; Name: '*='), // Core.u
    {0156}(Index: 298; Format: nffFunction; OperatorPrecedence: 0; Name: 'SetBase'), // Engine.u
    {0157}(Index: 299; Format: nffFunction; OperatorPrecedence: 0; Name: 'SetRotation'), // Engine.u
    {0158}(Index: 300; Format: nffFunction; OperatorPrecedence: 0; Name: 'MirrorVectorByNormal'), // Core.u
    {0159}(Index: 301; Format: nffFunction; OperatorPrecedence: 0; Name: 'FinishInterpolation'), // Engine.u
    {0160}(Index: 303; Format: nffFunction; OperatorPrecedence: 0; Name: 'IsA'), // Core.u
    {0161}(Index: 304; Format: nffFunction; OperatorPrecedence: 0; Name: 'AllActors'), // Engine.u
    {0162}(Index: 305; Format: nffFunction; OperatorPrecedence: 0; Name: 'ChildActors'), // Engine.u
    {0163}(Index: 306; Format: nffFunction; OperatorPrecedence: 0; Name: 'BasedActors'), // Engine.u
    {0164}(Index: 307; Format: nffFunction; OperatorPrecedence: 0; Name: 'TouchingActors'), // Engine.u
    {0165}(Index: 308; Format: nffFunction; OperatorPrecedence: 0; Name: 'ZoneActors'), // Engine.u
    {0166}(Index: 309; Format: nffFunction; OperatorPrecedence: 0; Name: 'TraceActors'), // Engine.u
    {0167}(Index: 310; Format: nffFunction; OperatorPrecedence: 0; Name: 'RadiusActors'), // Engine.u
    {0168}(Index: 311; Format: nffFunction; OperatorPrecedence: 0; Name: 'VisibleActors'), // Engine.u
    {0169}(Index: 312; Format: nffFunction; OperatorPrecedence: 0; Name: 'VisibleCollidingActors'), // Engine.u
    {0170}(Index: 314; Format: nffFunction; OperatorPrecedence: 0; Name: 'Warp'), // Engine.u
    {0171}(Index: 315; Format: nffFunction; OperatorPrecedence: 0; Name: 'UnWarp'), // Engine.u
    {0172}(Index: 316; Format: nffOperator; OperatorPrecedence: 20; Name: '+'), // Core.u
    {0173}(Index: 317; Format: nffOperator; OperatorPrecedence: 20; Name: '-'), // Core.u
    {0174}(Index: 318; Format: nffOperator; OperatorPrecedence: 34; Name: '+='), // Core.u
    {0175}(Index: 319; Format: nffOperator; OperatorPrecedence: 34; Name: '-='), // Core.u
    {0176}(Index: 320; Format: nffFunction; OperatorPrecedence: 0; Name: 'RotRand'), // Core.u
    {0177}(Index: 464; Format: nffFunction; OperatorPrecedence: 0; Name: 'StrLen'), // Engine.u
    {0178}(Index: 465; Format: nffFunction; OperatorPrecedence: 0; Name: 'DrawText'), // Engine.u
    {0179}(Index: 466; Format: nffFunction; OperatorPrecedence: 0; Name: 'DrawTile'), // Engine.u
    {0180}(Index: 467; Format: nffFunction; OperatorPrecedence: 0; Name: 'DrawActor'), // Engine.u
    {0181}(Index: 468; Format: nffFunction; OperatorPrecedence: 0; Name: 'DrawTileClipped'), // Engine.u
    {0182}(Index: 469; Format: nffFunction; OperatorPrecedence: 0; Name: 'DrawTextClipped'), // Engine.u
    {0183}(Index: 470; Format: nffFunction; OperatorPrecedence: 0; Name: 'TextSize'), // Engine.u
    {0184}(Index: 471; Format: nffFunction; OperatorPrecedence: 0; Name: 'DrawClippedActor'), // Engine.u
    {0185}(Index: 472; Format: nffFunction; OperatorPrecedence: 0; Name: 'DrawText'), // Engine.u
    {0186}(Index: 473; Format: nffFunction; OperatorPrecedence: 0; Name: 'DrawTile'), // Engine.u
    {0187}(Index: 474; Format: nffFunction; OperatorPrecedence: 0; Name: 'DrawColoredText'), // Engine.u
    {0188}(Index: 475; Format: nffFunction; OperatorPrecedence: 0; Name: 'ReplaceTexture'), // Engine.u
    {0189}(Index: 476; Format: nffFunction; OperatorPrecedence: 0; Name: 'TextSize'), // Engine.u
    {0190}(Index: 480; Format: nffFunction; OperatorPrecedence: 0; Name: 'DrawPortal'), // Engine.u
    {0191}(Index: 500; Format: nffFunction; OperatorPrecedence: 0; Name: 'MoveTo'), // Engine.u
    {0192}(Index: 502; Format: nffFunction; OperatorPrecedence: 0; Name: 'MoveToward'), // Engine.u
    {0193}(Index: 504; Format: nffFunction; OperatorPrecedence: 0; Name: 'StrafeTo'), // Engine.u
    {0194}(Index: 506; Format: nffFunction; OperatorPrecedence: 0; Name: 'StrafeFacing'), // Engine.u
    {0195}(Index: 508; Format: nffFunction; OperatorPrecedence: 0; Name: 'TurnTo'), // Engine.u
    {0196}(Index: 510; Format: nffFunction; OperatorPrecedence: 0; Name: 'TurnToward'), // Engine.u
    {0197}(Index: 512; Format: nffFunction; OperatorPrecedence: 0; Name: 'MakeNoise'), // Engine.u
    {0198}(Index: 514; Format: nffFunction; OperatorPrecedence: 0; Name: 'LineOfSightTo'), // Engine.u
    {0199}(Index: 517; Format: nffFunction; OperatorPrecedence: 0; Name: 'FindPathToward'), // Engine.u
    {0200}(Index: 518; Format: nffFunction; OperatorPrecedence: 0; Name: 'FindPathTo'), // Engine.u
    {0201}(Index: 519; Format: nffFunction; OperatorPrecedence: 0; Name: 'describeSpec'), // Engine.u
    {0202}(Index: 520; Format: nffFunction; OperatorPrecedence: 0; Name: 'actorReachable'), // Engine.u
    {0203}(Index: 521; Format: nffFunction; OperatorPrecedence: 0; Name: 'pointReachable'), // Engine.u
    {0204}(Index: 522; Format: nffFunction; OperatorPrecedence: 0; Name: 'ClearPaths'), // Engine.u
    {0205}(Index: 523; Format: nffFunction; OperatorPrecedence: 0; Name: 'EAdjustJump'), // Engine.u
    {0206}(Index: 524; Format: nffFunction; OperatorPrecedence: 0; Name: 'FindStairRotation'), // Engine.u
    {0207}(Index: 525; Format: nffFunction; OperatorPrecedence: 0; Name: 'FindRandomDest'), // Engine.u
    {0208}(Index: 526; Format: nffFunction; OperatorPrecedence: 0; Name: 'PickWallAdjust'), // Engine.u
    {0209}(Index: 527; Format: nffFunction; OperatorPrecedence: 0; Name: 'WaitForLanding'), // Engine.u
    {0210}(Index: 529; Format: nffFunction; OperatorPrecedence: 0; Name: 'AddPawn'), // Engine.u
    {0211}(Index: 530; Format: nffFunction; OperatorPrecedence: 0; Name: 'RemovePawn'), // Engine.u
    {0212}(Index: 531; Format: nffFunction; OperatorPrecedence: 0; Name: 'PickTarget'), // Engine.u
    {0213}(Index: 532; Format: nffFunction; OperatorPrecedence: 0; Name: 'PlayerCanSeeMe'), // Engine.u
    {0214}(Index: 533; Format: nffFunction; OperatorPrecedence: 0; Name: 'CanSee'), // Engine.u
    {0215}(Index: 534; Format: nffFunction; OperatorPrecedence: 0; Name: 'PickAnyTarget'), // Engine.u
    {0216}(Index: 536; Format: nffFunction; OperatorPrecedence: 0; Name: 'SaveConfig'), // Core.u
    {0217}(Index: 539; Format: nffFunction; OperatorPrecedence: 0; Name: 'GetMapName'), // Engine.u
    {0218}(Index: 540; Format: nffFunction; OperatorPrecedence: 0; Name: 'FindBestInventoryPath'), // Engine.u
    {0219}(Index: 544; Format: nffFunction; OperatorPrecedence: 0; Name: 'ResetKeyboard'), // Engine.u
    {0220}(Index: 545; Format: nffFunction; OperatorPrecedence: 0; Name: 'GetNextSkin'), // Engine.u
    {0221}(Index: 546; Format: nffFunction; OperatorPrecedence: 0; Name: 'UpdateURL'), // Engine.u
    {0222}(Index: 547; Format: nffFunction; OperatorPrecedence: 0; Name: 'GetURLMap'), // Engine.u
    {0223}(Index: 548; Format: nffFunction; OperatorPrecedence: 0; Name: 'FastTrace'), // Engine.u
    {0224}(Index: 549; Format: nffOperator; OperatorPrecedence: 20; Name: '-'), // Engine.u
    {0225}(Index: 550; Format: nffOperator; OperatorPrecedence: 16; Name: '*'), // Engine.u
    {0226}(Index: 551; Format: nffOperator; OperatorPrecedence: 20; Name: '+'), // Engine.u
    {0227}(Index: 552; Format: nffOperator; OperatorPrecedence: 16; Name: '*'), // Engine.u
    {0228}(Index: 1033; Format: nffFunction; OperatorPrecedence: 0; Name: 'RandRange'), // Core.u
    {0229}(Index: 3969; Format: nffFunction; OperatorPrecedence: 0; Name: 'MoveSmooth'), // Engine.u
    {0230}(Index: 3970; Format: nffFunction; OperatorPrecedence: 0; Name: 'SetPhysics'), // Engine.u
    {0231}(Index: 3971; Format: nffFunction; OperatorPrecedence: 0; Name: 'AutonomousPhysics') // Engine.u
    );

  // Property Flags
  CPF_Edit = $00000001;                 // Property is user-settable in the editor.
  CPF_Const = $00000002;                // Actor's property always matches class's default actor property.
  CPF_Input = $00000004;                // Variable is writable by the input system.
  CPF_ExportObject = $00000008;         // Object can be exported with actor.
  CPF_OptionalParm = $00000010;         // Optional parameter (if CPF_Param is set).
  CPF_Net = $00000020;                  // Property is relevant to network replication.
  CPF_ConstRef = $00000040;             // Reference to a constant object.
  CPF_Parm = $00000080;                 // Function/When call parameter.
  CPF_OutParm = $00000100;              // Value is copied out after function call.
  CPF_SkipParm = $00000200;             // Property is a short-circuitable evaluation function parm.
  CPF_ReturnParm = $00000400;           // Return value.
  CPF_CoerceParm = $00000800;           // Coerce args into this function parameter.
  CPF_Native = $00001000;               // Property is native: C++ code is responsible for serializing it.
  CPF_Transient = $00002000;            // Property is transient: shouldn't be saved, zero-filled at load time.
  CPF_Config = $00004000;               // Property should be loaded/saved as permanent profile.
  CPF_Localized = $00008000;            // Property should be loaded as localizable text.
  CPF_Travel = $00010000;               // Property travels across levels/servers.
  CPF_EditConst = $00020000;            // Property is uneditable in the editor.
  CPF_GlobalConfig = $00040000;         // Load config from base class, not subclass.
  CPF_OnDemand = $00100000;             // Object or dynamic array loaded on demand only.
  CPF_New = $00200000;                  // Automatically create inner object.
  CPF_NeedCtorLink = $00400000;         // Fields need construction/destruction.
  // NOTE : for vars with no category specified, the category name is the class

 // Function flags.
  FUNC_Final = $00000001;               // Function is final (prebindable, non-overridable function).
  FUNC_Defined = $00000002;             // Function has been defined (not just declared).
  FUNC_Iterator = $00000004;            // Function is an iterator.
  FUNC_Latent = $00000008;              // Function is a latent state function.
  FUNC_PreOperator = $00000010;         // Unary operator is a prefix operator.
  FUNC_Singular = $00000020;            // Function cannot be reentered.
  FUNC_Net = $00000040;                 // Function is network-replicated.
  FUNC_NetReliable = $00000080;         // Function should be sent reliably on the network.
  FUNC_Simulated = $00000100;           // Function executed on the client side.
  FUNC_Exec = $00000200;                // Executable from command line.
  FUNC_Native = $00000400;              // Native function.
  FUNC_Event = $00000800;               // Event function.
  FUNC_Operator = $00001000;            // Operator function.
  FUNC_Static = $00002000;              // Static function.
  FUNC_NoExport = $00004000;            // Don't export intrinsic function to C++.
  FUNC_Const = $00008000;               // Function doesn't modify this object.
  FUNC_Invariant = $00010000;           // Return value is purely dependent on parameters; no state dependencies or internal state changes.

  // Base flags.
  CLASS_Abstract = $00001;              // Class is abstract and can't be instantiated directly.
  CLASS_Compiled = $00002;              // Script has been compiled successfully.
  CLASS_Config = $00004;                // Load object configuration at construction time.
  CLASS_Transient = $00008;             // This object type can't be saved; null it out at save time.
  CLASS_Parsed = $00010;                // Successfully parsed.
  CLASS_Localized = $00020;             // Class contains localized text.
  CLASS_SafeReplace = $00040;           // Objects of this class can be safely replaced with default or NULL.
  CLASS_RuntimeStatic = $00080;         // Objects of this class are static during gameplay.
  CLASS_NoExport = $00100;              // Don't export to C++ header.
  CLASS_NoUserCreate = $00200;          // Don't allow users to create in the editor.
  CLASS_PerObjectConfig = $00400;       // Handle object configuration on a per-object basis, rather than per-class.
  CLASS_NativeReplication = $00800;     // Replication handled in C++

  // State flags.
  STATE_Editable = $00000001;           // State should be user-selectable in UnrealEd.
  STATE_Auto = $00000002;               // State is automatic (the default state).
  STATE_Simulated = $00000004;          // State executes on client side.

implementation

resourcestring
  rsUnknownPropertyType = 'Unknown type 0x%-2.2x, scan stopped';
  rsUnknownStruct = 'Unknown struct "%s", scan stopped';
  rsErrorNoUTPackage = 'ERROR! This isn''t an Unreal package';
  rsExceptionProcessingPackage = 'Exception processing package %s';
  rsExceptionReadingProperty = 'Exception reading property %s';
  rsUnknown = 'Unknown';
  rsWarning = 'Warning!';
  rsNotImplemented = 'Not Implemented';
  rsUnknownOpcode = 'Unknown OpCode 0x%-2.2x. Will try to continue.';
  rsInvalidNativeIndex = 'Invalid native function index 0x%-2.2x';

type
  TUTClassRegistry = record
    class_name: string;
    class_class: TUTObjectClass;
  end;
  TUTClassEquivalence = record
    class_name: string;
    equivalent_class_name: string;
  end;

var
  RegisteredUTClasses: array of TUTClassRegistry;
  UTClassEquivalences: array of TUTClassEquivalence;

procedure AddUTClassEquivalence(classname, equivalentclass: string);
begin
  setlength(UTClassEquivalences, length(UTClassEquivalences) + 1);
  with UTClassEquivalences[high(UTClassEquivalences)] do
    begin
      class_name := lowercase(classname);
      equivalent_class_name := lowercase(equivalentclass);
    end;
end;

procedure ClearUTClassEquivalences;
begin
  setlength(UTClassEquivalences, 0);
end;

procedure RegisterUTObjectClass(classname: string; classclass: TUTObjectClass);
var
  a, n: integer;
begin
  classname := lowercase(classname);
  a := 0;
  n := -1;
  while (a <= high(RegisteredUTClasses)) and (n = -1) do
    if RegisteredUTClasses[a].class_name = classname then
      n := a
    else
      inc(a);
  if n = -1 then
    begin
      setlength(RegisteredUTClasses, length(RegisteredUTClasses) + 1);
      n := high(RegisteredUTClasses);
    end;
  with RegisteredUTClasses[n] do
    begin
      class_name := classname;
      class_class := classclass;
    end;
end;

function GeTUTObjectClass(classname: string): TUTObjectClass;
var
  a: integer;
begin
  classname := lowercase(classname);
  for a := 0 to high(UTClassEquivalences) do
    if UTClassEquivalences[a].class_name = classname then
      begin
        classname := UTClassEquivalences[a].equivalent_class_name;
        break;
      end;
  a := 0;
  result := TUTObject;
  while (a < length(RegisteredUTClasses)) and (result = TUTObject) do
    if RegisteredUTClasses[a].class_name = classname then
      result := RegisteredUTClasses[a].class_class
    else
      inc(a);
end;

{ TUTProperty }

function TUTProperty.GetDescription: string;
var
  a, b, l: integer;
  vn, vd, vs, v: string;
  vv: variant;
  vt: TUTPropertyType;
  ds: char;
begin
  if not FIsInitialized then
    begin
      result := '';
      exit;
    end;
  ds := DecimalSeparator;
  DecimalSeparator := '.';
  result := name;
  if (FArrayIndex > -1) then result := result + format('[%d]', [FArrayIndex]);
  if specifictypename <> '' then result := result + ' (' + specifictypename + ')';
  if propertytype <> otNone then
    begin
      vs := '';
      for a := 0 to ValueCount - 1 do
        begin
          GetValue(a, vn, vv, vd, vt);
          case vt of
            otNone:
              begin
                v := '{None}';
              end;
            otObject, otName:
              begin
                b := vv;
                v := format('%d', [b]);
              end;
            otByte, otInt, otWord:
              begin
                if vt = otByte then
                  l := 2
                else if vt = otWord then
                  l := 4
                else
                  l := 8;
                b := vv;
                v := format('0x%-.*x', [l, b]);
              end;
            otFloat:
              begin
                ds := DecimalSeparator;
                DecimalSeparator := '.';
                v := format('%f', [single(vv)]);
                DecimalSeparator := ds;
              end;
            otBool:
              begin
                if vv then
                  v := 'True'
                else
                  v := 'False';
              end;
            otStr, otString:
              begin
                v := '"' + vv + '"';
              end;
            otBuffer:
              begin
                v := '';
                for b := 0 to length(vv) - 1 do
                  v := v + format('%-2.2x ', [vv[b]]);
              end
          else
            v := '{unknown type}';
          end;
          if PropertyType = otStruct then
            vs := vs + ' ; ' + vn + '=' + v
          else
            vs := vs + ' ; ' + v;
          if vd <> '' then vs := vs + ' (' + vd + ')';
        end;
      if vs <> '' then delete(vs, 1, 3);
      if vs <> '' then result := result + ' = ';
      if PropertyType = otStruct then
        result := result + '< ' + vs + ' >'
      else
        result := result + vs;
    end;
  DecimalSeparator := ds;
end;

function TUTProperty.GetFirstValue: variant;
var
  name, description: string;
  _type: TUTPropertyType;
begin
  GetValue(0, name, result, description, _type);
end;

function TUTProperty.GetValueTypeName(t: TUTPropertyType): string;
begin
  case t of
    otNone: result := '';
    otByte: result := 'Byte';
    otInt: result := 'Int';
    otBool: result := 'Bool';
    otFloat: result := 'Float';
    otObject: result := 'Object';
    otName: result := 'Name';
    otString: result := 'String';
    otClass: result := 'Class';
    otArray: result := 'Array';
    otStruct: result := FTypeName;
    otVector: result := 'Vector';
    otRotator: result := 'Rotator';
    otStr: result := 'Str';
    otMap: result := 'Map';
    otFixedArray: result := 'FixedArray';

    otWord: result := 'Word';
    otBuffer:
      result := 'Buffer'
  else
    result := '';
  end;
end;

function TUTProperty.GetTypeName: string;
begin
  if FTypeName <> '' then
    result := FTypeName
  else
    result := GetValueTypeName(PropertyType);
end;

procedure TUTProperty.GetValue(i: integer; var valuename: string;
  var value: variant; var valuedescription: string;
  var valuetype: TUTPropertyType);
var
  w: word;
  int: integer;
  s: single;
  dbl: double;
  t, enum, lcname: string;
  bo: boolean;
  b: byte;
  rgba: array[0..3] of byte;
  vector: array[0..2] of single;
  plane: array[0..3] of single;
  rotator: array[0..2] of integer;
  pointregion: packed record
    zone: integer;
    ileaf: integer;
    zonenumber: integer;
  end;
  scale: packed record
    x, y, z, sheerrate: single;
    sheeraxis: byte;
  end;
  adropspark: packed record
    _type, heat: byte;
    x, y: byte;
    x_speed, y_speed: byte;
    Age, ExpTime: byte;
  end;
const                                   // do not localize this strings
  enumSheerAxis: array[0..6] of string = ('SHEER_None', 'SHEER_XY', 'SHEER_XZ', 'SHEER_YX', 'SHEER_YZ', 'SHEER_ZX', 'SHEER_ZY');
  enumCsgOper: array[0..4] of string = ('CSG_Active', 'CSG_Add', 'CSG_Substract', 'CSG_Intersect', 'CSG_Deintersect');
  enumDrawType: array[0..7] of string = ('DT_None', 'DT_Sprite', 'DT_Mesh', 'DT_Brush', 'DT_RopeSprite', 'DT_VerticalSprite', 'DT_TerraForm', 'DT_SpriteAnimOnce');
  enumStyle: array[0..4] of string = ('STY_None', 'STY_Normal', 'STY_Masked', 'STY_Translucent', 'STY_Modulated');
  enumLightEffect: array[0..19] of string = ('LE_None', 'LE_TorchWaver', 'LE_FireWaver', 'LE_WateryShimmer', 'LE_Searchlight', 'LE_SlowWave', 'LE_FastWave', 'LE_CloudCast', 'LE_StaticSpot', 'LE_Shock', 'LE_Disco', 'LE_Warp', 'LE_Spotlight', 'LE_NonIncidence', 'LE_Shell', 'LE_OmniBumpMap', 'LE_Interference', 'LE_Cylinder', 'LE_Rotor', 'LE_Unused');
  enumLightType: array[0..9] of string = ('LT_None', 'LT_Steady', 'LT_Pulse', 'LT_Blink', 'LT_Flicker', 'LT_Strobe', 'LT_BackdropLight', 'LT_SubtlePulse', 'LT_TexturePaletteOnce', 'LT_TexturePaletteLoop');
  enumPhysics: array[0..11] of string = ('PHYS_None', 'PHYS_Walking', 'PHYS_Falling', 'PHYS_Swimming', 'PHYS_Flying', 'PHYS_Rotating', 'PHYS_Projectile', 'PHYS_Rolling', 'PHYS_Interpolating', 'PHYS_MovingBrush', 'PHYS_Spider', 'PHYS_Trailer');
  enumRemoteRole: array[0..4] of string = ('ROLE_None', 'ROLE_DumbProxy', 'ROLE_SimulatedProxy', 'ROLE_AutonomousProxy', 'ROLE_Authority');
  enumDropType: array[0..19] of string = ('DROP_FixedDepth', 'DROP_PhaseSpot', 'DROP_ShallowSpot', 'DROP_HalfAmpl', 'DROP_RandomMover', 'DROP_FixedRandomMover', 'DROP_WhirlyThing', 'DROP_BigWhirly', 'DROP_HorizontalLine', 'DROP_VerticalLine', 'DROP_DiagonalLine1', 'DROP_DiagonalLine2', 'DROP_HorizontalOsc', 'DROP_VerticalOsc', 'DROP_DiagonalOsc1', 'DROP_DiagonalOsc2', 'DROP_RainDrops', 'DROP_AreaClamp', 'DROP_LeakyTap', 'DROP_DrippyTap');
  enumDrawMode: array[0..4] of string = ('DRAW_Normal', 'DRAW_Lathe', 'DRAW_Lathe_2', 'DRAW_Lathe_3', 'DRAW_Lathe_4');
  enumSparkType: array[0..28] of string = ('SPARK_Burn', 'SPARK_Sparkle', 'SPARK_Pulse', 'SPARK_Signal', 'SPARK_Blaze', 'SPARK_OzHasSpoken', 'SPARK_Cone', 'SPARK_BlazeRight', 'SPARK_BlazeLeft', 'SPARK_Cylinder', 'SPARK_Cylinder3D', 'SPARK_Lissajous', 'SPARK_Jugglers', 'SPARK_Emit', 'SPARK_Fountain', 'SPARK_Flocks', 'SPARK_Eels', 'SPARK_Organic', 'SPARK_WanderOrganic', 'SPARK_RandomCloud', ',SPARK_CustomCloud', 'SPARK_LocalCloud', 'SPARK_Starts', 'SPARK_LineLightning', 'SPARK_RampLightning', 'SPARK_SphereLightning', 'SPARK_Wheel', 'SPARK_Gametes', 'SPARK_Sprinkler');
  enumTimeDistribution: array[0..2] of string = ('DIST_Constant', 'DIST_Uniform', 'DIST_Gaussian');
  enumAttitudeToPlayer: array[0..6] of string = ('ATTITUDE_Fear', 'ATTITUDE_Hate', 'ATTITUDE_Frenzy', 'ATTITUDE_Threaten', 'ATTITUDE_Ignore', 'ATTITUDE_Friendly', 'ATTITUDE_Follow');
  enumIntelligence: array[0..3] of string = ('BRAINS_None', 'BRAINS_Reptile', 'BRAINS_Mammal', 'BRAINS_Human');
  enumBumpType: array[0..2] of string = ('BT_PlayerBump', 'BT_PawnBump', 'BT_AnyBump');
  enumMoverEncroachType: array[0..3] of string = ('ME_StopWhenEncroach', 'ME_ReturnWhenEncroach', 'ME_CrushWhenEncroach', 'ME_IgnoreWhenEncroach');
  enumMoverGlideType: array[0..1] of string = ('MV_MoveByTime', 'MV_GlideByTime');
  enumLODSet: array[0..2] of string = ('LODSET_None', 'LODSET_World', 'LODSET_Skin');
  enumCompFormat: array[0..5] of string = ('TEXF_P8', 'TEXF_RGBA7', 'TEXF_RGB16', 'TEXF_DXT1', 'TEXF_RGB8', 'TEXF_RGBA8');
  // can also be: TEXF_P8 ,  TEXF_RGB32 ,  TEXF_RGB64 ,  TEXF_DXT1 ,  TEXF_RGB24
  // TODO : make enums global and complete them
begin
  if self = nil then exit;
  if not FIsInitialized then
    begin
      valuename := '';
      value := '0';
      valuedescription := '';
      valuetype := otNone;
      exit;
    end;
  valuename := name;
  value := 0;
  valuedescription := '';
  valuetype := PropertyType;
  case PropertyType of
    otNone:
      begin
        value := 0;
      end;
    otByte:
      begin
        b := FValue[0];
        lcname := lowercase(name);
        if (lcname = 'csgoper') and (b <= high(enumCsgOper)) then
          enum := enumCsgOper[b]
        else if (lcname = 'drawtype') and (b <= high(enumDrawType)) then
          enum := enumDrawType[b]
        else if (lcname = 'lighteffect') and (b <= high(enumLightEffect)) then
          enum := enumLightEffect[b]
        else if (lcname = 'lighttype') and (b <= high(enumLightType)) then
          enum := enumLightType[b]
        else if (lcname = 'physics') and (b <= high(enumPhysics)) then
          enum := enumPhysics[b]
        else if (lcname = 'remoterole') and (b <= high(enumRemoteRole)) then
          enum := enumRemoteRole[b]
        else if (lcname = 'droptype') and (b <= high(enumDropType)) then
          enum := enumDropType[b]
        else if (lcname = 'drawmode') and (b <= high(enumDrawMode)) then
          enum := enumDrawMode[b]
        else if (lcname = 'sparktype') and (b <= high(enumSparkType)) then
          enum := enumSparkType[b]
        else if ((lcname = 'style') or (lcname = 'muzzleflashstyle')) and (b <= high(enumStyle)) then
          enum := enumStyle[b]
        else if (lcname = 'timedistribution') and (b <= high(enumTimeDistribution)) then
          enum := enumTimeDistribution[b]
        else if (lcname = 'attitudetoplayer') and (b <= high(enumAttitudeToPlayer)) then
          enum := enumAttitudeToPlayer[b]
        else if (lcname = 'intelligence') and (b <= high(enumIntelligence)) then
          enum := enumIntelligence[b]
        else if (lcname = 'bumptype') and (b <= high(enumBumpType)) then
          enum := enumBumpType[b]
        else if (lcname = 'moverencroachtype') and (b <= high(enumMoverEncroachType)) then
          enum := enumMoverEncroachType[b]
        else if (lcname = 'moverglidetype') and (b <= high(enumMoverGlideType)) then
          enum := enumMoverGlideType[b]
        else if (lcname = 'lodset') and (b <= high(enumLODSet)) then
          enum := enumLODSet[b]
        else if (lcname = 'compformat') and (b <= high(enumCompFormat)) then
          enum := enumCompFormat[b]
        else
          enum := '';                   //format('0x%-2.2x',[b]);
        value := b;
        valuedescription := enum;
      end;
    otObject:
      begin
        move(FValue[0], int, 4);
        value := int;
        if int = 0 then
          valuedescription := 'None'
        else
          begin
            if int >= 0 then
              valuedescription := FOwner.Exported[int - 1].ClassName
            else
              valuedescription := FOwner.Imported[-int - 1].ClassName;
            if valuedescription = '' then valuedescription := 'Class';
            valuedescription := valuedescription + '''' + FOwner.GetObjectPath(-1, int);
          end;
      end;
    otInt:
      begin
        move(FValue[0], int, 4);
        value := int;
      end;
    otFloat:
      begin
        move(FValue[0], s, 4);
        dbl := s;
        value := dbl;
      end;
    otBool:
      begin
        move(FValue[0], bo, 1);
        value := bo;
      end;
    otWord:
      begin
        move(FValue[0], w, 2);
        value := w;
      end;
    otName:
      begin
        move(FValue[0], int, 4);
        value := int;
        valuedescription := FOwner.FNameTableList[int];
      end;
    otStr, otString:
      begin
        setlength(t, length(FValue));
        move(FValue[0], t[1], length(FValue));
        while (t <> '') and (t[length(t)] = #0) do
          delete(t, length(t), 1);
        value := t;
      end;
    otBuffer:
      begin
        value := FValue;
      end;
    otStruct:
      begin
        if lowercase(FTypeName) = 'color' then
          begin
            move(FValue[0], rgba[0], sizeof(rgba));
            case i of
              0:
                begin
                  valuename := 'Red';
                  valuetype := otByte;
                end;
              1:
                begin
                  valuename := 'Green';
                  valuetype := otByte;
                end;
              2:
                begin
                  valuename := 'Blue';
                  valuetype := otByte;
                end;
              3:
                begin
                  valuename := 'Alpha';
                  valuetype := otByte;
                end;
            end;
            value := rgba[i];
          end
        else if lowercase(FTypeName) = 'vector' then
          begin
            move(FValue[0], vector, sizeof(vector));
            case i of
              0:
                begin
                  valuename := 'X';
                  valuetype := otFloat;
                end;
              1:
                begin
                  valuename := 'Y';
                  valuetype := otFloat;
                end;
              2:
                begin
                  valuename := 'Z';
                  valuetype := otFloat;
                end;
            end;
            value := vector[i];
          end
        else if lowercase(FTypeName) = 'rotator' then
          begin
            move(FValue[0], rotator, sizeof(rotator));
            case i of
              0:
                begin
                  valuename := 'Pitch';
                  valuetype := otInt;
                end;
              1:
                begin
                  valuename := 'Yaw';
                  valuetype := otInt;
                end;
              2:
                begin
                  valuename := 'Roll';
                  valuetype := otInt;
                end;
            end;
            value := rotator[i];
          end
        else if lowercase(FTypeName) = 'pointregion' then
          begin
            move(FValue[0], pointregion, sizeof(pointregion));
            case i of
              0:
                begin
                  valuename := 'Zone';
                  value := pointregion.zone;
                  valuetype := otObject;
                end;
              1:
                begin
                  valuename := 'iLeaf';
                  value := pointregion.ileaf;
                  valuetype := otInt;
                end;
              2:
                begin
                  valuename := 'ZoneNumber';
                  value := pointregion.zonenumber;
                  valuetype := otInt;
                end;
            end;
          end
        else if lowercase(FTypeName) = 'scale' then
          begin
            move(FValue[0], scale, sizeof(scale));
            case i of
              0:
                begin
                  valuename := 'X';
                  value := scale.x;
                  valuetype := otFloat;
                end;
              1:
                begin
                  valuename := 'Y';
                  value := scale.y;
                  valuetype := otFloat;
                end;
              2:
                begin
                  valuename := 'Z';
                  value := scale.z;
                  valuetype := otFloat;
                end;
              3:
                begin
                  valuename := 'SheerRate';
                  value := scale.sheerrate;
                  valuetype := otFloat;
                end;
              4:
                begin
                  valuename := 'SheerAxis';
                  value := scale.sheeraxis;
                  valuedescription := enumSheerAxis[scale.sheeraxis];
                  valuetype := otByte;
                end;
            end;
          end
        else if lowercase(FTypeName) = 'plane' then
          begin
            move(FValue[0], plane, sizeof(plane));
            case i of
              0:
                begin
                  valuename := 'X';
                  valuetype := otFloat;
                end;
              1:
                begin
                  valuename := 'Y';
                  valuetype := otFloat;
                end;
              2:
                begin
                  valuename := 'Z';
                  valuetype := otFloat;
                end;
              3:
                begin
                  valuename := 'W';
                  valuetype := otFloat;
                end;
            end;
            value := vector[i];
          end
        else if lowercase(FTypeName) = 'sphere' then
          begin
            move(FValue[0], plane, sizeof(plane));
            case i of
              0:
                begin
                  valuename := 'X';
                  valuetype := otFloat;
                end;
              1:
                begin
                  valuename := 'Y';
                  valuetype := otFloat;
                end;
              2:
                begin
                  valuename := 'Z';
                  valuetype := otFloat;
                end;
              3:
                begin
                  valuename := 'W';
                  valuetype := otFloat;
                end;
            end;
            value := vector[i];
          end
        else if (lowercase(FTypeName) = 'adrop') or (lowercase(FTypeName) = 'aspark') then
          begin
            move(FValue[0], adropspark, sizeof(adropspark));
            case i of
              0:
                begin
                  valuename := 'Type';
                  value := adropspark._type;
                  valuetype := otByte;
                  if (lowercase(FTypeName) = 'adrop') then
                    valuedescription := enumDropType[adropspark._type]
                  else
                    valuedescription := enumSparkType[adropspark._type];
                end;
              1:
                begin
                  valuename := 'Heat';
                  value := adropspark.heat;
                  valuetype := otByte;
                end;
              2:
                begin
                  valuename := 'X';
                  value := adropspark.x;
                  valuetype := otByte;
                end;
              3:
                begin
                  valuename := 'Y';
                  value := adropspark.y;
                  valuetype := otByte;
                end;
              4:
                begin
                  valuename := 'XSpeed';
                  value := adropspark.x_speed;
                  valuetype := otByte;
                end;
              5:
                begin
                  valuename := 'YSpeed';
                  value := adropspark.y_speed;
                  valuetype := otByte;
                end;
              6:
                begin
                  valuename := 'Age';
                  value := adropspark.age;
                  valuetype := otByte;
                end;
              7:
                begin
                  valuename := 'ExpTime';
                  value := adropspark.exptime;
                  valuetype := otByte;
                end;
            end;
          end
        else
          begin
            valuename := 'unknown';
            value := 0;
            valuetype := otByte;
          end;
      end;
  else
    value := FValue;
  end;
end;

function TUTProperty.GetValueCount: integer;
begin
  if not FIsInitialized then
    begin
      result := 0;
      exit;
    end;
  case PropertyType of
    otNone: result := 0;
    otStruct:
      begin
        if lowercase(FTypeName) = 'color' then
          result := 4
        else if lowercase(FTypeName) = 'vector' then
          result := 3
        else if lowercase(FTypeName) = 'pointregion' then
          result := 3
        else if lowercase(FTypeName) = 'rotator' then
          result := 3
        else if lowercase(FTypeName) = 'scale' then
          result := 5
        else if lowercase(FTypeName) = 'plane' then
          result := 4
        else if lowercase(FTypeName) = 'sphere' then
          result := 4
        else if lowercase(FTypeName) = 'adrop' then
          result := 8
        else if lowercase(FTypeName) = 'aspark' then
          result := 8
        else
          result := 1;
      end
  else
    result := 1;
  end;
end;

procedure TUTProperty.SetProperty(Owner: TUTPackage; n: string; i: integer;
  t: TUTPropertyType; var value; valuesize: integer; typename: string);
begin
  FOwner := owner;
  FName := n;
  FArrayIndex := i;
  FPropertyType := t;
  setlength(FValue, valuesize);
  if valuesize > 0 then move(value, FValue[0], valuesize);
  FTypeName := typename;
  FIsInitialized := true;
end;

{ TUTPropertyList }

constructor TUTPropertyList.Create;
begin
  FProperties := tlist.create;
end;

destructor TUTPropertyList.Destroy;
begin
  FProperties.free;
end;

procedure TUTPropertyList.Clear;
var
  a: integer;
begin
  for a := 0 to FProperties.count - 1 do
    TUTProperty(FProperties[a]).free;
  FProperties.clear;
end;

function TUTPropertyList.GetProperty(i: integer): TUTProperty;
begin
  try
    result := FProperties[i];
  except
    result := nil;
  end;
end;

function TUTPropertyList.GetPropertyCount: integer;
begin
  result := FProperties.count;
end;

function TUTPropertyList.NewProperty: TUTProperty;
begin
  FProperties.add(TUTProperty.create);
  result := FProperties[FProperties.count - 1];
end;

function TUTPropertyList.GetPropertyByName(name: string): TUTProperty;
var
  a: integer;
begin
  result := nil;
  name := lowercase(name);
  a := 0;
  while a < FProperties.count do
    if lowercase(TUTProperty(FProperties[a]).name) = name then
      begin
        result := FProperties[a];
        break;
      end
    else
      inc(a);
end;

function TUTPropertyList.GetPropertyListDescriptions: string;
var
  a: integer;
begin
  result := '';
  for a := 0 to FProperties.count - 1 do
    result := result + #1 + TUTProperty(FProperties[a]).Description;
  delete(result, 1, 1);
end;

procedure TUTPropertyList.FixArrayIndices;
var
  a, b: integer;
  more: boolean;
  function PropertyIsArray(const name: string): boolean;
  var
    a: integer;
  const
    // TODO : make array properties list global and complete them
    arrayprops: array[0..22] of string =
    ('paths', 'visnoreachpaths', 'upstreampaths', 'prunedpaths', 'aiprofile',
      'multiskins', 'lensaflare', 'lensflareoffset', 'lensflarescale', 'delay',
      'gain', 'outevents', 'outdelays', 'objshots', 'objdesc', 'damageevent',
      'damageeventthreshold', 'tags', 'events', 'opentimes', 'closetimes',
      'splats', 'internaltime');
  begin
    result := false;
    for a := 0 to high(arrayprops) do
      if arrayprops[a] = name then
        begin
          result := true;
          break;
        end;
  end;
begin
  // fix array index of first elements
  for a := 0 to FProperties.Count - 1 do
    if (PropertyByPosition[a].ArrayIndex = -1) then
      begin
        more := PropertyIsArray(PropertyByPosition[a].Name);
        if not more then
          for b := 0 to FProperties.Count - 1 do
            begin
              if (a <> b) and (PropertyByPosition[b].Name = PropertyByPosition[a].Name) then
                begin
                  more := true;
                  break;
                end;
            end;
        if more then PropertyByPosition[a].FArrayIndex := 0;
      end;
end;

function TUTPropertyList.GetPropertyByNameValue(name: string): variant;
var
  p: TUTProperty;
begin
  p := GetPropertyByName(name);
  if p = nil then
    result := 0
  else
    result := p.value;
end;

function TUTPropertyList.GetPropertyValue(i: integer): variant;
var
  p: TUTProperty;
begin
  p := GetProperty(i);
  if p = nil then
    result := 0
  else
    result := p.value;
end;

{ TUTPackage }

procedure TUTPackage.read_buffer(var buffer; const size: integer; stream: tstream);
begin
  if stream = nil then stream := Fstr;
  fillchar(buffer, size, 0);
  stream.ReadBuffer(buffer, size);
end;

function TUTPackage.read_asciiz(stream: tstream): string;
var
  b: byte;
begin
  result := '';
  repeat
    b := read_byte(stream);
    if b <> 0 then result := result + chr(b);
  until b = 0;
end;

function TUTPackage.read_sizedascii(stream: tstream): string;
var
  namesize: byte;
begin
  namesize := read_idx(stream);
  setlength(result, namesize);
  read_buffer(result[1], namesize, stream);
end;

function TUTPackage.read_sizedasciiz(stream: tstream): string;
begin
  result := read_sizedascii(stream);
  setlength(result, length(result) - 1);
end;

function TUTPackage.read_doublesizedasciiz(stream: tstream): string;
begin
  read_idx(stream);                     // tamao
  result := read_sizedasciiz(stream);
end;

function TUTPackage.read_int(stream: tstream): integer;
begin
  if stream = nil then stream := Fstr;
  stream.ReadBuffer(result, 4);
end;

function TUTPackage.read_word(stream: tstream): word;
begin
  if stream = nil then stream := Fstr;
  stream.ReadBuffer(result, 2);
end;

function TUTPackage.read_byte(stream: tstream): byte;
begin
  if stream = nil then stream := Fstr;
  stream.ReadBuffer(result, 1);
end;

function TUTPackage.read_bool(stream: tstream): boolean;
var
  src: word;
begin
  if stream = nil then stream := Fstr;
  stream.ReadBuffer(src, 2);            // first byte=D3 (?), second byte=0 for True?
  result := ((src shr 8) = 0);
end;

function TUTPackage.read_float(stream: tstream): single;
begin
  if stream = nil then stream := Fstr;
  stream.ReadBuffer(result, 4);
end;

function TUTPackage.read_idx(stream: tstream): integer;
var
  b0, b1, b2, b3, b4: byte;
begin
  result := 0;
  b0 := read_byte(stream);
  if (b0 and $40) <> 0 then
    begin
      b1 := read_byte(stream);
      if (b1 and $80) <> 0 then
        begin
          b2 := read_byte(stream);
          if (b2 and $80) <> 0 then
            begin
              b3 := read_byte(stream);
              if (b3 and $80) <> 0 then
                begin
                  b4 := read_byte(stream);
                  result := b4;
                end;
              result := (result shl 7) or (b3 and $7F);
            end;
          result := (result shl 7) or (b2 and $7F);
        end;
      result := (result shl 7) or (b1 and $7F);
    end;
  result := (result shl 6) or (b0 and $3F);
  if (b0 and $80) <> 0 then result := -result;
end;

function TUTPackage.Read_Name(stream: tstream): string;
begin
  if FVersion >= 64 then
    begin
      result := Read_SizedASCII(stream);
      setlength(result, length(result) - 1); // remove ending #0
    end
  else
    begin
      result := Read_ASCIIZ(stream);
    end;
end;

function TUTPackage.GetObjectFlagsText(const e: cardinal): string;
begin
  result := '';
  // TODO : add support for unknown flag bits
  if (e and RF_LoadForClient) <> 0 then result := result + ', RF_LoadForClient';
  if (e and RF_NotForClient) <> 0 then result := result + ', RF_NotForClient';
  if (e and RF_LoadForServer) <> 0 then result := result + ', RF_LoadForServer';
  if (e and RF_NotForServer) <> 0 then result := result + ', RF_NotForServer';
  if (e and RF_LoadForEdit) <> 0 then result := result + ', RF_LoadForEdit';
  if (e and RF_NotForEdit) <> 0 then result := result + ', RF_NotForEdit';
  if (e and RF_Public) <> 0 then result := result + ', RF_Public';
  if (e and RF_Standalone) <> 0 then result := result + ', RF_Standalone';
  if (e and RF_Native) <> 0 then result := result + ', RF_Native';
  if (e and RF_SourceModified) <> 0 then result := result + ', RF_SourceModified';
  if (e and RF_Transactional) <> 0 then result := result + ', RF_Transactional';
  if (e and RF_HasStack) <> 0 then result := result + ', RF_HasStack';
  if (e and RF_Transient) <> 0 then result := result + ', RF_Transient';
  if (e and RF_Unreachable) <> 0 then result := result + ', RF_Unreachable';
  if (e and RF_TagImp) <> 0 then result := result + ', RF_TagImp';
  if (e and RF_TagExp) <> 0 then result := result + ', RF_TagExp';
  if (e and RF_TagGarbage) <> 0 then result := result + ', RF_TagGarbage';
  if (e and RF_NeedLoad) <> 0 then result := result + ', RF_NeedLoad';
  if (e and RF_HighlightedName) <> 0 then result := result + ', RF_HighlightedName';
  if (e and RF_InSingularFunc) <> 0 then result := result + ', RF_InSingularFunc';
  if (e and RF_Suppress) <> 0 then result := result + ', RF_Suppress';
  if (e and RF_InEndState) <> 0 then result := result + ', RF_InEndState';
  if (e and RF_PreLoading) <> 0 then result := result + ', RF_PreLoading';
  if (e and RF_Destroyed) <> 0 then result := result + ', RF_Destroyed';
  if (e and RF_NeedPostLoad) <> 0 then result := result + ', RF_NeedPostLoad';
  if (e and RF_Marked) <> 0 then result := result + ', RF_Marked';
  if (e and RF_ErrorShutdown) <> 0 then result := result + ', RF_ErrorShutdown';
  if (e and RF_DebugPostLoad) <> 0 then result := result + ', RF_DebugPostLoad';
  if (e and RF_DebugSerialize) <> 0 then result := result + ', RF_DebugSerialize';
  if (e and RF_DebugDestroy) <> 0 then result := result + ', RF_DebugDestroy';

  if (e and RF_Unk_00000080) <> 0 then result := result + ', RF_Unk_00000080';
  if (e and RF_Unk_00000100) <> 0 then result := result + ', RF_Unk_00000100';

  if result <> '' then delete(result, 1, 1);
end;

function TUTPackage.GetObjectPath(limit, index: integer): string;
var
  i: TUTImportTableObjectData;
  e: TUTExportTableObjectData;
  s: string;
begin
  result := '';
  while limit <> 0 do
    begin
      if index = 0 then
        break
      else if index < 0 then
        begin
          if - index - 1 < Fimporttablelist.count then
            begin
              i := TUTImportTableObjectData(Fimporttablelist[-index - 1]);
              s := Fnametablelist[i.UTObjectIndex] + '.';
              index := i.UTpackageindex;
            end
          else
            begin
              s := '';
              result := '';
              limit := 1;
            end;
        end
      else if index > 0 then
        begin
          if index - 1 < Fexporttablelist.count then
            begin
              e := TUTExportTableObjectData(Fexporttablelist[index - 1]);
              s := Fnametablelist[e.UTObjectIndex] + '.';
              index := e.UTpackageindex;
            end
          else
            begin
              s := '';
              result := '';
              limit := 1;
            end;
        end;
      result := s + result;
      dec(limit);
    end;
  if copy(result, length(result), 1) = '.' then setlength(result, length(result) - 1);
end;

procedure TUTPackage.Process;
var
  name: string;
  int, i, a: integer;
  needsfix: boolean;
  lw, namecount, nameoffset, exportcount, exportoffset, importcount, importoffset,
    heritagecount, heritageoffset: cardinal;
  guid: TGUID;
  id: TUTImportTableObjectData;
  ed: TUTExportTableObjectData;
  maxposition, position: integer;
begin
  try
    maxposition := 100;
    try
      DoOnProgress(0, 100);
      StartReadingPackage;
      lw := cardinal(read_int(Fstr));
      if lw <> $9E2A83C1 then raise exception.create(rsErrorNoUTPackage);
      FVersion := read_word(Fstr);
      FLicenseeMode := read_word(FStr);
      FFlags := read_int(Fstr);
      namecount := read_int(Fstr);
      nameoffset := read_int(Fstr);
      exportcount := read_int(Fstr);
      exportoffset := read_int(Fstr);
      importcount := read_int(Fstr);
      importoffset := read_int(Fstr);
      setlength(Fheritagetablelist, 0);
      if FVersion < 68 then
        begin
          heritagecount := read_int(Fstr);
          heritageoffset := read_int(Fstr);
        end
      else
        begin
          heritagecount := 1;
          heritageoffset := Fstr.Position;
          guid := read_guid(Fstr);      // will be read again below
          int := read_int(Fstr);
          setlength(FGenerationInfo, int);
          for i := 0 to int - 1 do
            begin
              FGenerationInfo[i].exportcount := read_int(Fstr); // number of exports
              FGenerationInfo[i].namecount := read_int(Fstr); // number of names
            end;
        end;
      maxposition := namecount + exportcount * 2 + importcount * 2 + heritagecount;
      position := 0;
      // name table
      Seek(nameoffset);
      Fnametablelist.clear;
      for i := 1 to namecount do
        begin
          name := Read_Name(Fstr);
          int := read_int(Fstr);
          needsfix := false;
          a := 1;
          while a <= length(name) do
            begin
              if not (name[a] in ['a'..'z', 'A'..'Z', '0'..'9']) then
                begin
                  needsfix := true;
                  name := copy(name, 1, a - 1) + format('%-2.2x', [ord(name[a])]) + copy(name, a + 1, length(name));
                  inc(a);
                end;
              inc(a);
            end;
          if needsfix then name := 'Obj_' + name;
          Fnametablelist.AddObject(name, pointer(int));
          inc(position);
          DoOnProgress(position, maxposition);
        end;
      // export table
      Seek(exportoffset);
      for i := 0 to Fexporttablelist.count - 1 do
        TUTExportTableObjectData(Fexporttablelist[i]).free;
      Fexporttablelist.clear;
      for i := 1 to exportcount do
        begin
          ed := TUTExportTableObjectData.create;
          ed.UTClassIndex := read_idx(Fstr); // class index
          ed.UTSuperIndex := read_idx(Fstr); // super index
          ed.UTPackageIndex := read_int(Fstr); // package index
          ed.UTObjectIndex := read_idx(Fstr); // object name
          ed.Flags := read_int(Fstr);   // object flags
          ed.SerialSize := read_idx(Fstr); // serial size
          if ed.SerialSize > 0 then
            ed.SerialOffset := read_idx(Fstr) // serial offset
          else
            ed.SerialOffset := 0;
          FExportTableList.add(ed);
          inc(position);
          DoOnProgress(position, maxposition);
        end;
      // import table
      Seek(importoffset);
      for i := 0 to Fimporttablelist.count - 1 do
        TUTImportTableObjectData(Fimporttablelist[i]).free;
      Fimporttablelist.clear;
      for i := 1 to importcount do
        begin
          id := TUTImportTableObjectData.create;
          id.UTClassPackageIndex := read_idx(Fstr); // ClassPackage
          id.UTClassIndex := read_idx(Fstr); // ClassName
          id.UTPackageIndex := read_int(Fstr); // PackageIndex
          id.UTObjectIndex := read_idx(Fstr); // ObjectName
          Fimporttablelist.add(id);
          inc(position);
          DoOnProgress(position, maxposition);
        end;
      {if FVersion < 68 then
        begin}
          // heritage table
      Seek(heritageoffset);
      setlength(Fheritagetablelist, heritagecount);
      for i := 0 to heritagecount - 1 do
        begin
          guid := read_guid(Fstr);
          Fheritagetablelist[i] := guid;
          inc(position);
          DoOnProgress(position, maxposition);
        end;
      {end;}

      for i := 0 to FImportTableList.count - 1 do
        begin
          id := TUTImportTableObjectData(FImportTableList[i]);
          id.UTpackagename := GetObjectPath(-1, id.UTPackageIndex);
          id.UTobjectname := Fnametablelist[id.UTObjectIndex];
          id.UTclasspackagename := Fnametablelist[id.UTClassPackageIndex];
          id.UTclassname := Fnametablelist[id.UTClassIndex];
          inc(position);
          DoOnProgress(position, maxposition);
        end;

      for i := 0 to Fexporttablelist.count - 1 do
        begin
          ed := TUTExportTableObjectData(Fexporttablelist[i]);
          ed.UTPackageName := GetObjectPath(-1, ed.UTPackageIndex);
          ed.UTobjectname := Fnametablelist[ed.UTObjectIndex];
          ed.UTclassname := GetObjectPath(1, ed.UTClassIndex);
          ed.UTsupername := GetObjectPath(-1, ed.UTSuperIndex);
          ed.UTObject := GeTUTObjectClass(ed.classname).create(self, i);
          inc(position);
          DoOnProgress(position, maxposition);
        end;

    finally
      EndReadingPackage;
      DoOnProgress(maxposition, maxposition);
    end;
  except
    on e: exception do raise exception.create(format(rsExceptionProcessingPackage, [FPackage]));
  end;
end;

procedure TUTPackage.StartReadingPackage;
begin
  if (FReadingPackageCount = 0) then
    Fstr := tfilestream.create(FPackage, fmOpenRead + fmShareDenyNone);
  inc(FReadingPackageCount);
end;

procedure TUTPackage.EndReadingPackage;
begin
  dec(FReadingPackageCount);
  if (FReadingPackageCount = 0) then
    begin
      Fstr.free;
      Fstr := nil;
    end;
end;

function TUTPackage.ReadProperty(prop: TUTProperty; stream: tstream): boolean;
var
  n, nlc, struct, nstruct: string;
  b: integer;
  infobyte, info_type, info_size, index_b, index_c, index_d, index_e: byte; index, v, size: integer;
  info_array: boolean;
  buffer: array of byte;
  f: single;
begin
  n := '<unknown>';
  try
    setlength(buffer, 256);
    b := read_idx(stream);
    n := inttostr(b);
    n := Fnametablelist[b];
    nlc := lowercase(n);
    index := -1;
    if nlc <> 'none' then               // do not localize properties names and types
      begin
        infobyte := read_byte(stream);
        info_type := infobyte and $0F;
        info_size := (infobyte shr 4) and 7;
        info_array := ((infobyte and $80) <> 0);

        if info_type = otStruct then
          begin
            struct := Fnametablelist[read_idx(stream)];
            nstruct := lowercase(struct);
          end;

        case info_size of
          0: size := 1;
          1: size := 2;
          2: size := 4;
          3: size := 12;
          4: size := 16;
          5: size := read_byte(stream);
          6: size := read_word(stream);
          7:
            size := read_int(stream)
        else
          size := 0;
        end;

        if info_array and (info_type <> otBool) then
          begin
            index_b := read_byte(stream);
            if (index_b and $80) = 0 then
              index := index_b
            else if (index_b and $C0) = $80 then
              begin
                index_c := read_byte(stream);
                index := ((index_b and $7F) shl 8) or index_c;
              end
            else
              begin
                index_c := read_byte(stream);
                index_d := read_byte(stream);
                index_e := read_byte(stream);
                index := ((index_b and $3F) shl 24) or (index_c shl 16) or (index_d shl 8) or index_e;
              end;
          end;

        case info_type of
          otByte:                       // Byte
            begin
              v := read_byte(stream);
              prop.SetProperty(self, n, index, otByte, v, 1);
            end;
          otInt:                        // Integer
            begin
              v := read_int(stream);
              prop.SetProperty(self, n, index, otInt, v, 4);
            end;
          otBool:                       // Boolean
            begin
              if info_array then
                v := 1
              else
                v := 0;
              prop.SetProperty(self, n, index, otBool, v, 4);
            end;
          otFloat:                      // Float
            begin
              f := read_float(stream);
              prop.SetProperty(self, n, index, otFloat, f, 4);
            end;
          otObject:                     // Object Reference
            begin
              v := read_idx(stream);
              prop.SetProperty(self, n, index, otObject, v, 4);
            end;
          otName:                       // Name Reference (Tag)
            begin
              v := read_idx(stream);
              prop.SetProperty(self, n, index, otName, v, 4);
            end;
          otString:
            begin
              setlength(buffer, size);
              read_buffer(buffer[0], size, stream);
              prop.SetProperty(self, n, index, otString, buffer[0], size);
            end;
          {
          otClass:
          otArray:
          }
          otStruct:                     // Struct
            begin
              if nstruct = 'pointregion' then
                begin
                  // use a buffer because it contains indices and so the size is not fixed
                  b := read_idx(stream);
                  move(b, buffer[0], 4);
                  b := read_int(stream);
                  move(b, buffer[4], 4);
                  b := read_idx(stream);
                  move(b, buffer[8], 4);
                  size := 12;           // override size
                end
              else
                read_buffer(buffer[0], size, stream);
              prop.SetProperty(self, n, index, otStruct, buffer[0], size, struct);
            end;
          {
          otVector:
          otRotator:
          }
          otStr:                        // Str
            begin
              v := read_idx(stream);    // length including #0
              setlength(buffer, v);
              read_buffer(buffer[0], v, stream);
              prop.SetProperty(self, n, index, otStr, buffer[0], v);
            end
            {
            otMap:
            otFixedArray:
            }
        else
          begin                         // Unknown type
            if size > sizeof(buffer) then
              read_buffer(buffer[0], sizeof(buffer), stream)
            else
              read_buffer(buffer[0], size, stream);
            prop.SetProperty(self, n, index, otBuffer, v, 0, format(rsUnknownPropertyType, [info_type]));
            result := false;
            exit;
          end;
        end;
      end
    else
      prop.SetProperty(self, n, index, otNone, v, 0);
    result := (nlc <> 'none');
  except
    on e: exception do raise exception.create(format(rsExceptionReadingProperty, [n]));
  end;
end;

constructor TUTPackage.Create(package: string = '');
begin
  FNameTablelist := tstringlist.create;
  FImportTablelist := tlist.create;
  FExportTablelist := tlist.create;
  setlength(FHeritageTableList, 0);
  FReadingPackageCount := 0;
  Fstr := nil;
  if package <> '' then
    begin
      FPackage := package;
      Process;
    end;
end;

destructor TUTPackage.Destroy;
var
  a: integer;
begin
  EndReadingPackage;
  Fnametablelist.free;
  for a := 0 to Fexporttablelist.count - 1 do
    TUTExportTableObjectData(Fexporttablelist[a]).free;
  Fexporttablelist.free;
  for a := 0 to Fimporttablelist.count - 1 do
    TUTImportTableObjectData(Fimporttablelist[a]).free;
  Fimporttablelist.free;
  inherited;
end;

function TUTPackage.GetName(i: integer): string;
begin
  if (i >= 0) and (i < FNameTableList.count) then
    result := FNameTableList[i]
  else
    result := '';
end;

function TUTPackage.GetExport(i: integer): TUTExportTableObjectData;
begin
  if (i >= 0) and (i < FExportTableList.count) then
    result := TUTExportTableObjectData(FExportTableList[i])
  else
    result := nil;
end;

function TUTPackage.GetImport(i: integer): TUTImportTableObjectData;
begin
  if (i >= 0) and (i < FImportTableList.count) then
    result := TUTImportTableObjectData(FImportTableList[i])
  else
    result := nil;
end;

function TUTPackage.GetHeritage(i: integer): TGUID;
begin
  if (i >= 0) and (i < length(FHeritageTableList)) then
    result := FHeritageTableList[i]
  else
    fillchar(result, sizeof(TGUID), 0);
end;

procedure TUTPackage.Seek(p: integer);
begin
  Fstr.seek(p, soFromBeginning);
end;

procedure TUTPackage.Initialize(package: string);
begin
  FPackage := package;
  Process;
end;

function TUTPackage.GetNameFlags(i: integer): integer;
begin
  if (i >= 0) and (i < FNameTableList.count) then
    result := integer(Fnametablelist.objects[i])
  else
    result := 0;
end;

function TUTPackage.GetExportCount: integer;
begin
  result := FExportTableList.count;
end;

function TUTPackage.GetHeritageCount: integer;
begin
  result := length(FHeritageTableList);
end;

function TUTPackage.GetImportCount: integer;
begin
  result := FImportTableList.count;
end;

function TUTPackage.GetNameCount: integer;
begin
  result := FNameTableList.count;
end;

procedure TUTPackage.SaveDataBlock(filename: string; position, size: integer);
var
  str2: tfilestream;
begin
  Seek(position);
  str2 := tfilestream.create(filename, fmCreate);
  try
    str2.CopyFrom(Fstr, size);
  finally
    str2.free;
  end;
end;

function TUTPackage.GetPackagePosition: integer;
begin
  result := Fstr.position;
end;

function TUTPackage.GetStream: TStream;
begin
  result := Fstr;
end;

function TUTPackage.GetExportIndex(objectname, classname: string): integer;
var
  a: integer;
  found: boolean;
  ed: TUTExportTableObjectData;
begin
  a := 0;
  found := false;
  objectname := lowercase(objectname);
  classname := lowercase(classname);
  while not found and (a < Fexporttablelist.count) do
    begin
      ed := TUTExportTableObjectData(Fexporttablelist[a]);
      if (lowercase(ed.UTobjectname) = objectname) and (lowercase(ed.UTclassname) = classname) then
        found := true
      else
        inc(a);
    end;
  if found then
    result := a
  else
    result := -1;
end;

function TUTPackage.GetNameIndex(objectname: string): integer;
begin
  result := Fnametablelist.indexof(objectname);
end;

function TUTPackage.EncodeIndex(i: integer): string;
var
  v: integer;
  b0, b1, b2, b3, b4: byte;
begin
  result := '';
  v := abs(i);
  if i >= 0 then
    b0 := 0
  else
    b0 := $80;
  if v < $40 then
    b0 := b0 + v
  else
    b0 := b0 + $40 + (v and $3F);
  result := result + chr(b0);
  if (b0 and $40) <> 0 then
    begin
      v := v shr 6;
      if v < $80 then
        b1 := v
      else
        b1 := (v and $7F) + $80;
      result := result + chr(b1);
      if (b1 and $80) <> 0 then
        begin
          v := v shr 7;
          if v < $80 then
            b2 := v
          else
            b2 := (v and $7F) + $80;
          result := result + chr(b2);
          if (b2 and $80) <> 0 then
            begin
              v := v shr 7;
              if v < $80 then
                b3 := v
              else
                b3 := (v and $7F) + $80;
              result := result + chr(b3);
              if (b3 and $80) <> 0 then
                begin
                  v := v shr 7;
                  b4 := v;
                  result := result + chr(b4);
                end;
            end;
        end;
    end;
end;

function TUTPackage.FindObject(where: TUTPackageObjectLocations; what: TUTPackageFindWhatSet;
  packagename, objectname, classname: string; start: integer): integer;
var
  a: integer;
  ok: boolean;
begin
  objectname := lowercase(objectname);
  packagename := lowercase(packagename);
  classname := lowercase(classname);
  case where of
    utolNames:
      begin
        a := start;
        result := -1;
        while (a < FNameTableList.count) and (result = -1) do
          if (lowercase(FNameTableList[a]) = objectname) then
            result := a
          else
            inc(a);
      end;
    utolExports:
      begin
        a := start;
        result := -1;
        while (a < ExportedCount) and (result = -1) do
          begin
            ok := true;
            if (utfwName in what) then
              ok := ok and (lowercase(Exported[a].UTObjectName) = objectname);
            if (utfwClass in what) then
              ok := ok and (lowercase(Exported[a].UTClassName) = classname);
            if (utfwPackage in what) then
              ok := ok and (lowercase(Exported[a].UTPackageName) = packagename);
            if ok then
              result := a
            else
              inc(a);
          end;
      end;
    utolImports:
      begin
        a := start;
        result := -1;
        while (a < ImportedCount) and (result = -1) do
          begin
            ok := true;
            if (utfwName in what) then
              ok := ok and (lowercase(Imported[a].UTObjectName) = objectname);
            if (utfwClass in what) then
              ok := ok and (lowercase(Imported[a].UTClassName) = classname);
            if (utfwPackage in what) then
              ok := ok and (lowercase(Imported[a].UTPackageName) = packagename);
            if ok then
              result := a
            else
              inc(a);
          end;
      end
  else
    result := -1;
  end;
end;

procedure TUTPackage.ReadAllObjects;
var
  a, b: integer;
begin
  b := ExportedCount - 1;
  DoOnProgress(0, b);
  for a := 0 to b do
    begin
      Exported[a].UTObject.ReadObject;
      DoOnProgress(a, b);
    end;
end;

procedure TUTPackage.DoOnProgress(position, maxposition: integer);
begin
  if assigned(FOnProgress) then
    if maxposition = 0 then
      FOnProgress(self, 0)
    else
      FOnProgress(self, position * 100 div maxposition);
end;

function TUTPackage.GetGeneration(i: integer): TUTPackage_GenerationInfo;
begin
  result := FGenerationInfo[i];
end;

function TUTPackage.GetGenerationCount: integer;
begin
  result := length(FGenerationInfo);
end;

function TUTPackage.IndentText(indent, txt: string): string;
var
  str: tstringlist;
  a: integer;
  ending_crlf: boolean;
begin
  ending_crlf := (copy(txt, length(txt) - 1, 2) = #13#10);
  str := tstringlist.create;
  str.text := txt;
  for a := 0 to str.count - 1 do
    str[a] := indent + str[a];
  result := str.text;
  str.free;
  if not ending_crlf then delete(result, length(result) - 1, 2);
end;

function TUTPackage.read_qword(stream: tstream): int64;
begin
  result := read_int(stream);
  result := result shl 32;
  result := result or read_int(stream);
end;

function TUTPackage.read_guid(stream: tstream): TGuid;
begin
  read_buffer(result, sizeof(result), stream);
end;

{ TUTObject }

constructor TUTObject.create(owner: TUTPackage; exportedindex: integer);
begin
  FOwner := owner;
  FExportedIndex := exportedindex;
  Buffer := nil;
  FProperties := TUTPropertyList.create;
  InitializeObject;
end;

destructor TUTObject.destroy;
begin
  ReleaseObject;
  FProperties.free;
  inherited;
end;

procedure TUTObject.ReadObject(interpret: boolean);
var
  p: integer;
begin
  if (buffer = nil) then
    begin
      FOwner.StartReadingPackage;
      try
        buffer := TMemoryStream.create;
        p := FOwner.Position;
        FStartInPackage := UTSerialOffset;
        FOwner.Seek(FStartInPackage);
        if UTSerialSize > 0 then
          buffer.CopyFrom(FOwner.GetStream, UTSerialSize);
        FOwner.Seek(p);
        buffer.Seek(0, soFromBeginning);
        if interpret and (buffer.size > 0) then
          begin
            InitializeObject;
            InterpretObject;
            buffer.Seek(0, soFromBeginning);
          end;
      finally
        FOwner.EndReadingPackage;
      end;
    end;
end;

procedure TUTObject.ReleaseObject;
begin
  if buffer <> nil then
    begin
      buffer.free;
      buffer := nil;
    end;
  FProperties.clear;
end;

procedure TUTObject.InitializeObject;
begin
  FProperties.clear;
end;

procedure TUTObject.InterpretObject;
var
  node: integer;
begin
  if (UTFlags and RF_HasStack) <> 0 then
    begin
      node := FOwner.read_idx(buffer);  // StateFrame.Node
      FOwner.read_idx(buffer);          // StateFrame.StateNode
      FOwner.read_qword(buffer);        // StateFrame.ProbeMask
      FOwner.read_int(buffer);          // StateFrame.LatentAction
      if node <> 0 then FOwner.read_idx(buffer); // Offset
    end;
  if UTClassIndex <> 0 then ReadProperties;
end;

procedure TUTObject.ReadProperties;
var
  more: boolean;
begin
  repeat
    more := FOwner.ReadProperty(FProperties.New, buffer);
  until not more;
  FProperties.FixArrayIndices;
end;

function TUTObject.GetClassName: string;
begin
  result := FOwner.Exported[FExportedIndex].ClassName;
end;

function TUTObject.GetObjectname: string;
begin
  result := FOwner.Exported[FExportedIndex].UTObjectName;
end;

function TUTObject.GetClassIndex: integer;
begin
  result := FOwner.Exported[FExportedIndex].UTClassIndex;
end;

function TUTObject.GetObjectIndex: integer;
begin
  result := FOwner.Exported[FExportedIndex].UTObjectIndex;
end;

function TUTObject.GetPackageIndex: integer;
begin
  result := FOwner.Exported[FExportedIndex].UTPackageIndex;
end;

function TUTObject.GetPackageName: string;
begin
  result := FOwner.Exported[FExportedIndex].UTPackageName;
end;

function TUTObject.GetSuperIndex: integer;
begin
  result := FOwner.Exported[FExportedIndex].UTSuperIndex;
end;

function TUTObject.GetSuperName: string;
begin
  result := FOwner.Exported[FExportedIndex].UTSuperName;
end;

function TUTObject.GetSerialOffset: integer;
begin
  result := FOwner.Exported[FExportedIndex].SerialOffset;
end;

function TUTObject.GetSerialSize: integer;
begin
  result := FOwner.Exported[FExportedIndex].SerialSize;
end;

function TUTObject.GetFlags: longword;
begin
  result := FOwner.Exported[FExportedIndex].Flags;
end;

procedure TUTObject.check_initialized;
begin
  assert(buffer <> nil, UTObjectName + ': ReadObject was not called for this object!');
  {if buffer = nil then
    raise exception.create(UTObjectName + ': ReadObject was not called for this object!');}
  // could also make the ReadObject call, but the programmer would not release the object
end;

function TUTObject.GetProperties: TUTPropertyList;
begin
  check_initialized;
  result := FProperties;
end;

function TUTObject.GetFullName: string;
begin
  result := FOwner.GetObjectPath(-1, FExportedIndex + 1);
end;

{ TUTObject_NoHeader }

procedure TUTObject_NoHeader.InterpretObject;
begin
  // We dont call "inherited" because these objects do not have a header before the properties.
  ReadProperties;
end;

{ TUTObjectClassPalette }

function TUTObjectClassPalette.GetColor(n: integer): TColor;
begin
  check_initialized;
  result := TColor(FColors[n]);
end;

function TUTObjectClassPalette.GetColorCount: integer;
begin
  check_initialized;
  result := FColorCount;
end;

function TUTObjectClassPalette.GetNewPalette: HPalette;
var
  palette_struct: PLogPalette;
  buffer: array[0..4 + 256 * 4 - 1] of byte;
  y: integer;
begin
  check_initialized;
  palette_struct := @buffer;
  palette_struct^.palVersion := $300;
  palette_struct^.palNumEntries := 256;
  {$IFOPT R+}
  {$DEFINE RangeCheck}
  {$ENDIF}
  {$R-}
  for y := 0 to 255 do
    begin
      move(FColors[y], palette_struct^.palPalEntry[y], 4);
      palette_struct^.palPalEntry[y].peFlags := 0;
    end;
  {$IFDEF RangeCheck}
  {$R+}
  {$ENDIF}
  result := CreatePalette(palette_struct^);
end;

procedure TUTObjectClassPalette.InitializeObject;
begin
  inherited;
  FColorCount := 0;
  setlength(FColors, 0);
end;

procedure TUTObjectClassPalette.InterpretObject;
var
  a: integer;
begin
  inherited;
  FColorCount := FOwner.read_idx(buffer);
  setlength(FColors, FColorCount);
  for a := 0 to FColorCount - 1 do
    FOwner.read_buffer(FColors[a], 4, buffer);
end;

procedure TUTObjectClassPalette.ReleaseObject;
begin
  FColorCount := 0;
  setlength(FColors, 0);
  inherited;
end;

{ TUTObjectClassSound }

function TUTObjectClassSound.GetData: string;
begin
  check_initialized;
  result := FData;
end;

function TUTObjectClassSound.GetFormat: string;
begin
  check_initialized;
  result := FFormat;
end;

procedure TUTObjectClassSound.InitializeObject;
begin
  inherited;
  FFormat := '';
  FData := '';
end;

procedure TUTObjectClassSound.InterpretObject;
begin
  inherited;
  FFormat := FOwner.Names[FOwner.read_idx(buffer)];
  setlength(FData, FOwner.read_idx(buffer));
  FOwner.read_buffer(FData[1], length(FData), buffer);
end;

procedure TUTObjectClassSound.ReleaseObject;
begin
  FFormat := '';
  FData := '';
  inherited;
end;

procedure TUTObjectClassSound.SaveToFile(filename: string);
var
  f: file;
begin
  check_initialized;
  assignfile(f, filename);
  rewrite(f, 1);
  try
    blockwrite(f, FData[1], length(FData));
  finally
    closefile(f);
  end;
end;

{ TUTObjectClassTextBuffer }

function TUTObjectClassTextBuffer.GetData: string;
begin
  check_initialized;
  result := FData;
end;

procedure TUTObjectClassTextBuffer.InitializeObject;
begin
  inherited;
  FData := '';
end;

procedure TUTObjectClassTextBuffer.InterpretObject;
begin
  inherited;
  FOwner.read_int(buffer);              // pos
  FOwner.read_int(buffer);              // top
  setlength(FData, FOwner.read_idx(buffer) - 1);
  FOwner.read_buffer(FData[1], length(FData), buffer);
  FOwner.read_byte(buffer);             // null
end;

procedure TUTObjectClassTextBuffer.ReleaseObject;
begin
  FData := '';
  inherited;
end;

procedure TUTObjectClassTextBuffer.SaveToFile(filename: string);
var
  f: file;
begin
  check_initialized;
  assignfile(f, filename);
  rewrite(f, 1);
  try
    blockwrite(f, FData[1], length(FData));
  finally
    closefile(f);
  end;
end;

{ TUTObjectClassMusic }

function TUTObjectClassMusic.GetData: string;
begin
  check_initialized;
  result := FData;
end;

function TUTObjectClassMusic.GetFormat: string;
begin
  check_initialized;
  result := FFormat;
end;

procedure TUTObjectClassMusic.InitializeObject;
begin
  inherited;
  FFormat := '';
  FData := '';
end;

procedure TUTObjectClassMusic.InterpretObject;
begin
  inherited;
  FFormat := FOwner.Names[0];
  FOwner.Read_word(buffer);             // numchunks?
  FOwner.read_int(buffer);              // always 1 ?
  setlength(FData, FOwner.read_idx(buffer));
  FOwner.read_buffer(FData[1], length(FData), buffer);
end;

procedure TUTObjectClassMusic.ReleaseObject;
begin
  FFormat := '';
  FData := '';
  inherited;
end;

procedure TUTObjectClassMusic.SaveToFile(filename: string);
var
  f: file;
begin
  check_initialized;
  assignfile(f, filename);
  rewrite(f, 1);
  try
    blockwrite(f, FData[1], length(FData));
  finally
    closefile(f);
  end;
end;

{ TUTObjectClassFont }

procedure TUTObjectClassFont.GetCharacterInfo(i: integer; var texture, x,
  y, w, h: integer);
begin
  check_initialized;
  if (i >= 0) and (i <= high(FCharacters)) then
    begin
      texture := FCharacters[i].texture;
      x := FCharacters[i].x;
      y := FCharacters[i].y;
      w := FCharacters[i].w;
      h := FCharacters[i].h;
    end
  else
    begin
      texture := -1;
      x := -1;
      y := -1;
      w := -1;
      h := -1;
    end;
end;

procedure TUTObjectClassFont.InitializeObject;
begin
  inherited;
  setlength(FCharacters, 0);
end;

procedure TUTObjectClassFont.InterpretObject;
var
  numtextures, charidx, a, b, numchars, idx: integer;
begin
  inherited;
  numtextures := FOwner.read_byte(buffer);
  setlength(FCharacters, 256);
  charidx := 0;
  for a := 1 to numtextures do
    begin
      idx := FOwner.read_idx(buffer);   // texture inedx
      numchars := FOwner.read_idx(buffer); // character count
      for b := 1 to numchars do
        begin
          FCharacters[charidx].texture := idx;
          FCharacters[charidx].x := FOwner.read_int(buffer);
          FCharacters[charidx].y := FOwner.read_int(buffer);
          FCharacters[charidx].w := FOwner.read_int(buffer);
          FCharacters[charidx].h := FOwner.read_int(buffer);
          inc(charidx);
        end;
    end;
  setlength(FCharacters, charidx);
end;

procedure TUTObjectClassFont.ReleaseObject;
begin
  setlength(FCharacters, 0);
  inherited;
end;

{ TUTObjectClassPolys }

function TUTObjectClassPolys.GetPolygonCount: integer;
begin
  check_initialized;
  result := high(FPolygons);
end;

function TUTObjectClassPolys.GetPolygon(n: integer): TUT_Struct_Polygon;
begin
  check_initialized;
  result := FPolygons[n];
end;

procedure TUTObjectClassPolys.InitializeObject;
begin
  inherited;
  setlength(FPolygons, 0);
end;

procedure TUTObjectClassPolys.InterpretObject;
var
  p, v: integer;
  function coord: single;
  var
    i: integer;
    sign: boolean;
  const
    sign_flag = $80000000;
  begin
    i := Fowner.read_int(buffer);
    sign := (i and sign_flag) = sign_flag;
    i := i and not sign_flag;
    move(i, result, 4);
    if sign then result := -result;
  end;
begin
  inherited;
  FOwner.read_int(buffer);              // num polys?
  setlength(FPolygons, Fowner.read_int(buffer)); // max polys?
  for p := 0 to high(FPolygons) do
    begin
      setlength(FPolygons[p].Vertex, FOwner.read_idx(buffer));
      with FPolygons[p] do
        begin
          Base.x := coord;
          Base.y := coord;
          Base.z := coord;
          Normal.x := coord;
          Normal.y := coord;
          Normal.z := coord;
          TextureU.x := coord;
          TextureU.y := coord;
          TextureU.z := coord;
          TextureV.x := coord;
          TextureV.y := coord;
          TextureV.z := coord;
          for v := 0 to high(Vertex) do
            begin
              Vertex[v].x := coord;
              Vertex[v].y := coord;
              Vertex[v].z := coord;
            end;
          PolyFlags := FOwner.read_int(buffer);
          Actor := FOwner.read_idx(buffer);
          Texture := FOwner.read_idx(buffer);
          ItemName := FOwner.read_idx(buffer);
          iLink := FOwner.read_idx(buffer);
          iBrushPoly := FOwner.read_idx(buffer);
          pan_u := FOwner.read_word(buffer);
          pan_v := FOwner.read_word(buffer);
          if pan_u > $8000 then pan_u := integer($FFFF0000 or cardinal(pan_u));
          if pan_v > $8000 then pan_v := integer($FFFF0000 or cardinal(pan_v));
        end;
    end;
end;

procedure TUTObjectClassPolys.ReleaseObject;
begin
  setlength(FPolygons, 0);
  inherited;
end;

{ TUTObjectClassTexture }

function TUTObjectClassTexture.GetCompMipMap(i: integer): TBitmap;
begin
  check_initialized;
  result := FCompMipMaps[i];
end;

function TUTObjectClassTexture.GetCompMipMapCount: integer;
begin
  check_initialized;
  result := length(FCompMipMaps);
end;

function TUTObjectClassTexture.GetMipMap(i: integer): TBitmap;
begin
  check_initialized;
  result := FMipMaps[i];
end;

function TUTObjectClassTexture.GetMipMapCount: integer;
begin
  check_initialized;
  result := length(FMipMaps);
end;

procedure TUTObjectClassTexture.InitializeObject;
begin
  inherited;
  setlength(FMipMaps, 0);
  setlength(FCompMipMaps, 0);
end;

procedure TUTObjectClassTexture.InterpretObject;
var
  a, data_pos, x, palette, block_size: integer;
  ed2: TUTExportTableObjectData;
  color: array[0..3] of word;
  yy: array[0..3] of byte;
  real_color: array[0..3] of tcolor;
  y, c, bx, by, yd4, xd4: integer;
  yv: byte;
  compformat: integer;
const
  TEXF_P8 = 0;
  TEXF_RGBA7 = 1;                       // TEXF_RGB32 ?
  TEXF_RGB16 = 2;                       // TEXF_RGB64 ?
  TEXF_DXT1 = 3;
  TEXF_RGB8 = 4;                        // TEXF_RGB24 ?
  TEXF_RGBA8 = 5;                       // none?
begin
  inherited;
  // Read MipMaps
  setlength(FMipMaps, FOwner.read_byte(buffer));
  for a := 0 to high(FMipMaps) do
    begin
      if FOwner.Version >= 63 then
        begin
          y := FOwner.read_int(buffer) - FStartInPackage; // position after block
          block_size := FOwner.read_idx(buffer); // block size
          data_pos := buffer.position;
        end
      else
        begin
          block_size := FOwner.read_idx(buffer); // block size
          data_pos := buffer.position;
          y := data_pos + block_size;
        end;
      buffer.seek(y, soFromBeginning);
      FMipMaps[a] := tbitmap.create;
      FMipMaps[a].pixelformat := pf8Bit;
      FMipMaps[a].width := FOwner.read_int(buffer);
      FMipMaps[a].height := FOwner.read_int(buffer);
      buffer.seek(data_pos, soFromBeginning);
      // Extract palette
      palette := Properties['palette']; // do not localize
      if palette > 0 then
        begin
          ed2 := FOwner.Exported[palette - 1];
          ed2.UTObject.ReadObject;
          FMipMaps[a].palette := TUTObjectClassPalette(ed2.UTObject).GetPalette;
          ed2.UTObject.ReleaseObject;
        end;
      //else : don't supports imported palettes
      // Extract bitmap
      if block_size <> 0 then           // block_size=0 in some Texture descendants
        for y := 0 to FMipMaps[a].height - 1 do
          FOwner.read_buffer(FMipMaps[a].scanline[y]^, FMipMaps[a].width, buffer);
      FOwner.read_int(buffer);          // Width
      FOwner.read_int(buffer);          // Height
      FOwner.read_byte(buffer);         // Width bits
      FOwner.read_byte(buffer);         // Height bits
    end;
  // Read Compressed MipMaps
  if Properties['bHasComp'] then
    begin
      compformat := Properties['CompFormat'];
      setlength(FCompMipMaps, FOwner.read_byte(buffer));
      for a := 0 to high(FCompMipMaps) do
        begin
          if FOwner.Version >= 63 then
            begin
              y := FOwner.read_int(buffer) - FStartInPackage; // position after block
              block_size := FOwner.read_idx(buffer); // block size
              data_pos := buffer.position;
            end
          else
            begin
              block_size := FOwner.read_idx(buffer); // block size
              data_pos := buffer.position;
              y := data_pos + block_size;
            end;
          buffer.seek(y, soFromBeginning);
          FCompMipMaps[a] := tbitmap.create;
          FCompMipMaps[a].width := FOwner.read_int(buffer);
          FCompMipMaps[a].height := FOwner.read_int(buffer);
          buffer.seek(data_pos, soFromBeginning);
          if compformat = TEXF_P8 then
            begin
              FCompMipMaps[a].pixelformat := pf8Bit;
              // Extract palette
              palette := Properties['palette']; // do not localize
              if palette > 0 then
                begin
                  ed2 := FOwner.Exported[palette - 1];
                  ed2.UTObject.ReadObject;
                  FCompMipMaps[a].palette := TUTObjectClassPalette(ed2.UTObject).GetPalette;
                  ed2.UTObject.ReleaseObject;
                end;
              //else : don't supports imported palettes
            end
          else
            FCompMipMaps[a].pixelformat := pf24Bit;
          // Extract bitmap
          if block_size <> 0 then       // block_size=0 in some Texture descendants
            begin
              case compformat of
                TEXF_P8:
                  begin                 // Palettized
                    for y := 0 to FCompMipMaps[a].height - 1 do
                      FOwner.read_buffer(FCompMipMaps[a].scanline[y]^, FCompMipMaps[a].width, buffer);
                  end;
                TEXF_DXT1:
                  begin                 // DirectX-1
                    // The texture needs to have a size divisible by 4
                    // This format allows for one transparent color, but we dont use that.
                    yd4 := (FCompMipMaps[a].height div 4);
                    xd4 := (FCompMipMaps[a].width div 4);
                    if yd4 = 0 then yd4 := 1;
                    if xd4 = 0 then xd4 := 1;
                    for y := 0 to yd4 - 1 do
                      for x := 0 to xd4 - 1 do
                        begin
                          // read 4x4 block
                          color[0] := FOwner.read_word(buffer);
                          color[1] := FOwner.read_word(buffer);
                          yy[0] := FOwner.read_byte(buffer);
                          yy[1] := FOwner.read_byte(buffer);
                          yy[2] := FOwner.read_byte(buffer);
                          yy[3] := FOwner.read_byte(buffer);
                          // get colors
                          for c := 0 to 1 do
                            real_color[c] := rgb(((color[c] and $001F) shl 3),
                              ((color[c] and $07E0) shr 3),
                              ((color[c] and $F800) shr 8)
                              );
                          if color[0] > color[1] then
                            begin
                              real_color[2] := rgb((2 * GetRValue(real_color[0]) + GetRValue(real_color[1])) div 3,
                                (2 * GetGValue(real_color[0]) + GetGValue(real_color[1])) div 3,
                                (2 * GetBValue(real_color[0]) + GetBValue(real_color[1])) div 3);
                              real_color[3] := rgb((GetRValue(real_color[0]) + 2 * GetRValue(real_color[1])) div 3,
                                (GetGValue(real_color[0]) + 2 * GetGValue(real_color[1])) div 3,
                                (GetBValue(real_color[0]) + 2 * GetBValue(real_color[1])) div 3);
                            end
                          else
                            begin
                              real_color[2] := rgb((GetRValue(real_color[0]) + GetRValue(real_color[1])) div 2,
                                (GetGValue(real_color[0]) + GetGValue(real_color[1])) div 2,
                                (GetBValue(real_color[0]) + GetBValue(real_color[1])) div 2);
                              real_color[3] := 0;
                            end;
                          // get pixels
                          for by := 0 to 3 do
                            begin
                              yv := yy[by];
                              for bx := 0 to 3 do
                                begin
                                  if (y * 4 + by < FCompMipMaps[a].height) and (x * 4 + bx < FCompMipMaps[a].width) then
                                    move(real_color[yv and 3], PByteArray(FCompMipMaps[a].scanline[y * 4 + by])^[3 * (x * 4 + bx)], 3);
                                  yv := yv shr 2;
                                end;
                            end;
                        end;
                  end
              else
                buffer.Seek(block_size, soFromCurrent);
              end;
            end;
          FOwner.read_int(buffer);      // Width
          FOwner.read_int(buffer);      // Height
          FOwner.read_byte(buffer);     // Width bits
          FOwner.read_byte(buffer);     // Height bits
        end;
    end;
end;

procedure TUTObjectClassTexture.ReleaseObject;
var
  a: integer;
begin
  for a := 0 to high(FMipMaps) do
    FMipMaps[a].free;
  setlength(FMipMaps, 0);
  for a := 0 to high(FCompMipMaps) do
    FCompMipMaps[a].free;
  setlength(FCompMipMaps, 0);
  inherited;
end;

procedure TUTObjectClassTexture.SaveCompMipMapToFile(mipmap: integer;
  filename: string);
begin
  check_initialized;
  FCompMipMaps[mipmap].savetofile(filename);
end;

procedure TUTObjectClassTexture.SaveMipMapToFile(mipmap: integer;
  filename: string);
begin
  check_initialized;
  FMipMaps[mipmap].savetofile(filename);
end;

{ TUTObjectClassPrimitive }

procedure TUTObjectClassPrimitive.InitializeObject;
begin
  inherited;
  fillchar(FPrimitiveBoundingBox, sizeof(FPrimitiveBoundingBox), 0);
  fillchar(FPrimitiveBoundingSphere, sizeof(FPrimitiveBoundingSphere), 0);
end;

procedure TUTObjectClassPrimitive.InterpretObject;
begin
  inherited;
  // UPrimitive.BoundingBox
  FPrimitiveBoundingBox.Min.x := FOwner.read_float(buffer);
  FPrimitiveBoundingBox.Min.y := FOwner.read_float(buffer);
  FPrimitiveBoundingBox.Min.z := FOwner.read_float(buffer);
  FPrimitiveBoundingBox.Max.x := FOwner.read_float(buffer);
  FPrimitiveBoundingBox.Max.y := FOwner.read_float(buffer);
  FPrimitiveBoundingBox.Max.z := FOwner.read_float(buffer);
  FPrimitiveBoundingBox.Valid := FOwner.read_byte(buffer);
  // UPrimitive.BoundingSphere
  FPrimitiveBoundingSphere.Center.x := FOwner.read_float(buffer);
  FPrimitiveBoundingSphere.Center.y := FOwner.read_float(buffer);
  FPrimitiveBoundingSphere.Center.z := FOwner.read_float(buffer);
  if FOwner.Version > 61 then
    FPrimitiveBoundingSphere.Radius := FOwner.read_float(buffer)
  else
    FPrimitiveBoundingSphere.Radius := -1;
end;

procedure TUTObjectClassPrimitive.ReleaseObject;
begin
  fillchar(FPrimitiveBoundingBox, sizeof(FPrimitiveBoundingBox), 0);
  fillchar(FPrimitiveBoundingSphere, sizeof(FPrimitiveBoundingSphere), 0);
  inherited;
end;

{ TUTObjectClassMesh }

function TUTObjectClassMesh.GetAnimFrames: integer;
begin
  check_initialized;
  result := FAnimFrames;
end;

function TUTObjectClassMesh.GetAnimSeq(i: integer): TUT_Struct_AnimSeq;
begin
  check_initialized;
  result := FAnimSeqs[i];
end;

function TUTObjectClassMesh.GetAnimSeqCount: integer;
begin
  check_initialized;
  result := length(FAnimSeqs);
end;

function TUTObjectClassMesh.GetBoundingBox(
  i: integer): TUT_Struct_BoundingBox;
begin
  check_initialized;
  result := FBoundingBoxes[i];
end;

function TUTObjectClassMesh.GetBoundingBoxCount: integer;
begin
  check_initialized;
  result := length(FBoundingBoxes);
end;

function TUTObjectClassMesh.GetBoundingSphere(
  i: integer): TUT_Struct_BoundingSphere;
begin
  check_initialized;
  result := FBoundingSpheres[i];
end;

function TUTObjectClassMesh.GetBoundingSphereCount: integer;
begin
  check_initialized;
  result := length(FBoundingSpheres);
end;

function TUTObjectClassMesh.GetConnect(i: integer): TUT_Struct_Connects;
begin
  check_initialized;
  result := FConnects[i];
end;

function TUTObjectClassMesh.GetConnectsCount: integer;
begin
  check_initialized;
  result := length(FConnects);
end;

function TUTObjectClassMesh.GetTexture(i: integer): TUT_Struct_Texture;
begin
  check_initialized;
  result := FTextures[i];
end;

function TUTObjectClassMesh.GetTextureLOD(i: integer): single;
begin
  check_initialized;
  result := FTextureLOD[i];
end;

function TUTObjectClassMesh.GetTextureLODCount: integer;
begin
  check_initialized;
  result := length(FTextureLOD);
end;

function TUTObjectClassMesh.GetTexturesCount: integer;
begin
  check_initialized;
  result := length(FTextures);
end;

function TUTObjectClassMesh.GetTri(i: integer): TUT_Struct_Tri;
begin
  check_initialized;
  result := FTris[i];
end;

function TUTObjectClassMesh.GetTrisCount: integer;
begin
  check_initialized;
  result := length(FTris);
end;

function TUTObjectClassMesh.GetVert(i: integer): TUT_Struct_Vert;
begin
  check_initialized;
  result := FVerts[i];
end;

function TUTObjectClassMesh.GetVertLink(i: integer): integer;
begin
  check_initialized;
  result := FVertLinks[i];
end;

function TUTObjectClassMesh.GetVertLinksCount: integer;
begin
  check_initialized;
  result := length(FVertLinks);
end;

function TUTObjectClassMesh.GetVertsCount: integer;
begin
  check_initialized;
  result := length(FVerts);
end;

procedure TUTObjectClassMesh.InitializeObject;
begin
  inherited;
  setlength(FVerts, 0);
  setlength(FTris, 0);
  setlength(FTextures, 0);
  setlength(FAnimSeqs, 0);
  setlength(FConnects, 0);
  setlength(FBoundingBoxes, 0);
  setlength(FBoundingSpheres, 0);
  setlength(FVertLinks, 0);
  setlength(FTextureLOD, 0);
end;

procedure TUTObjectClassMesh.InterpretObject;
var
  a, a2, seekpos, size, size2, p, game: integer;
  xyz64: int64;
  x, y, z: single;
  xyz: cardinal;
begin
  inherited;
  // UMesh.Verts
  if FOwner.Version > 61 then
    seekpos := FOwner.read_int(buffer) - FStartInPackage
  else
    seekpos := 0;
  size := FOwner.read_idx(buffer);
  p := buffer.position;
  if (size <> 0) and (seekpos > 0) and ((seekpos - p) div size = 8) then
    game := 1                           // DeusEx
  else
    game := 0;                          // UT
  setlength(FVerts, size);
  case game of
    1:
      begin                             // DeusEX vertices have more resolution
        for a := 0 to size - 1 do
          begin
            FOwner.read_buffer(xyz64, 8, buffer);
            x := (xyz64 and $FFFF) / 256;
            y := ((xyz64 shr 16) and $FFFF) / 256;
            z := ((xyz64 shr 32) and $FFFF) / 256;
            if y > 128 then y := y - 256;
            if x > 128 then x := x - 256;
            if z > 128 then z := z - 256;
            FVerts[a].x := x;
            FVerts[a].y := y;
            FVerts[a].z := z;
          end
      end;
  else
    begin                               // 0 = UT
      for a := 0 to size - 1 do
        begin
          FOwner.read_buffer(xyz, 4, buffer);
          x := (xyz and $7FF) / 8;
          y := ((xyz shr 11) and $7FF) / 8;
          z := ((xyz shr 22) and $3FF) / 4;
          if y > 128 then y := y - 256;
          if x > 128 then x := x - 256;
          if z > 128 then z := z - 256;
          FVerts[a].x := x;
          FVerts[a].y := y;
          FVerts[a].z := z;
        end;
    end;
  end;
  // UMesh.Tris
  if FOwner.Version > 61 then FOwner.read_int(buffer);
  size := FOwner.read_idx(buffer);
  setlength(FTris, size);
  for a := 0 to size - 1 do
    begin
      FTris[a].vertexindex1 := FOwner.read_word(buffer);
      FTris[a].vertexindex2 := FOwner.read_word(buffer);
      FTris[a].vertexindex3 := FOwner.read_word(buffer);
      FTris[a].u1 := FOwner.read_byte(buffer);
      FTris[a].v1 := FOwner.read_byte(buffer);
      FTris[a].u2 := FOwner.read_byte(buffer);
      FTris[a].v2 := FOwner.read_byte(buffer);
      FTris[a].u3 := FOwner.read_byte(buffer);
      FTris[a].v3 := FOwner.read_byte(buffer);
      FTris[a].flags := FOwner.read_int(buffer);
      FTris[a].textureindex := FOwner.read_int(buffer);
    end;
  // UMesh.AnimSeqs
  size := FOwner.read_idx(buffer);
  setlength(Fanimseqs, size);
  for a := 0 to size - 1 do
    begin
      FAnimSeqs[a].name := FOwner.read_idx(buffer); // name index
      FAnimSeqs[a].group := FOwner.read_idx(buffer); // group index
      FAnimSeqs[a].startframe := FOwner.read_int(buffer); // startframe
      FAnimSeqs[a].numframes := FOwner.read_int(buffer); // numframes
      size2 := FOwner.read_idx(buffer);
      setlength(FAnimSeqs[a].functions, size2);
      for a2 := 0 to size2 - 1 do
        begin
          FAnimSeqs[a].functions[a2].time := FOwner.read_float(buffer); //time
          FAnimSeqs[a].functions[a2]._function := FOwner.read_idx(buffer); // function name index
        end;
      FAnimSeqs[a].rate := FOwner.read_float(buffer); //rate
    end;
  // UMesh.Connects
  if FOwner.Version > 61 then FOwner.read_int(buffer);
  size := FOwner.read_idx(buffer);
  setlength(FConnects, size);
  for a := 0 to size - 1 do
    begin
      FConnects[a].NumVertTriangles := FOwner.read_int(buffer); // numverttriangles
      FConnects[a].TriangleListOffset := FOwner.read_int(buffer); // trianglelistoffset
    end;
  // UMesh.BoundingBox
  FBoundingBox.Min.x := FOwner.read_float(buffer);
  FBoundingBox.Min.y := FOwner.read_float(buffer);
  FBoundingBox.Min.z := FOwner.read_float(buffer);
  FBoundingBox.Max.x := FOwner.read_float(buffer);
  FBoundingBox.Max.y := FOwner.read_float(buffer);
  FBoundingBox.Max.z := FOwner.read_float(buffer);
  FBoundingBox.Valid := FOwner.read_byte(buffer);
  // UMesh.BoundingSphere
  FBoundingSphere.Center.x := FOwner.read_float(buffer);
  FBoundingSphere.Center.y := FOwner.read_float(buffer);
  FBoundingSphere.Center.z := FOwner.read_float(buffer);
  if FOwner.Version > 61 then
    FBoundingSphere.Radius := FOwner.read_float(buffer)
  else
    FBoundingSphere.Radius := -1;
  // UMesh.VertLinks
  if FOwner.Version > 61 then FOwner.read_int(buffer);
  size := FOwner.read_idx(buffer);
  setlength(FVertLinks, size);
  for a := 0 to size - 1 do
    FVertLinks[a] := FOwner.read_int(buffer);
  // UMesh.Textures
  size := FOwner.read_idx(buffer);
  setlength(FTextures, size);
  for a := 0 to size - 1 do
    FTextures[a].value := FOwner.read_idx(buffer);
  // UMesh.BoundingBoxes
  size := FOwner.read_idx(buffer);
  setlength(FBoundingBoxes, size);
  for a := 0 to size - 1 do
    begin
      FBoundingBoxes[a].Min.x := FOwner.read_float(buffer);
      FBoundingBoxes[a].Min.y := FOwner.read_float(buffer);
      FBoundingBoxes[a].Min.z := FOwner.read_float(buffer);
      FBoundingBoxes[a].Max.x := FOwner.read_float(buffer);
      FBoundingBoxes[a].Max.y := FOwner.read_float(buffer);
      FBoundingBoxes[a].Max.z := FOwner.read_float(buffer);
      FBoundingBoxes[a].Valid := FOwner.read_byte(buffer);
    end;
  // UMesh.BoundingSpheres
  size := FOwner.read_idx(buffer);
  setlength(FBoundingSpheres, size);
  for a := 0 to size - 1 do
    begin
      FBoundingSpheres[a].Center.x := FOwner.read_float(buffer);
      FBoundingSpheres[a].Center.y := FOwner.read_float(buffer);
      FBoundingSpheres[a].Center.z := FOwner.read_float(buffer);
      if FOwner.Version > 61 then
        FBoundingSpheres[a].Radius := FOwner.read_float(buffer)
      else
        FBoundingSpheres[a].Radius := -1;
    end;
  // UMesh.FrameVerts
  FFrameVerts := FOwner.read_int(buffer);
  // UMesh.AnimFrames
  FAnimFrames := FOwner.read_int(buffer);
  // UMesh.ANDFlags
  FANDFlags := FOwner.read_int(buffer);
  // UMesh.ORFlags
  FORFlags := FOwner.read_int(buffer);
  // UMesh.Scale
  FScale.x := FOwner.read_float(buffer);
  FScale.y := FOwner.read_float(buffer);
  FScale.z := FOwner.read_float(buffer);
  // UMesh.Origin
  FOrigin.x := FOwner.read_float(buffer);
  FOrigin.y := FOwner.read_float(buffer);
  FOrigin.z := FOwner.read_float(buffer);
  // UMesh.RotOrigin
  FRotOrigin.Pitch := FOwner.read_int(buffer);
  FRotOrigin.Yaw := FOwner.read_int(buffer);
  FRotOrigin.Roll := FOwner.read_int(buffer);
  // UMesh.CurPoly
  FCurPoly := FOwner.read_int(buffer);
  // UMesh.CurVertex
  FCurVertex := FOwner.read_int(buffer);
  // UMesh.TextureLOD
  if FOwner.Version >= 66 then
    begin
      size := FOwner.read_idx(buffer);
      setlength(FTextureLOD, size);
      for a := 0 to size - 1 do
        FTextureLOD[a] := FOwner.read_FLOAT(buffer);
    end
  else if FOwner.Version = 65 then
    begin
      setlength(FTextureLOD, 1);
      FTextureLOD[0] := FOwner.read_FLOAT(buffer);
    end;
end;

procedure TUTObjectClassMesh.ReleaseObject;
begin
  setlength(FVerts, 0);
  setlength(FTris, 0);
  setlength(FTextures, 0);
  setlength(FAnimSeqs, 0);
  setlength(FConnects, 0);
  setlength(FBoundingBoxes, 0);
  setlength(FBoundingSpheres, 0);
  setlength(FVertLinks, 0);
  setlength(FTextureLOD, 0);
  inherited;
end;

procedure TUTObjectClassMesh.PrepareExporter(exporter: TUT_MeshExporter; frames: TIntegerArray);
const
  material_colors: array[0..5] of TColor = (clBlue, clRed, clLime, clYellow, clAqua, clSilver);
var
  m, firstvert, v, f: integer;
begin
  check_initialized;
  if length(frames) = 0 then
    begin
      setlength(frames, 1);
      frames[0] := 0;
    end;
  setlength(exporter.Materials, length(FTextures));
  for m := 0 to high(FTextures) do
    begin
      exporter.Materials[m].name := 'Material' + inttostr(FTextures[m].Value);
      if m <= high(material_colors) then
        begin
          exporter.Materials[m].diffusecolor[0] := material_colors[m] and $FF;
          exporter.Materials[m].diffusecolor[1] := (material_colors[m] shr 8) and $FF;
          exporter.Materials[m].diffusecolor[2] := (material_colors[m] shr 16) and $FF;
        end
      else
        begin
          exporter.Materials[m].diffusecolor[0] := random(256);
          exporter.Materials[m].diffusecolor[1] := random(256);
          exporter.Materials[m].diffusecolor[2] := random(256);
        end;
    end;
  exporter.AnimationFrames := length(frames);
  setlength(exporter.Faces, length(FTris));
  setlength(exporter.Vertices, 3 * length(frames) * length(FTris));
  v := 0;
  for f := 0 to high(frames) do
    begin
      firstvert := frames[f] * FFrameVerts;
      for m := 0 to high(FTris) do
        begin
          inc(v);
          exporter.Vertices[v].x := FVerts[firstvert + FTris[m].VertexIndex1].X;
          exporter.Vertices[v].y := FVerts[firstvert + FTris[m].VertexIndex1].Y;
          exporter.Vertices[v].z := FVerts[firstvert + FTris[m].VertexIndex1].Z;
          exporter.Vertices[v].U := FTris[m].U1;
          exporter.Vertices[v].V := FTris[m].V1;
          inc(v);
          exporter.Vertices[v].x := FVerts[firstvert + FTris[m].VertexIndex2].X;
          exporter.Vertices[v].y := FVerts[firstvert + FTris[m].VertexIndex2].Y;
          exporter.Vertices[v].z := FVerts[firstvert + FTris[m].VertexIndex2].Z;
          exporter.Vertices[v].U := FTris[m].U2;
          exporter.Vertices[v].V := FTris[m].V2;
          inc(v);
          exporter.Vertices[v].x := FVerts[firstvert + FTris[m].VertexIndex3].X;
          exporter.Vertices[v].y := FVerts[firstvert + FTris[m].VertexIndex3].Y;
          exporter.Vertices[v].z := FVerts[firstvert + FTris[m].VertexIndex3].Z;
          exporter.Vertices[v].U := FTris[m].U3;
          exporter.Vertices[v].V := FTris[m].V3;
          if f = 0 then
            begin
              exporter.Faces[m].VertexIndex1 := v - 2;
              exporter.Faces[m].VertexIndex2 := v - 1;
              exporter.Faces[m].VertexIndex3 := v;
              exporter.Faces[m].MaterialIndex := FTris[m].TextureIndex;
              exporter.Faces[m].Flags := FTris[m].Flags;
            end;
        end;
    end;
end;

procedure TUTObjectClassMesh.Save_3DS(filename: string; frame: integer = 0;
  smoothing: TUT_3DStudioExporter_Smoothing = exp3ds_smooth_None; MirrorX: boolean = false);
var
  exporter: TUT_3DStudioExporter;
  frames: TIntegerArray;
begin
  check_initialized;
  exporter := TUT_3DStudioExporter.create;
  setlength(frames, 1);
  frames[0] := frame;
  PrepareExporter(exporter, frames);
  exporter.mirrorX := mirrorX;
  exporter.smoothing := smoothing;
  exporter.Save(filename);
  exporter.free;
end;

procedure TUTObjectClassMesh.Save_Unreal3D(filename: string);
var
  exporter: TUT_Unreal3DExporter;
  frames: TIntegerArray;
  a: integer;
begin
  check_initialized;
  exporter := TUT_Unreal3DExporter.create;
  setlength(frames, FAnimFrames);
  for a := 0 to FAnimFrames - 1 do
    frames[a] := a;
  PrepareExporter(exporter, frames);
  exporter.Save(filename);
  exporter.free;
end;

procedure TUTObjectClassMesh.Save_UnrealUC(filename: string);
var
  k: char;
  uc, parent_class, basename: string;
  a, b, script_idx: integer;
  ed2: TUTExportTableObjectData;
  id: TUTImportTableObjectData;
  str_uc: tfilestream;
begin
  check_initialized;
  // Find a class with the same name as the mesh
  a := FOwner.FindObject(utolExports, [utfwName, utfwClass, utfwPackage], '', UTobjectname, '');
  if a <> -1 then
    begin
      ed2 := FOwner.Exported[a];
      ed2.UTObject.ReadObject;
      parent_class := FOwner.GetObjectPath(1, TUTObjectClassClass(ed2.UTObject).SuperField);
      script_idx := TUTObjectClassClass(ed2.UTObject).ScriptText - 1;
      ed2.UTObject.ReleaseObject;
    end
  else
    begin
      parent_class := 'TournamentPlayer'; // do not localize these strings
      // Find the correct script object
      script_idx := FOwner.FindObject(utolExports, [utfwName, utfwClass, utfwPackage],
        UTobjectname, 'ScriptText', 'TextBuffer');
    end;
  if script_idx <> -1 then
    begin                               // the script was found
      ed2 := FOwner.Exported[script_idx];
      ed2.UTObject.ReadObject;
      TUTObjectClassTextBuffer(ed2.UTObject).SaveToFile(filename);
      ed2.UTObject.ReleaseObject;
    end
  else
    begin                               // the script wasn't found, we must recreate it
      k := DecimalSeparator;
      decimalSeparator := '.';
      basename := UTobjectname;
      // do not localize
      uc := '//============================================================================='#13#10;
      uc := uc + format('// %s.'#13#10, [basename]);
      uc := uc + '//============================================================================='#13#10;
      uc := uc + format('class %s extends %s;'#13#10#13#10, [basename, parent_class]);
      uc := uc + format('#exec MESH IMPORT MESH=%s ANIVFILE=MODELS\%s_a.3d DATAFILE=MODELS\%s_d.3d' {+' X=0 Y=0 Z=0'}, [basename, basename, basename]);
      uc := uc + ' MLOD=0'#13#10;
      uc := uc + format('#exec MESH ORIGIN MESH=%s X=%f Y=%f Z=%f YAW=%f ROLL=%f PITCH=%f'#13#10#13#10,
        [basename, FOrigin.x, FOrigin.y, FOrigin.z, FRotorigin.Yaw / 256, FRotOrigin.Roll / 256, FRotOrigin.Pitch / 256]);
      for a := 0 to high(FAnimSeqs) do
        begin
          uc := uc + format('#exec MESH SEQUENCE MESH=%s SEQ=%-9s STARTFRAME=%d NUMFRAMES=%d', [basename, FOwner.Names[FAnimSeqs[a].name], FAnimSeqs[a].startframe, FAnimSeqs[a].numframes]);
          if FAnimSeqs[a].rate <> 30 then uc := uc + format(' RATE=%f', [FAnimSeqs[a].rate]);
          if FOwner.Names[FAnimSeqs[a].group] <> 'None' then uc := uc + format(' GROUP=%s', [FOwner.Names[FAnimSeqs[a].group]]);
          uc := uc + #13#10;
        end;
      uc := uc + #13#10;
      for a := 0 to high(FTextures) do
        if FTextures[a].value > 0 then
          begin
            ed2 := FOwner.Exported[FTextures[a].value - 1];
            uc := uc + format('#exec TEXTURE IMPORT NAME=%s FILE=%s.PCX GROUP=%s'#13#10,
              [ed2.UTobjectname, ed2.UTobjectname, ed2.UTpackagename]);
            // TODO : TUTObjectClassMesh.Save_UnrealUC : FLAGS=%d should put correct flags...
          end
        else if FTextures[a].value < 0 then
          begin
            id := FOwner.Imported[-FTextures[a].value - 1];
            uc := uc + format('#exec OBJ LOAD FILE=%s.utx PACKAGE=%s'#13#10, [id.UTobjectname, id.UTpackagename]);
          end;
      uc := uc + #13#10;
      for a := 0 to high(FTextures) do
        if FTextures[a].value > 0 then
          begin
            ed2 := FOwner.Exported[FTextures[a].value - 1];
            uc := uc + format('#exec MESHMAP SETTEXTURE MESHMAP=%s NUM=%d TEXTURE=%s'#13#10, [basename, a, ed2.UTobjectname]);
          end
        else if FTextures[a].value < 0 then
          begin
            id := FOwner.Imported[-FTextures[a].value - 1];
            uc := uc + format('#exec MESHMAP SETTEXTURE MESHMAP=%s NUM=%d TEXTURE=%s'#13#10, [basename, a, id.UTpackagename + '.' + id.UTobjectname]);
          end;
      uc := uc + #13#10;
      //uc:=uc+format('#exec MESHMAP NEW   MESHMAP=%s MESH=%s'#13#10,[basename,basename]);
      if (FScale.x <> 0.1) or (FScale.y <> 0.1) or (FScale.z <> 0.2) then
        uc := uc + format('#exec MESHMAP SCALE MESHMAP=%s X=%f Y=%f Z=%f'#13#10#13#10, [basename, FScale.x, FScale.y, FScale.z]);
      // FScale is incorrect, sometimes is 0 and it shouldnt be.
      for a := 0 to high(FAnimSeqs) do
        for b := 0 to high(FAnimSeqs[a].functions) do
          uc := uc + format('#exec MESH NOTIFY MESH=%s SEQ=%-9s TIME=%f FUNCTION=%s'#13#10, [basename, FOwner.Names[FAnimSeqs[a].name], FAnimSeqs[a].functions[b].time, FOwner.Names[FAnimSeqs[a].functions[b]._function]]);
      uc := uc + #13#10;
      (*uc:=uc+#13#10+
      'defaultproperties'#13#10+
      '{'#13#10+
      '    DrawType=DT_Mesh'#13#10+
      format('    Mesh=%s'#13#10,[basename])+
      '}'#13#10;*)
      DecimalSeparator := k;
      str_uc := tfilestream.create(filename, fmCreate);
      try
        str_uc.write(uc[1], length(uc));
      finally
        str_uc.free;
      end;
    end;
end;

{ TUTObjectClassLodMesh }

function TUTObjectClassLodMesh.GetCollapsePointThus(i: integer): word;
begin
  check_initialized;
  result := FCollapsePointThus[i];
end;

function TUTObjectClassLodMesh.GetCollapsePointThusCount: integer;
begin
  check_initialized;
  result := length(FCollapsePointThus);
end;

function TUTObjectClassLodMesh.GetCollapseWedgeThus(i: integer): word;
begin
  check_initialized;
  result := FCollapseWedgeThus[i];
end;

function TUTObjectClassLodMesh.GetCollapseWedgeThusCount: integer;
begin
  check_initialized;
  result := length(FCollapseWedgeThus);
end;

function TUTObjectClassLodMesh.GetFace(i: integer): TUT_Struct_Face;
begin
  check_initialized;
  result := FFaces[i];
end;

function TUTObjectClassLodMesh.GetFaceCount: integer;
begin
  check_initialized;
  result := length(FFaces);
end;

function TUTObjectClassLodMesh.GetFaceLevel(i: integer): word;
begin
  check_initialized;
  result := FFaceLevel[i];
end;

function TUTObjectClassLodMesh.GetFaceLevelCount: integer;
begin
  check_initialized;
  result := length(FFaceLevel);
end;

function TUTObjectClassLodMesh.GetMaterial(
  i: integer): TUT_Struct_Material;
begin
  check_initialized;
  result := FMaterials[i];
end;

function TUTObjectClassLodMesh.GetMaterialCount: integer;
begin
  check_initialized;
  result := length(FMaterials);
end;

function TUTObjectClassLodMesh.GetRemapAnimVerts(i: integer): word;
begin
  check_initialized;
  result := FRemapAnimVerts[i];
end;

function TUTObjectClassLodMesh.GetRemapAnimVertsCount: integer;
begin
  check_initialized;
  result := length(FRemapAnimVerts);
end;

function TUTObjectClassLodMesh.GetSpecialFace(i: integer): TUT_Struct_Face;
begin
  check_initialized;
  result := FSpecialFaces[i];
end;

function TUTObjectClassLodMesh.GetSpecialFaceCount: integer;
begin
  check_initialized;
  result := length(FSpecialFaces);
end;

function TUTObjectClassLodMesh.GetWedge(i: integer): TUT_Struct_Wedge;
begin
  check_initialized;
  result := FWedges[i];
end;

function TUTObjectClassLodMesh.GetWedgeCount: integer;
begin
  check_initialized;
  result := length(FWedges);
end;

procedure TUTObjectClassLodMesh.InitializeObject;
begin
  inherited;
  setlength(FWedges, 0);
  setlength(FFaces, 0);
  setlength(FMaterials, 0);
  setlength(FCollapsePointThus, 0);
  setlength(FFaceLevel, 0);
  setlength(FCollapseWedgeThus, 0);
  setlength(FSpecialFaces, 0);
  setlength(FRemapAnimVerts, 0);
end;

procedure TUTObjectClassLodMesh.InterpretObject;
var
  size, a: integer;
begin
  inherited;
  // ULodMesh.CollapsePointThus
  size := FOwner.read_idx(buffer);
  setlength(FCollapsePointThus, size);
  for a := 0 to size - 1 do
    FCollapsePointThus[a] := FOwner.read_WORD(buffer);
  // ULodMesh.FaceLevel
  size := FOwner.read_idx(buffer);
  setlength(FFaceLevel, size);
  for a := 0 to size - 1 do
    FFaceLevel[a] := FOwner.read_WORD(buffer);
  // UlodMesh.Faces
  size := FOwner.read_idx(buffer);
  setlength(FFaces, size);
  for a := 0 to size - 1 do
    begin
      FFaces[a].WedgeIndex1 := FOwner.read_word(buffer);
      FFaces[a].WedgeIndex2 := FOwner.read_word(buffer);
      FFaces[a].WedgeIndex3 := FOwner.read_word(buffer);
      FFaces[a].MatIndex := FOwner.read_word(buffer);
    end;
  // ULodMesh.CollapseWedgeThus
  size := FOwner.read_idx(buffer);
  setlength(FCollapseWedgeThus, size);
  for a := 0 to size - 1 do
    FCollapseWedgeThus[a] := FOwner.read_WORD(buffer);
  // ULodMesh.Wedges
  size := FOwner.read_idx(buffer);
  setlength(FWedges, size);
  for a := 0 to size - 1 do
    begin
      FWedges[a].VertexIndex := FOwner.read_word(buffer);
      FWedges[a].U := FOwner.read_byte(buffer);
      FWedges[a].V := FOwner.read_byte(buffer);
    end;
  // ULodMesh.Materials
  size := FOwner.read_idx(buffer);
  setlength(FMaterials, size);
  for a := 0 to size - 1 do
    begin
      FMaterials[a].flags := FOwner.read_int(buffer);
      FMaterials[a].textureindex := FOwner.read_int(buffer);
    end;
  // ULodMesh.SpecialFaces
  size := FOwner.read_idx(buffer);
  setlength(FSpecialFaces, size);
  for a := 0 to size - 1 do
    begin
      FSpecialFaces[a].WedgeIndex1 := FOwner.read_word(buffer);
      FSpecialFaces[a].WedgeIndex2 := FOwner.read_word(buffer);
      FSpecialFaces[a].WedgeIndex3 := FOwner.read_word(buffer);
      FSpecialFaces[a].MatIndex := FOwner.read_word(buffer);
    end;
  // ULodMesh.ModelVerts
  FModelVerts := FOwner.read_int(buffer);
  // ULodMesh.SpecialVerts
  FSpecialVerts := FOwner.read_int(buffer);
  // ULodMesh.MeshScaleMax
  FMeshScaleMax := FOwner.read_FLOAT(buffer);
  // ULodMesh.LODHysteresis
  FLODHysteresis := FOwner.read_FLOAT(buffer);
  // ULodMesh.LODStrength
  FLODStrength := FOwner.read_FLOAT(buffer);
  // ULodMesh.LODMinVerts
  FLODMinVerts := FOwner.read_int(buffer);
  // ULodMesh.LODMorph
  FLODMorph := FOwner.read_FLOAT(buffer);
  // ULodMesh.LODZDisplace
  FLODZDisplace := FOwner.read_FLOAT(buffer);
  // ULodMesh.RemapAnimVerts
  size := FOwner.read_idx(buffer);
  setlength(FRemapAnimVerts, size);
  for a := 0 to size - 1 do
    FRemapAnimVerts[a] := FOwner.read_WORD(buffer);
  // ULodMesh.OldFrameVerts
  FOldFrameVerts := FOwner.read_int(buffer);

end;

procedure TUTObjectClassLodMesh.ReleaseObject;
begin
  setlength(FWedges, 0);
  setlength(FFaces, 0);
  setlength(FMaterials, 0);
  setlength(FCollapsePointThus, 0);
  setlength(FFaceLevel, 0);
  setlength(FCollapseWedgeThus, 0);
  setlength(FSpecialFaces, 0);
  setlength(FRemapAnimVerts, 0);
  inherited;
end;

procedure TUTObjectClassLodMesh.PrepareExporter(exporter: TUT_MeshExporter;
  frames: TIntegerArray);
const
  material_colors: array[0..5] of TColor = (clBlue, clRed, clLime, clYellow, clAqua, clSilver);
var
  matname: string;
  m, firstvert, f, v: integer;
begin
  check_initialized;
  if length(frames) = 0 then
    begin
      setlength(frames, 1);
      frames[0] := 0;
    end;
  setlength(exporter.Materials, length(FMaterials));
  for m := 0 to high(FMaterials) do
    begin
      matname := 'SKIN' + inttostr(FMaterials[m].textureindex);
      if (FMaterials[m].flags and (PF_TwoSided or PF_Modulated)) = (PF_TwoSided or PF_Modulated) then
        matname := matname + '.MODU'    //LATED'
      else if (FMaterials[m].flags and (PF_TwoSided or PF_Translucent)) = (PF_TwoSided or PF_Translucent) then
        matname := matname + '.TRAN'    //SLUCENT'
      else if (FMaterials[m].flags and (PF_TwoSided or PF_Masked)) = (PF_TwoSided or PF_Masked) then
        matname := matname + '.MASK'    //ED'
      else if (FMaterials[m].flags and PF_TwoSided) = PF_TwoSided then
        matname := matname + '.TWOS'    //IDED'
      else if (FMaterials[m].flags and PF_NotSolid) = PF_NotSolid then
        matname := 'WEAPON';
      matname := copy(matname, 1, 10);
      exporter.Materials[m].name := matname;
      if m <= high(material_colors) then
        begin
          exporter.Materials[m].diffusecolor[0] := material_colors[m] and $FF;
          exporter.Materials[m].diffusecolor[1] := (material_colors[m] shr 8) and $FF;
          exporter.Materials[m].diffusecolor[2] := (material_colors[m] shr 16) and $FF;
        end
      else
        begin
          exporter.Materials[m].diffusecolor[0] := random(256);
          exporter.Materials[m].diffusecolor[1] := random(256);
          exporter.Materials[m].diffusecolor[2] := random(256);
        end;
    end;
  exporter.AnimationFrames := length(frames);
  setlength(exporter.Vertices, length(FWedges) * length(frames));
  for f := 0 to high(frames) do
    begin
      firstvert := frames[f] * FFrameVerts;
      for m := 0 to high(FWedges) do
        begin
          v := f * length(FWedges) + m;
          exporter.Vertices[v].x := FVerts[FSpecialVerts + firstvert + FWedges[m].VertexIndex].X;
          exporter.Vertices[v].y := FVerts[FSpecialVerts + firstvert + FWedges[m].VertexIndex].Y;
          exporter.Vertices[v].z := FVerts[FSpecialVerts + firstvert + FWedges[m].VertexIndex].Z;
          exporter.Vertices[v].U := FWedges[m].U;
          exporter.Vertices[v].V := FWedges[m].V;
        end;
    end;
  setlength(exporter.Faces, length(FFaces));
  for m := 0 to high(FFaces) do
    begin
      exporter.Faces[m].VertexIndex1 := FFaces[m].WedgeIndex1;
      exporter.Faces[m].VertexIndex2 := FFaces[m].WedgeIndex2;
      exporter.Faces[m].VertexIndex3 := FFaces[m].WedgeIndex3;
      exporter.Faces[m].MaterialIndex := FFaces[m].MatIndex;
      exporter.Faces[m].Flags := FMaterials[FFaces[m].MatIndex].Flags;
    end;
end;

procedure TUTObjectClassLodMesh.Save_3DS(filename: string; frame: integer = 0;
  smoothing: TUT_3DStudioExporter_Smoothing = exp3ds_smooth_None; MirrorX: boolean = false);
var
  exporter: TUT_3DStudioExporter;
  frames: TIntegerArray;
begin
  check_initialized;
  exporter := TUT_3DStudioExporter.create;
  setlength(frames, 1);
  frames[0] := frame;
  PrepareExporter(exporter, frames);
  exporter.mirrorx := mirrorx;
  exporter.smoothing := smoothing;
  exporter.Save(filename);
  exporter.free;
end;

procedure TUTObjectClassLodMesh.Save_Unreal3D(filename: string);
var
  exporter: TUT_Unreal3DExporter;
  frames: TIntegerArray;
  a: integer;
begin
  check_initialized;
  exporter := TUT_Unreal3DExporter.create;
  setlength(frames, FAnimFrames);
  for a := 0 to FAnimFrames - 1 do
    frames[a] := a;
  PrepareExporter(exporter, frames);
  exporter.Save(filename);
  exporter.free;
end;

procedure TUTObjectClassLodMesh.Save_UnrealUC(filename: string);
var
  k: char;
  uc, parent_class, basename: string;
  a, b, script_idx: integer;
  ed2: TUTExportTableObjectData;
  id: TUTImportTableObjectData;
  str_uc: tfilestream;
begin
  check_initialized;
  a := FOwner.FindObject(utolExports, [utfwName, utfwClass, utfwPackage], '', UTobjectname, '');
  if a <> -1 then
    begin
      ed2 := FOwner.Exported[a];
      ed2.UTObject.ReadObject;
      parent_class := FOwner.GetObjectPath(1, TUTObjectClassClass(ed2.UTObject).SuperField);
      script_idx := TUTObjectClassClass(ed2.UTObject).ScriptText - 1;
      ed2.UTObject.ReleaseObject;
    end
  else
    begin
      parent_class := 'TournamentPlayer'; // do not localize these strings
      script_idx := FOwner.FindObject(utolExports, [utfwName, utfwClass, utfwPackage],
        UTobjectname, 'ScriptText', 'TextBuffer');
    end;
  if script_idx <> -1 then
    begin
      ed2 := FOwner.Exported[script_idx];
      ed2.UTObject.ReadObject;
      TUTObjectClassTextBuffer(ed2.UTObject).SaveToFile(filename);
      ed2.UTObject.ReleaseObject;
    end
  else
    begin
      k := DecimalSeparator;
      decimalSeparator := '.';
      basename := UTobjectname;
      // do not localize
      uc := '//============================================================================='#13#10;
      uc := uc + format('// %s.'#13#10, [basename]);
      uc := uc + '//============================================================================='#13#10;
      uc := uc + format('class %s extends %s;'#13#10#13#10, [basename, parent_class]);
      uc := uc + format('#exec MESH IMPORT MESH=%s ANIVFILE=MODELS\%s_a.3d DATAFILE=MODELS\%s_d.3d' {+' X=0 Y=0 Z=0'}, [basename, basename, basename]);
      uc := uc + #13#10;
      uc := uc + format('#exec MESH LODPARAMS MESH=%s HYSTERESIS=%f STRENGTH=%f MINVERTS=%f MORPH=%f ZDISP=%f'#13#10,
        [basename, FLODHysteresis, FLODStrength, FLODMinVerts, FLODMorph, FLODZDisplace]);
      uc := uc + format('#exec MESH ORIGIN MESH=%s X=%f Y=%f Z=%f YAW=%f ROLL=%f PITCH=%f'#13#10#13#10,
        [basename, FOrigin.x, FOrigin.y, FOrigin.z, FRotorigin.Yaw / 256, FRotOrigin.Roll / 256, FRotOrigin.Pitch / 256]);
      for a := 0 to high(FAnimSeqs) do
        begin
          uc := uc + format('#exec MESH SEQUENCE MESH=%s SEQ=%-9s STARTFRAME=%d NUMFRAMES=%d', [basename, FOwner.Names[FAnimSeqs[a].name], FAnimSeqs[a].startframe, FAnimSeqs[a].numframes]);
          if FAnimSeqs[a].rate <> 30 then uc := uc + format(' RATE=%f', [FAnimSeqs[a].rate]);
          if FOwner.Names[FAnimSeqs[a].group] <> 'None' then uc := uc + format(' GROUP=%s', [FOwner.Names[FAnimSeqs[a].group]]);
          uc := uc + #13#10;
        end;
      uc := uc + #13#10;
      for a := 0 to high(FTextures) do
        if FTextures[a].value > 0 then
          begin
            ed2 := FOwner.Exported[FTextures[a].value - 1];
            uc := uc + format('#exec TEXTURE IMPORT NAME=%s FILE=%s.PCX GROUP=%s'#13#10,
              [ed2.UTobjectname, ed2.UTobjectname, ed2.UTpackagename]);
            // TODO : TUTObjectClassLodMesh.Save_UnrealUC : FLAGS=%d should put correct flags...
          end
        else if FTextures[a].value < 0 then
          begin
            id := FOwner.Imported[-FTextures[a].value - 1];
            uc := uc + format('#exec OBJ LOAD FILE=%s.utx PACKAGE=%s'#13#10, [id.UTobjectname, id.UTpackagename]);
          end;
      uc := uc + #13#10;
      for a := 0 to high(FTextures) do
        if FTextures[a].value > 0 then
          begin
            ed2 := FOwner.Exported[FTextures[a].value - 1];
            uc := uc + format('#exec MESHMAP SETTEXTURE MESHMAP=%s NUM=%d TEXTURE=%s'#13#10, [basename, a, ed2.UTobjectname]);
          end
        else if FTextures[a].value < 0 then
          begin
            id := FOwner.Imported[-FTextures[a].value - 1];
            uc := uc + format('#exec MESHMAP SETTEXTURE MESHMAP=%s NUM=%d TEXTURE=%s'#13#10, [basename, a, id.UTpackagename + '.' + id.UTobjectname]);
          end;
      uc := uc + #13#10;
      //uc:=uc+format('#exec MESHMAP NEW   MESHMAP=%s MESH=%s'#13#10,[basename,basename]);
      if (FScale.x <> 0.1) or (FScale.y <> 0.1) or (FScale.z <> 0.2) then
        uc := uc + format('#exec MESHMAP SCALE MESHMAP=%s X=%f Y=%f Z=%f'#13#10#13#10, [basename, FScale.x, FScale.y, FScale.z]);
      // FScale is incorrect, sometimes is 0 and it shouldnt be.
      for a := 0 to high(FAnimSeqs) do
        for b := 0 to high(FAnimSeqs[a].functions) do
          uc := uc + format('#exec MESH NOTIFY MESH=%s SEQ=%-9s TIME=%f FUNCTION=%s'#13#10, [basename, FOwner.Names[FAnimSeqs[a].name], FAnimSeqs[a].functions[b].time, FOwner.Names[FAnimSeqs[a].functions[b]._function]]);
      uc := uc + #13#10;
      (*uc:=uc+#13#10+
      'defaultproperties'#13#10+
      '{'#13#10+
      '    DrawType=DT_Mesh'#13#10+
      format('    Mesh=%s'#13#10,[basename])+
      '}'#13#10;*)
      DecimalSeparator := k;
      str_uc := tfilestream.create(filename, fmCreate);
      try
        str_uc.write(uc[1], length(uc));
      finally
        str_uc.free;
      end;
    end;
end;

{ TUTObjectClassFireTexture }

function TUTObjectClassFireTexture.GetSpark(i: integer): TUT_Struct_Spark;
begin
  check_initialized;
  result := FSparks[i];
end;

function TUTObjectClassFireTexture.GetSparkCount: integer;
begin
  check_initialized;
  result := length(FSparks);
end;

procedure TUTObjectClassFireTexture.InitializeObject;
begin
  inherited;
  setlength(FSparks, 0);
end;

procedure TUTObjectClassFireTexture.InterpretObject;
var
  s: integer;
begin
  inherited;
  setlength(FSparks, FOwner.read_idx(buffer));
  for s := 0 to high(FSparks) do
    FOwner.read_buffer(FSparks[s], sizeof(FSparks[s]), buffer);
end;

procedure TUTObjectClassFireTexture.ReleaseObject;
begin
  setlength(FSparks, 0);
  inherited;
end;

{ TUTObjectClassField }

function TUTObjectClassField.GetNext: integer;
begin
  check_initialized;
  result := FNext;
end;

function TUTObjectClassField.GetSuperField: integer;
begin
  check_initialized;
  result := FSuperField;
end;

procedure TUTObjectClassField.InitializeObject;
begin
  inherited;
  FSuperField := 0;
  FNext := 0;
end;

procedure TUTObjectClassField.InterpretObject;
begin
  inherited;
  FSuperField := FOwner.read_idx(buffer);
  FNext := FOwner.read_idx(buffer);
end;

procedure TUTObjectClassField.ReleaseObject;
begin
  FSuperField := 0;
  FNext := 0;
  inherited;
end;

{ TUTObjectEnum }

function TUTObjectClassEnum.GetCount: integer;
begin
  check_initialized;
  result := length(FValues);
end;

function TUTObjectClassEnum.GetDeclaration: string;
var
  a: integer;
begin
  check_initialized;
  result := '';
  for a := 0 to count - 1 do
    begin
      result := result + EnumName[a];
      if a < count - 1 then result := result + ',';
      result := result + #13#10;
    end;
  result := FOwner.IndentText(#9, result);
  result := 'enum ' + UTobjectname + ' {'#13#10 + result + '}';
end;

function TUTObjectClassEnum.GetValue(i: integer): integer;
begin
  check_initialized;
  if (i >= 0) and (i < length(FValues)) then
    result := FValues[i]
  else
    result := 0;
end;

function TUTObjectClassEnum.GetValueName(i: integer): string;
begin
  check_initialized;
  if (i >= 0) and (i < length(FValues)) then
    result := FOwner.Names[FValues[i]]
  else
    result := '';
end;

procedure TUTObjectClassEnum.InitializeObject;
begin
  inherited;
  setlength(FValues, 0);
end;

procedure TUTObjectClassEnum.InterpretObject;
var
  a: integer;
begin
  inherited;
  setlength(FValues, FOwner.read_idx(buffer));
  for a := 0 to high(FValues) do
    FValues[a] := FOwner.read_idx(buffer);
end;

procedure TUTObjectClassEnum.ReleaseObject;
begin
  setlength(FValues, 0);
  inherited;
end;

{ TUTObjectClassConst }

function TUTObjectClassConst.GetDeclaration: string;
begin
  check_initialized;
  result := 'const ' + UTobjectname + '=' + Value;
end;

function TUTObjectClassConst.GetValue: string;
begin
  check_initialized;
  result := FValue;
end;

procedure TUTObjectClassConst.InitializeObject;
begin
  inherited;
  FValue := '';
end;

procedure TUTObjectClassConst.InterpretObject;
begin
  inherited;
  FValue := FOwner.read_sizedasciiz(buffer);
end;

procedure TUTObjectClassConst.ReleaseObject;
begin
  FValue := '';
  inherited;
end;

{ TUTObjectClassProperty }

function TUTObjectClassProperty.GetArrayDimension: integer;
begin
  check_initialized;
  result := FArrayDimension;
end;

function TUTObjectClassProperty.GetCategory: string;
begin
  check_initialized;
  result := FCategory;
end;

function TUTObjectClassProperty.GetDeclaration(context, cn: string): string;
var
  flags: string;
begin
  check_initialized;
  flags := GetFlags(cn);
  if (context <> '') and (copy(flags, 1, 1) <> '(') then context := context + ' ';
  result := format('%s%s%s %s', [context, flags, TypeName, UTObjectName]);
  if FArrayDimension > 1 then result := result + format('[%d]', [FArrayDimension]);
end;

function TUTObjectClassProperty.GetElementSize: integer;
begin
  check_initialized;
  result := FElementSize;
end;

function TUTObjectClassProperty.GetFlags(cn: string): string;
var
  flags, varcategory: string;
begin
  check_initialized;
  flags := '';
  if (PropertyFlags and CPF_Edit) <> 0 then
    begin
      if category = cn then
        varcategory := ''
      else
        varcategory := category;
      varcategory := '(' + varcategory + ')';
      varcategory := varcategory + ' ';
    end
  else
    varcategory := '';
  if (PropertyFlags and CPF_Const) <> 0 then flags := flags + 'const ';
  if (PropertyFlags and CPF_Input) <> 0 then flags := flags + 'input ';
  if (PropertyFlags and CPF_ExportObject) <> 0 then flags := flags + 'exportobject ';
  if (PropertyFlags and CPF_OptionalParm) <> 0 then flags := flags + 'optional ';
  //if (PropertyFlags and CPF_Net) <> 0 then flags := flags + 'net ';
  if (PropertyFlags and CPF_ConstRef) <> 0 then flags := flags + 'constref ';
  if ((PropertyFlags and CPF_OutParm) <> 0) and
    ((PropertyFlags and CPF_ReturnParm) = 0) then flags := flags + 'out ';
  if (PropertyFlags and CPF_SkipParm) <> 0 then flags := flags + 'skip ';
  if (PropertyFlags and CPF_CoerceParm) <> 0 then flags := flags + 'coerce ';
  if (PropertyFlags and CPF_Native) <> 0 then flags := flags + 'native ';
  if (PropertyFlags and CPF_Transient) <> 0 then flags := flags + 'transient ';
  if ((PropertyFlags and CPF_Config) <> 0) and
    ((PropertyFlags and CPF_GlobalConfig) = 0) then flags := flags + 'config ';
  if (PropertyFlags and CPF_Localized) <> 0 then flags := flags + 'localized ';
  if (PropertyFlags and CPF_Travel) <> 0 then flags := flags + 'travel ';
  if (PropertyFlags and CPF_EditConst) <> 0 then flags := flags + 'editconst ';
  if (PropertyFlags and CPF_GlobalConfig) <> 0 then flags := flags + 'globalconfig ';
  if (PropertyFlags and CPF_OnDemand) <> 0 then flags := flags + 'ondemand ';
  if (PropertyFlags and CPF_New) <> 0 then flags := flags + 'new ';
  //if (variable.PropertyFlags and CPF_NeedCtorLink) <> 0 then flags := flags + 'needctorlink ';
  if (UTflags and RF_Public) = 0 then flags := flags + 'private ';
  result := varcategory + flags;
end;

function TUTObjectClassProperty.GetPropertyFlags: longword;
begin
  check_initialized;
  result := FPropertyFlags;
end;

function TUTObjectClassProperty.GetRepOffset: word;
begin
  check_initialized;
  result := FRepOffset;
end;

procedure TUTObjectClassProperty.InitializeObject;
begin
  inherited;
  FArrayDimension := 0;
  FElementSize := 0;
  FPropertyFlags := 0;
  FCategory := '';
  FRepOffset := $FFFF;
end;

procedure TUTObjectClassProperty.InterpretObject;
begin
  inherited;
  FArrayDimension := FOwner.read_word(buffer);
  FElementSize := FOwner.read_word(buffer);
  FPropertyFlags := FOwner.read_int(buffer);
  FCategory := FOwner.Names[FOwner.read_idx(buffer)];
  if (FPropertyFlags and CPF_Net) <> 0 then FRepOffset := FOwner.read_word(buffer);
end;

procedure TUTObjectClassProperty.ReleaseObject;
begin
  FArrayDimension := 0;
  FElementSize := 0;
  FPropertyFlags := 0;
  FCategory := '';
  FRepOffset := $FFFF;
  inherited;
end;

function TUTObjectClassProperty.TypeName: string;
begin
  check_initialized;
  result := 'property';
end;

{ TUTObjectClassByteProperty }

function TUTObjectClassByteProperty.GetEnum: byte;
begin
  check_initialized;
  result := FEnum;
end;

procedure TUTObjectClassByteProperty.InitializeObject;
begin
  inherited;
  FEnum := 0;
end;

procedure TUTObjectClassByteProperty.InterpretObject;
begin
  inherited;
  FEnum := FOwner.read_idx(buffer);
end;

procedure TUTObjectClassByteProperty.ReleaseObject;
begin
  FEnum := 0;
  inherited;
end;

function TUTObjectClassByteProperty.TypeName: string;
begin
  check_initialized;
  if FEnum = 0 then
    result := 'byte'
  else
    result := FOwner.GetObjectPath(1, FEnum);
end;

{ TUTObjectClassObjectProperty }

function TUTObjectClassObjectProperty.GetObject: integer;
begin
  check_initialized;
  result := FObject;
end;

procedure TUTObjectClassObjectProperty.InitializeObject;
begin
  inherited;
  FObject := 0;
end;

procedure TUTObjectClassObjectProperty.InterpretObject;
begin
  inherited;
  FObject := FOwner.read_idx(buffer);
end;

procedure TUTObjectClassObjectProperty.ReleaseObject;
begin
  FObject := 0;
  inherited;
end;

function TUTObjectClassObjectProperty.TypeName: string;
begin
  check_initialized;
  result := FOwner.GetObjectPath(1, FObject);
end;

{ TUTObjectClassFixedArrayProperty }

function TUTObjectClassFixedArrayProperty.GetCount: integer;
begin
  check_initialized;
  result := FCount;
end;

function TUTObjectClassFixedArrayProperty.GetInner: integer;
begin
  check_initialized;
  result := FInner;
end;

procedure TUTObjectClassFixedArrayProperty.InitializeObject;
begin
  inherited;
  FInner := 0;
  FCount := 0;
end;

procedure TUTObjectClassFixedArrayProperty.InterpretObject;
begin
  inherited;
  FInner := FOwner.read_idx(buffer);
  FCount := FOwner.read_idx(buffer);
end;

procedure TUTObjectClassFixedArrayProperty.ReleaseObject;
begin
  FInner := 0;
  FCount := 0;
  inherited;
end;

function TUTObjectClassFixedArrayProperty.TypeName: string;
begin
  check_initialized;
  result := FOwner.GetObjectPath(1, FInner) + '[' + inttostr(FCount) + ']'; // TODO : the variable name should go in-between?
end;

{ TUTObjectClassArrayProperty }

function TUTObjectClassArrayProperty.GetInner: integer;
begin
  check_initialized;
  result := FInner;
end;

procedure TUTObjectClassArrayProperty.InitializeObject;
begin
  inherited;
  FInner := 0;
end;

procedure TUTObjectClassArrayProperty.InterpretObject;
begin
  inherited;
  FInner := FOwner.read_idx(buffer);
end;

procedure TUTObjectClassArrayProperty.ReleaseObject;
begin
  FInner := 0;
  inherited;
end;

function TUTObjectClassArrayProperty.TypeName: string;
var
  inner: string;
begin
  check_initialized;
  FOwner.Exported[FInner - 1].UTObject.ReadObject;
  inner := TUTObjectClassProperty(FOwner.Exported[FInner - 1].UTObject).TypeName;
  FOwner.Exported[FInner - 1].UTObject.ReleaseObject;
  result := 'array<' + inner + '>';
end;

{ TUTObjectClassMapProperty }

function TUTObjectClassMapProperty.GetKey: integer;
begin
  check_initialized;
  result := FKey;
end;

function TUTObjectClassMapProperty.GetValue: integer;
begin
  check_initialized;
  result := FValue;
end;

procedure TUTObjectClassMapProperty.InitializeObject;
begin
  inherited;
  FKey := 0;
  FValue := 0;
end;

procedure TUTObjectClassMapProperty.InterpretObject;
begin
  inherited;
  FKey := FOwner.read_idx(buffer);
  FValue := FOwner.read_idx(buffer);
end;

procedure TUTObjectClassMapProperty.ReleaseObject;
begin
  FKey := 0;
  FValue := 0;
  inherited;
end;

function TUTObjectClassMapProperty.TypeName: string;
begin
  check_initialized;
  result := 'map';                      // TODO : fix map typename
end;

{ TUTObjectClassClassProperty }

function TUTObjectClassClassProperty.GetClass: integer;
begin
  check_initialized;
  result := FClass;
end;

procedure TUTObjectClassClassProperty.InitializeObject;
begin
  inherited;
  FClass := 0;
end;

procedure TUTObjectClassClassProperty.InterpretObject;
begin
  inherited;
  FClass := FOwner.read_idx(buffer);
end;

procedure TUTObjectClassClassProperty.ReleaseObject;
begin
  FClass := 0;
  inherited;
end;

function TUTObjectClassClassProperty.TypeName: string;
begin
  check_initialized;
  result := FOwner.GetObjectPath(1, FClass);
  if lowercase(result) = 'object' then
    result := ''
  else
    result := '<' + result + '>';
  result := inherited TypeName + result;
end;

{ TUTObjectClassStructProperty }

function TUTObjectClassStructProperty.GetStruct: integer;
begin
  check_initialized;
  result := FStruct;
end;

procedure TUTObjectClassStructProperty.InitializeObject;
begin
  inherited;
  FStruct := 0;
end;

procedure TUTObjectClassStructProperty.InterpretObject;
begin
  inherited;
  FStruct := FOwner.read_idx(buffer);
end;

procedure TUTObjectClassStructProperty.ReleaseObject;
begin
  FStruct := 0;
  inherited;
end;

function TUTObjectClassStructProperty.TypeName: string;
begin
  check_initialized;
  result := FOwner.GetObjectPath(1, FStruct);
end;

{ TUTObjectClassStruct }

function TUTObjectClassStruct.GetDeclaration: string;
var
  c: integer;
  child: TUTObjectClassField;
  r: string;
begin
  check_initialized;
  r := '';
  c := FChildren;
  while c <> 0 do
    begin
      child := TUTObjectClassField(FOwner.Exported[c - 1].UTObject);
      child.ReadObject;
      try
        if child is TUTObjectClassProperty then
          r := r + TUTObjectClassProperty(child).GetDeclaration('var', UTObjectName) + ';'#13#10;
        if child is TUTObjectClassEnum then
          r := r + TUTObjectClassEnum(child).GetDeclaration + ';'#13#10;
        c := child.next;
      finally
        child.ReleaseObject;
      end;
    end;
  r := FOwner.IndentText(#9, r);
  result := 'struct ' + UTObjectName;
  if FSuperField = 0 then
    result := result + #13#10'{'#13#10
  else
    result := result + ' expands ' + FOwner.GetObjectPath(1, FSuperField) + #13#10'{'#13#10;
  result := result + r + '}';
end;

procedure TUTObjectClassStruct.InitializeObject;
begin
  inherited;
  FScriptText := 0;
  FChildren := 0;
  FFriendlyName := '';
  FLine := 0;
  FTextPos := 0;
  FScriptSize := 0;
  FScriptStart := 0;
  JumpList := nil;
  indent_level := 0;
  position_icode := 0;
  setlength(FLabelTable, 0);
end;

procedure TUTObjectClassStruct.InterpretObject;
begin
  inherited;
  FScriptText := FOwner.read_idx(buffer);
  FChildren := FOwner.read_idx(buffer);
  FFriendlyName := FOwner.Names[FOwner.read_idx(buffer)];
  FLine := FOwner.read_int(buffer);
  FTextPos := FOwner.read_int(buffer);
  FScriptSize := FOwner.read_int(buffer);
  FScriptStart := buffer.position;
  JumpList := tlist.create;
  position_icode := 0;
  indent_level := 0;
  setlength(FLabelTable, 0);
  SkipStatements;
end;

procedure TUTObjectClassStruct.ReleaseObject;
begin
  FScriptText := 0;
  FChildren := 0;
  FFriendlyName := '';
  FTextPos := 0;
  FLine := 0;
  FScriptSize := 0;
  FScriptStart := 0;
  freeandnil(JumpList);
  position_icode := 0;
  indent_level := 0;
  setlength(FLabelTable, 0);
  inherited;
end;

function TUTObjectClassStruct.ReadToken(OuterOperatorPrecedence: byte): string;
var
  b: byte;
  i1, i2, i3: integer;
  f1, f2, f3: single;
  r1, r2, r3, r4, msg: string;
  ds: char;
  wc: widestring;
  procedure read_parameters(var r: string);
  var
    r1: string;
  begin
    repeat
      r1 := ReadToken;
      if r1 = ')' then
        begin
          if copy(r, length(r), 1) = ',' then
            delete(r, length(r), 1);    // remove last comma
          r := r + r1;
        end
      else
        r := r + r1 + ',';
    until r1 = ')';
  end;
begin
  result := '';                         // gives a false warning
  check_initialized;
  inc(indent_level);
  b := FOwner.read_byte(buffer);
  inc(position_icode);
  case b of
    EX_JumpIfNot:
      begin
        i1 := FOwner.read_word(buffer);
        inc(position_icode, 2);
        r1 := ReadToken(OuterOperatorPrecedence); //(0); // the zero precedence will make it include parenthesis if it is an expression
        if (nest <> nil) and
          (i1 > position_icode) then    // we jump forward (the exclusion of an equal address is intentional)
          begin
            endnestlist.addobject('IF', pointer((position_icode shl 16) or i1));
            result := format('if ( %s )', [r1]);
            nest.add(pointer(NEST_If));
            need_semicolon := false;
          end
        else
          begin
            i2 := jumplist.indexof(pointer(i1));
            if i2 = -1 then i2 := jumplist.add(pointer(i1));
            //if copy(r1, 1, 1) <> '(' then r1 := '(' + r1 + ')';
            result := format('if (! %s ) goto JL%-4.4x', [r1, integer(jumplist[i2])]);
            need_semicolon := true;
          end;
      end;
    EX_LocalVariable, EX_InstanceVariable,
      EX_NativeParm:                    // EX_NativeParm will always be in the body of a native function
      begin
        i1 := FOwner.read_idx(buffer);
        inc(position_icode, 4);
        result := FOwner.GetObjectPath(1, i1);
      end;
    EX_DefaultVariable:
      begin
        i1 := FOwner.read_idx(buffer);
        inc(position_icode, 4);
        result := 'Default.' + FOwner.GetObjectPath(1, i1);
      end;
    EX_Return:
      begin
        result := ReadToken(OuterOperatorPrecedence);
        if result <> '' then result := ' ' + result;
        result := 'return' + result;
        need_semicolon := true;
      end;
    EX_Nothing:
      begin
        result := '';
      end;
    EX_Let, EX_LetBool:
      begin
        r1 := ReadToken(OuterOperatorPrecedence);
        r2 := ReadToken(OuterOperatorPrecedence);
        result := r1 + '=' + r2;
        need_semicolon := true;
      end;
    EX_ClassContext, EX_Context:
      begin
        result := ReadToken(OuterOperatorPrecedence) + '.';
        // following fields only used when class context is null -> not needed for source code
        FOwner.read_word(buffer);
        inc(position_icode, 2);         // wSkip
        FOwner.read_byte(buffer);
        inc(position_icode);            // bSize
        context_change := true;
        result := result + ReadToken(OuterOperatorPrecedence);
        context_change := false;
      end;
    EX_EndFunctionParms:
      begin
        result := ')';
      end;
    EX_Self:
      begin
        result := 'self';
      end;
    EX_IntConst:
      begin
        result := format('%d', [FOwner.read_int(buffer)]);
        inc(position_icode, 4);
      end;
    EX_ByteConst, EX_IntConstByte:
      begin
        result := format('%d', [FOwner.read_byte(buffer)]);
        inc(position_icode);
      end;
    EX_FloatConst:
      begin
        ds := DecimalSeparator;
        DecimalSeparator := '.';
        result := format('%f', [FOwner.read_float(buffer)]);
        DecimalSeparator := ds;
        inc(position_icode, 4);
      end;
    EX_ObjectConst:
      begin
        i1 := FOwner.read_idx(buffer);
        if i1 > 0 then
          result := FOwner.exported[i1 - 1].classname
        else
          result := FOwner.imported[-i1 - 1].classname;
        if result = '' then result := 'Class';
        result := format('%s''%s''', [result, FOwner.GetObjectPath(1, i1)]);
        inc(position_icode, 4);
      end;
    EX_NameConst:
      begin
        result := format('''%s''', [FOwner.Names[FOwner.read_idx(buffer)]]);
        inc(position_icode, 4);
      end;
    EX_StringConst:
      begin
        r1 := FOwner.read_asciiz(buffer);
        result := '"' + r1 + '"';
        inc(position_icode, length(r1) + 1);
      end;
    EX_UnicodeStringConst:
      begin
        wc := '';
        repeat
          i1 := FOwner.read_word(buffer);
          inc(position_icode, 2);
          if i1 > 0 then wc := wc + widechar(i1);
        until i1 = 0;
        result := '"' + wc + '"';
      end;
    EX_EatString:
      begin
        result := ReadToken(OuterOperatorPrecedence);
      end;
    EX_RotationConst:
      begin
        i1 := FOwner.read_int(buffer);
        inc(position_icode, 4);
        i2 := FOwner.read_int(buffer);
        inc(position_icode, 4);
        i3 := FOwner.read_int(buffer);
        inc(position_icode, 4);
        result := format('rot(%d,%d,%d)', [i1, i2, i3]);
      end;
    EX_VectorConst:
      begin
        f1 := FOwner.read_float(buffer);
        inc(position_icode, 4);
        f2 := FOwner.read_float(buffer);
        inc(position_icode, 4);
        f3 := FOwner.read_float(buffer);
        inc(position_icode, 4);
        ds := DecimalSeparator;
        DecimalSeparator := '.';
        result := format('vect(%f,%f,%f)', [f1, f2, f3]);
        DecimalSeparator := ds;
      end;
    EX_IntZero:
      begin
        result := '0';
      end;
    EX_IntOne:
      begin
        result := '1';
      end;
    EX_True:
      begin
        result := 'True';
      end;
    EX_False:
      begin
        result := 'False';
      end;
    EX_NoObject:
      begin
        result := 'None';
      end;
    EX_DynamicCast:
      begin
        i1 := FOwner.read_idx(buffer);
        inc(position_icode, 4);
        result := FOwner.GetObjectPath(1, i1) + '(' + ReadToken(OuterOperatorPrecedence) + ')';
      end;
    EX_MetaCast:
      begin
        i1 := FOwner.read_idx(buffer);
        inc(position_icode, 4);
        result := 'Class<' + FOwner.GetObjectPath(1, i1) + '>(' + ReadToken(OuterOperatorPrecedence) + ')';
      end;
    EX_StructMember:
      begin
        i1 := FOwner.read_idx(buffer);
        inc(position_icode, 4);
        result := ReadToken(OuterOperatorPrecedence) + '.' + FOwner.GetObjectPath(1, i1);
      end;
    EX_Skip:
      begin
        FOwner.read_word(buffer);
        inc(position_icode, 2);         // jump address
        result := ReadToken(OuterOperatorPrecedence);
      end;
    EX_VectorToRotator, EX_StringToRotator:
      begin
        result := 'rotator(' + ReadToken(OuterOperatorPrecedence) + ')';
      end;
    EX_RotatorToVector, EX_StringToVector:
      begin
        result := 'vector(' + ReadToken(OuterOperatorPrecedence) + ')';
      end;
    EX_ObjectToBool, EX_NameToBool, EX_StringToBool, EX_VectorToBool, EX_RotatorToBool:
      begin
        result := 'bool(' + ReadToken(OuterOperatorPrecedence) + ')';
      end;
    EX_StringToByte:
      begin
        result := 'byte(' + ReadToken(OuterOperatorPrecedence) + ')';
      end;
    EX_StringToInt:
      begin
        result := 'int(' + ReadToken(OuterOperatorPrecedence) + ')';
      end;
    EX_StringToFloat:
      begin
        result := 'float(' + ReadToken(OuterOperatorPrecedence) + ')';
      end;
    EX_ByteToString, EX_IntToString, EX_BoolToString, EX_FloatToString,
      EX_ObjectToString, EX_NameToString, EX_VectorToString, EX_RotatorToString:
      begin
        result := 'string(' + ReadToken(OuterOperatorPrecedence) + ')';
      end;
    EX_StringToName:
      begin
        result := 'name(' + ReadToken(OuterOperatorPrecedence) + ')';
      end;
    EX_ByteToInt, EX_ByteToBool, EX_ByteToFloat, EX_IntToByte,
      EX_IntToBool, EX_IntToFloat, EX_BoolToByte, EX_BoolToInt, EX_BoolToFloat,
      EX_FloatToByte, EX_FloatToInt, EX_FloatToBool, EX_BoolVariable:
      begin
        result := ReadToken(OuterOperatorPrecedence);
      end;
    EX_Jump:
      begin
        i1 := FOwner.read_word(buffer);
        inc(position_icode, 2);         // jump address
        // check "If" case (end of if block)
        if (nest <> nil) and
          (nest.count > 0) and
          (endnestlist.count > 0) and
          (nest[nest.count - 1] = pointer(NEST_If)) and // we are inside an If statement
        (position_icode = (integer(endnestlist.objects[endnestlist.count - 1]) and $FFFF)) and // and we are at the end of the If block
        (position_icode - 3 > integer(endnestlist.objects[endnestlist.count - 1]) shr 16) and // but we are not at the start (case of a break inside an if)
        (i1 >= position_icode) then     // and we jump forward
          begin
            result := #8'}'#13#10'else'#13#10'{'; // the initial backspace is to correct indentation
            endnestlist[endnestlist.count - 1] := 'ELSE';
            endnestlist.objects[endnestlist.count - 1] := pointer(integer((longword(endnestlist.objects[endnestlist.count - 1]) and $FFFF0000)) or i1);
            nest[nest.count - 1] := pointer(NEST_Else);
            need_semicolon := false;
          end
            // check "Switch" case (break statement -> end of switch block)
        else if (nest <> nil) and
          (nest.count > 0) and
          (nest[nest.count - 1] = pointer(NEST_Switch)) and // we are inside a Switch statement
        (i1 >= position_icode) then     // and we jump forward
          begin
            result := 'break';
            need_semicolon := true;
            if ((endnestlist.count = 0) or ((integer(endnestlist.objects[endnestlist.count - 1]) and $FFFF) <> i1)) then
              endnestlist.addobject('SWITCH-BREAK', pointer((position_icode shl 16) or i1));
          end
        else
          begin
            i2 := jumplist.indexof(pointer(i1));
            if i2 = -1 then i2 := jumplist.add(pointer(i1));
            result := format('goto JL%-4.4x', [integer(jumplist[i2])]);
            need_semicolon := true;
          end;
      end;
    EX_ArrayElement:
      begin
        r1 := ReadToken(OuterOperatorPrecedence); // index
        r2 := ReadToken(OuterOperatorPrecedence); // base element
        result := r2 + '[' + r1 + ']';
      end;
    EX_DynArrayElement:
      begin
        r1 := ReadToken(OuterOperatorPrecedence); // index
        r2 := ReadToken(OuterOperatorPrecedence); // base element: not checked ???
        result := r2 + '[' + r1 + ']';
      end;
    EX_Switch:
      begin
        FOwner.read_byte(buffer);
        inc(position_icode);            // switch size
        r1 := ReadToken(OuterOperatorPrecedence); // switch expression
        if copy(r1, 1, 1) <> '(' then r1 := '(' + r1 + ')';
        result := 'switch ' + r1;
        need_semicolon := false;
        if nest <> nil then
          nest.add(pointer(NEST_Switch))
        else
          result := result + ' {';
      end;
    EX_Case:
      begin
        i1 := FOwner.read_word(buffer);
        inc(position_icode, 2);         // address of next case
        if i1 = $FFFF then
          begin
            result := 'default:';
            // add end of switch here is we had no break statements (but we could displace some default case statements)
            if (nest <> nil) and
              (nest.count > 0) and
              (nest[nest.count - 1] = pointer(NEST_Switch)) and // we are inside a Switch statement
            ((endnestlist.count = 0) or (endnestlist[endnestlist.count - 1] <> 'SWITCH-BREAK')) then // we dont have a previous switch end (by break)
              endnestlist.addobject('SWITCH', pointer((position_icode shl 16) or position_icode));
          end
        else
          result := 'case ' + ReadToken(OuterOperatorPrecedence) + ':';
        need_semicolon := false;
        // we dont know where the switch ends except for any "break" inside the cases
      end;
    EX_Iterator:
      begin
        r1 := ReadToken(OuterOperatorPrecedence);
        FOwner.read_word(buffer);
        inc(position_icode, 2);         // end of loop
        result := format('foreach %s', [r1]);
        need_semicolon := false;
        if nest <> nil then
          nest.add(pointer(NEST_Foreach))
        else
          result := result + ' {';
      end;
    EX_IteratorPop:
      begin
      end;
    EX_IteratorNext:
      begin
        if nest <> nil then
          nest.delete(nest.count - 1)
        else
          result := '}';
      end;
    EX_Stop:
      begin
        result := 'stop';               // just to flag the end, then removed
      end;
    EX_Assert:
      begin
        FOwner.read_word(buffer);
        inc(position_icode, 2);         // line number
        result := 'assert (' + ReadToken(OuterOperatorPrecedence) + ')';
        need_semicolon := true;
      end;
    EX_GotoLabel:
      begin
        result := 'goto (' + ReadToken(OuterOperatorPrecedence) + ')';
        need_semicolon := true;
      end;
    EX_StructCmpEq:
      begin
        FOwner.read_idx(buffer);
        inc(position_icode, 4);
        r1 := ReadToken(OuterOperatorPrecedence);
        r2 := ReadToken(OuterOperatorPrecedence);
        result := r1 + '==' + r2;
      end;
    EX_StructCmpNe:
      begin
        FOwner.read_idx(buffer);
        inc(position_icode, 4);
        r1 := ReadToken(OuterOperatorPrecedence);
        r2 := ReadToken(OuterOperatorPrecedence);
        result := r1 + '!=' + r2;
      end;
    EX_New:
      begin
        r1 := ReadToken(OuterOperatorPrecedence); // outer
        r2 := ReadToken(OuterOperatorPrecedence); // name
        r3 := ReadToken(OuterOperatorPrecedence); // flags
        r4 := ReadToken(OuterOperatorPrecedence); // class
        if r2 <> '' then r2 := ',' + r2;
        if r3 <> '' then r3 := ',' + r3;
        if r4 <> '' then r4 := ',' + r4;
        result := 'new (' + r1 + r2 + r3 + r4 + ')';
      end;
    EX_LabelTable:
      begin
        // follows an EX_Stop, and it will be aligned to 4 (filled with EX_Nothing)
        repeat
          i1 := FOwner.read_idx(buffer); // label name
          inc(position_icode, 4);
          i2 := FOwner.read_int(buffer); // iCode
          inc(position_icode, 4);
          if (lowercase(FOwner.Names[i1]) <> 'none') then
            begin
              setlength(FLabelTable, length(FLabelTable) + 1);
              FLabelTable[high(FLabelTable)].name := i1;
              FLabelTable[high(FLabelTable)].icode := i2;
            end;
        until (lowercase(FOwner.Names[i1]) = 'none');
      end;
    EX_VirtualFunction:
      begin
        result := FOwner.Names[FOwner.read_idx(buffer)] + '(';
        inc(position_icode, 4);
        read_parameters(result);
        need_semicolon := true;
      end;
    EX_GlobalFunction:
      begin
        result := 'Global.' + FOwner.Names[FOwner.read_idx(buffer)] + '(';
        inc(position_icode, 4);
        read_parameters(result);
        need_semicolon := true;
      end;
    EX_FinalFunction:
      begin
        i1 := FOwner.read_idx(buffer);
        if (lowercase(FOwner.GetObjectPath(1, i1)) = lowercase(UTObjectName)) and
          (lowercase(FOwner.GetObjectPath(2, i1)) <> lowercase(FOwner.GetObjectPath(2, FExportedIndex))) then
          begin                         // If it is the same name as self but different owner
            // TODO : may fail if owner is an State ?
            // get called function owner
            r1 := FOwner.GetObjectPath(2, i1);
            i2 := pos('.', r1);
            r1 := copy(r1, 1, i2 - 1);
            // get owner from current function owner
            r2 := FOwner.GetObjectPath(2, FSuperField);
            i2 := pos('.', r2);
            r2 := copy(r2, 1, i2 - 1);
            // compare
            if lowercase(r1) = lowercase(r2) then
              result := 'Super.'
            else
              result := 'Super(' + r1 + ').';
          end
        else
          result := '';
        result := result + FOwner.GetObjectPath(1, i1) + '(';
        inc(position_icode, 4);
        read_parameters(result);
        need_semicolon := true;
      end
  else
    begin
      if b < EX_ExtendedNative then
        begin
          msg := format(rsUnknownOpcode, [b]);
          windows.messagebox(0, pchar(msg), pchar(rsWarning), mb_ok);
          result := format('UnknownOpcode0x%-2.2x(', [b]) + ReadToken(OuterOperatorPrecedence) + ')'; // we suppose it has function/conversion format
        end
      else                              // other native function
        begin
          if (b and $F0) = EX_ExtendedNative then // high native flag
            begin
              i1 := ((b - EX_ExtendedNative) shl 8) + FOwner.read_byte(buffer);
              inc(position_icode, 1);
            end
          else
            i1 := b;
          if i1 < EX_FirstNative then raise exception.create(format(rsInvalidNativeIndex, [i1]));
          i3 := -1;
          for i2 := 0 to high(NativeFunctions) do
            if NativeFunctions[i2].index = i1 then
              begin
                i3 := i2;
                break;
              end;
          if i3 <> -1 then
            begin
              if lowercase(NativeFunctions[i3].Name) = lowercase(UTObjectName) then
                begin                   // If it is the same name as self. Could also be recursion.
                  result := 'Super.';
                end
              else
                result := '';
              case NativeFunctions[i3].Format of
                nffFunction:
                  begin
                    result := result + NativeFunctions[i3].Name + '(';
                    read_parameters(result);
                  end;
                nffPreOperator:
                  begin
                    result := result + NativeFunctions[i3].Name + ReadToken(OuterOperatorPrecedence);
                    ReadToken(OuterOperatorPrecedence); // end params
                    if indent_level > 1 then result := ' ' + result;
                  end;
                nffPostOperator:
                  begin
                    result := ReadToken(OuterOperatorPrecedence) + result + NativeFunctions[i3].Name;
                    ReadToken(OuterOperatorPrecedence); // end params
                    if indent_level > 1 then result := result + ' ';
                  end;
                nffOperator:
                  begin
                    r1 := ReadToken(NativeFunctions[i3].OperatorPrecedence);
                    r2 := ReadToken(NativeFunctions[i3].OperatorPrecedence);
                    ReadToken(OuterOperatorPrecedence); // end params
                    result := r1 + ' ' + result + NativeFunctions[i3].Name + ' ' + r2;
                    if indent_level > 1 then
                      if NativeFunctions[i3].OperatorPrecedence > OuterOperatorPrecedence then
                        result := '(' + result + ')';
                  end;
              end;
            end
          else
            begin
              result := result + format('UnknownFunction%d(', [i1]);
              read_parameters(result);
            end;
          need_semicolon := true;
        end;
    end;
  end;
  dec(indent_level);
end;

function TUTObjectClassStruct.ReadStatement: string;
var
  a, previous_nest_level: integer;
  s: string;
begin
  if nest <> nil then
    previous_nest_level := nest.count
  else
    previous_nest_level := 0;
  result := ReadToken;
  // remove backspaces (when it is an Else)
  s := indent_chars;
  while copy(result, 1, 1) = #8 do
    begin
      delete(s, length(s), 1);
      delete(result, 1, 1);
    end;
  // add indent and semicolon
  if (result <> '') then
    begin
      result := FOwner.indenttext(s, result);
      if need_semicolon then result := result + ';';
    end;
  // change indent
  if nest <> nil then
    begin
      // add indenting and block open braces
      a := previous_nest_level;
      while a < nest.Count do
        begin
          if result <> '' then result := result + #13#10;
          result := result + indent_chars + '{';
          indent_chars := indent_chars + #9;
          inc(a);
        end;
      // decrease nest levels
      while (endnestlist.count > 0) and
        (integer(endnestlist.objects[endnestlist.count - 1]) and $FFFF = position_icode) do
        begin
          nest.delete(nest.count - 1);
          endnestlist.delete(endnestlist.count - 1);
        end;
      // add unindenting and block close braces
      a := previous_nest_level;
      while a > nest.Count do
        begin
          if result <> '' then result := result + #13#10;
          delete(indent_chars, length(indent_chars), 1);
          result := result + indent_chars + '}';
          dec(a);
        end;
    end;
end;

function TUTObjectClassStruct.ReadStatements(beautify: boolean): string;
var
  statement: string;
  code: tstringlist;
  a, i: integer;
begin
  check_initialized;
  result := '';
  // read code statements
  code := tstringlist.create;
  if beautify then
    begin
      nest := tlist.create;
      endnestlist := tstringlist.create;
    end;
  indent_chars := '';
  while position_icode < FScriptSize do
    begin
      need_semicolon := false;
      labellist := '';
      i := position_icode;
      statement := ReadStatement;
      if statement <> '' then code.addobject(statement, pointer(i));
    end;
  if beautify then
    begin
      freeandnil(nest);
      freeandnil(endnestlist);
    end;
  // add labels and indent
  a := 0;
  while a < code.count do
    begin
      if (integer(code.objects[a]) and $FFFF) >= 0 then
        begin
          // normal labels
          i := JumpList.indexof(pointer(integer(code.objects[a]) and $FFFF));
          if i >= 0 then
            begin
              code.InsertObject(a, format('JL%-4.4x:', [integer(code.objects[a]) and $FFFF]), pointer(-1));
              inc(a);
              jumplist.delete(i);
            end;
          // state labels
          for i := 0 to high(FLabelTable) do
            if FLabelTable[i].iCode = (integer(code.objects[a]) and $FFFF) then
              begin
                code.InsertObject(a, FOwner.Names[FLabelTable[i].name] + ':', pointer(-1));
                inc(a);
              end;
        end;
      code[a] := FOwner.indenttext(#9, code[a]);
      inc(a);
    end;
  result := code.text;
  code.free;
end;

procedure TUTObjectClassStruct.SkipStatements;
begin
  check_initialized;
  while position_icode < FScriptSize do
    ReadToken;
  setlength(FLabelTable, 0);
end;

function TUTObjectClassStruct.GetChildren: integer;
begin
  check_initialized;
  result := FChildren;
end;

function TUTObjectClassStruct.GetFriendlyName: string;
begin
  check_initialized;
  result := FFriendlyName;
end;

function TUTObjectClassStruct.GetLine: integer;
begin
  check_initialized;
  result := FLine;
end;

function TUTObjectClassStruct.GetScriptSize: integer;
begin
  check_initialized;
  result := FScriptSize;
end;

function TUTObjectClassStruct.GetScriptText: integer;
begin
  check_initialized;
  result := FScriptText;
end;

function TUTObjectClassStruct.GetTextPos: integer;
begin
  check_initialized;
  result := FTextPos;
end;

{ TUTObjectClassFunction }

procedure TUTObjectClassFunction.InitializeObject;
begin
  inherited;
  FiNative := 0;
  FRepOffset := $FFFF;
  FOperatorPrecedence := 0;
  FFunctionFlags := 0;
end;

procedure TUTObjectClassFunction.InterpretObject;
begin
  inherited;
  try
    if FOwner.Version <= 63 then FOwner.read_idx(buffer); // ParmsSize
    FiNative := FOwner.read_word(buffer);
    if FOwner.Version <= 63 then FOwner.read_idx(buffer); // NumParms
    FOperatorPrecedence := FOwner.read_byte(buffer);
    if FOwner.Version <= 63 then FOwner.read_idx(buffer); // ReturnValueOffset
    FFunctionFlags := FOwner.read_int(buffer);
    if (FunctionFlags and FUNC_Net) <> 0 then FRepOffset := FOwner.read_word(buffer);
  except
  end;
  try
    Decompile(false);
  except
  end;
end;

procedure TUTObjectClassFunction.ReleaseObject;
begin
  FiNative := 0;
  FRepOffset := $FFFF;
  FOperatorPrecedence := 0;
  FFunctionFlags := 0;
  inherited;
end;

function TUTObjectClassFunction.Decompile(beautify: boolean): string;
const
  indentation = #9;
var
  code, function_name, parameters, result_type,
    locals, function_flags: string;
  c: integer;
  prop: TUTObjectClassProperty;
  result_str: tstringlist;
begin
  check_initialized;
  buffer.seek(FScriptStart, soFromBeginning);
  result := '';
  code := '';
  try
    function_name := FFriendlyName;
    result_type := '';
    parameters := '';
    locals := '';
    function_flags := '';
    if (FFunctionFlags and FUNC_Native) <> 0 then
      begin
        if FiNative > 0 then
          function_flags := function_flags + format('native(%d) ', [FiNative])
        else
          function_flags := function_flags + 'native ';
      end;
    if (FFunctionFlags and FUNC_Static) <> 0 then function_flags := function_flags + 'static ';
    if (FFunctionFlags and FUNC_Final) <> 0 then function_flags := function_flags + 'final ';
    //if (FFunctionFlags and FUNC_Defined)<>0 then function_flags:=function_flags+'defined ';
    if (FFunctionFlags and FUNC_Iterator) <> 0 then function_flags := function_flags + 'iterator ';
    if (FFunctionFlags and FUNC_Latent) <> 0 then function_flags := function_flags + 'latent ';
    if (FFunctionFlags and FUNC_Singular) <> 0 then function_flags := function_flags + 'singular ';
    //if (FFunctionFlags and FUNC_Net) <> 0 then function_flags := function_flags + 'net ';
    //if (FFunctionFlags and FUNC_NetReliable) <> 0 then function_flags := function_flags + 'netreliable ';
    if (FFunctionFlags and FUNC_Simulated) <> 0 then function_flags := function_flags + 'simulated ';
    if (FFunctionFlags and FUNC_Exec) <> 0 then function_flags := function_flags + 'exec ';
    if (FFunctionFlags and FUNC_Event) <> 0 then function_flags := function_flags + 'event ';
    if (FFunctionFlags and FUNC_NoExport) <> 0 then function_flags := function_flags + 'noexport ';
    if (FFunctionFlags and FUNC_Const) <> 0 then function_flags := function_flags + 'const ';
    if (FFunctionFlags and FUNC_Invariant) <> 0 then function_flags := function_flags + 'invariant ';
    if (FFunctionFlags and FUNC_Operator) <> 0 then
      begin
        if (FFunctionFlags and FUNC_PreOperator) <> 0 then
          function_flags := function_flags + 'preoperator '
        else if FOperatorPrecedence = 0 then
          function_flags := function_flags + 'postoperator '
        else
          function_flags := function_flags + format('operator(%d) ', [FOperatorPrecedence]);
      end;
    if ((FFunctionFlags and FUNC_Operator) = 0) and
      ((FFunctionFlags and FUNC_Event) = 0) then
      function_flags := function_flags + 'function ';

    c := FChildren;
    while c <> 0 do
      begin
        dec(c);
        if not (FOwner.Exported[c].UTObject is TUTObjectClassProperty) then break;
        prop := TUTObjectClassProperty(FOwner.Exported[c].UTObject);
        try
          try
            prop.ReadObject;
            if (prop.PropertyFlags and CPF_Parm) <> 0 then
              begin                     // function parameter, could also be a return value
                if (prop.PropertyFlags and CPF_ReturnParm) <> 0 then
                  begin
                    result_type := prop.GetFlags(UTobjectname) + prop.TypeName;
                  end
                else
                  begin
                    parameters := parameters + prop.GetDeclaration('', UTobjectname) + ', ';
                  end;
              end
            else
              begin                     // local variable
                locals := locals + indentation + prop.GetDeclaration('local', UTobjectname) + ';'#13#10;
              end;
            c := prop.Next;
          finally
            prop.ReleaseObject;
          end;
        except
          c := 0;
        end;
      end;
    if parameters <> '' then delete(parameters, length(parameters) - 1, 2);
    if result_type <> '' then result_type := result_type + ' ';
    if locals <> '' then locals := locals + #13#10;
    code := format('%s%s%s (%s)', [function_flags, result_type, function_name, parameters]);
    if (FFunctionFlags and FUNC_Defined) <> 0 then
      begin
        position_icode := 0;
        code := code + format(#13#10'{'#13#10'%s', [locals]);
        code := code + ReadStatements(beautify);
        result_str := tstringlist.create;
        result_str.text := code;
        result_str.delete(result_str.count - 1); // remove last "return;" line
        code := result_str.text;
        code := code + '}'#13#10;
      end
    else
      code := code + ';'#13#10;         // native functions do not have code
  except
  end;
  result := code;
end;

function TUTObjectClassFunction.GetFunctionFlags: longword;
begin
  check_initialized;
  result := FFunctionFlags;
end;

function TUTObjectClassFunction.GetiNative: integer;
begin
  check_initialized;
  result := FiNative;
end;

function TUTObjectClassFunction.GetOperatorPrecedence: integer;
begin
  check_initialized;
  result := FOperatorPrecedence;
end;

function TUTObjectClassFunction.GetRepOffset: integer;
begin
  check_initialized;
  result := FRepOffset;
end;

{ TUTObjectClassIntProperty }

function TUTObjectClassIntProperty.TypeName: string;
begin
  check_initialized;
  result := 'int';
end;

{ TUTObjectClassBoolProperty }

function TUTObjectClassBoolProperty.TypeName: string;
begin
  check_initialized;
  result := 'bool';
end;

{ TUTObjectClassFloatProperty }

function TUTObjectClassFloatProperty.TypeName: string;
begin
  check_initialized;
  result := 'float';
end;

{ TUTObjectClassNameProperty }

function TUTObjectClassNameProperty.TypeName: string;
begin
  check_initialized;
  result := 'name';
end;

{ TUTObjectClassStrProperty }

function TUTObjectClassStrProperty.TypeName: string;
begin
  check_initialized;
  result := 'string';
end;

{ TUTObjectClassStringProperty }

function TUTObjectClassStringProperty.TypeName: string;
begin
  check_initialized;
  result := 'string';                   // old fixed-length string type
end;

{ TUTObjectClassState }

function TUTObjectClassState.Decompile(beautify: boolean): string;
var
  c: integer;
  child: TUTObject;
  statements, header, var_block, ignored_functions: string;
  result_str: tstringlist;
begin
  check_initialized;
  buffer.seek(FScriptStart, soFromBeginning);
  result := '';
  c := FChildren;
  var_block := '';
  ignored_functions := '';
  while c <> 0 do
    begin
      child := FOwner.Exported[c - 1].UTObject;
      try
        child.ReadObject;
        if child is TUTObjectClassField then
          begin
            if child is TUTObjectClassFunction then
              begin
                if (TUTObjectClassFunction(child).FunctionFlags and FUNC_Defined) <> 0 then
                  result := TUTObjectClassFunction(child).Decompile(beautify) + #13#10 + result
                else
                  ignored_functions := ignored_functions + ', ' + TUTObjectClassFunction(child).UTObjectName;
              end
            else if child is TUTObjectClassProperty then
              begin                     // it is a variable (may not exist in states?)
                var_block := var_block + TUTObjectClassProperty(child).GetDeclaration('local', UTobjectname) + ';'#13#10;
              end
            else if child is TUTObjectClassConst then
              begin                     // it is a const
                var_block := var_block + TUTObjectClassConst(child).GetDeclaration + ';'#13#10;
              end
            else if child is TUTObjectClassEnum then
              begin                     // it is an enum
                var_block := var_block + TUTObjectClassEnum(child).GetDeclaration + ';'#13#10#13#10;
              end
            else if (child is TUTObjectClassStruct) and
              not ((child is TUTObjectClassFunction) or (child is TUTObjectClassState)) then
              begin                     // it is an struct
                var_block := var_block + TUTObjectClassStruct(child).GetDeclaration + ';'#13#10#13#10;
              end;
            c := TUTObjectClassField(child).Next;
          end
        else
          c := 0;
      finally
        child.ReleaseObject;
      end;
    end;

  if ignored_functions <> '' then
    begin
      delete(ignored_functions, 1, 1);
      ignored_functions := 'ignores ' + ignored_functions + ';'#13#10#13#10;
    end;
  if var_block <> '' then var_block := var_block + #13#10;

  result := var_block + ignored_functions + result;
  position_icode := 0;
  header := '';
  if (FStateFlags and STATE_Auto) <> 0 then header := header + 'auto ';
  if (FStateFlags and STATE_Simulated) <> 0 then header := header + 'simulated ';
  if (FStateFlags and STATE_Editable) <> 0 then
    header := header + 'state() '
  else
    header := header + 'state ';
  header := header + UTObjectName;
  if SuperField <> 0 then header := header + ' expands ' + FOwner.GetObjectPath(1, SuperField);
  result := header + #13#10'{'#13#10 + FOwner.IndentText(#9, result);
  statements := ReadStatements(beautify);
  if statements <> '' then result := result + Statements;
  result_str := tstringlist.create;
  result_str.text := result;
  result_str.delete(result_str.count - 1); // remove last "stop;" line
  result := result_str.text;
  result := result + '}'#13#10;
end;

function TUTObjectClassState.GetIgnoreMask: int64;
begin
  check_initialized;
  result := FIgnoreMask;
end;

function TUTObjectClassState.GetLabelTableOffset: word;
begin
  check_initialized;
  result := FLabelTableOffset;
end;

function TUTObjectClassState.GetProbeMask: int64;
begin
  check_initialized;
  result := FProbeMask;
end;

function TUTObjectClassState.GetStateFlags: longword;
begin
  check_initialized;
  result := FStateFlags;
end;

procedure TUTObjectClassState.InitializeObject;
begin
  inherited;
  FProbeMask := 0;
  FIgnoreMask := 0;
  FStateFlags := 0;
  FLabelTableOffset := 0;
end;

procedure TUTObjectClassState.InterpretObject;
begin
  inherited;
  FProbeMask := FOwner.read_qword(buffer);
  FIgnoreMask := FOwner.read_qword(buffer);
  FLabelTableOffset := FOwner.read_word(buffer);
  FStateFlags := FOwner.read_int(buffer);
end;

procedure TUTObjectClassState.ReleaseObject;
begin
  FProbeMask := 0;
  FIgnoreMask := 0;
  FStateFlags := 0;
  FLabelTableOffset := 0;
  inherited;
end;

{ TUTObjectClassClass }

function TUTObjectClassClass.Decompile(beautify: boolean): string;
type
  TUTGUID = record
    case byte of
      0: (guid: TGUID);
      1: (A, B, C, D: integer);
      2: (n: int64);
  end;
var
  a, b, c, o, offs, is_reliable: integer;
  child: TUTObject;
  var_block, func_block, replication_block, within, flags, rep_condition, n, rep_type, bars: string;
  g: TUTGUID;
  replication_list: tstringlist;
begin
  check_initialized;
  buffer.seek(FScriptStart, soFromBeginning);
  result := '';
  replication_list := tstringlist.create;
  // TODO : add "import" clause: import (enum|package) <name> [from <package>]

  // read functions & states
  func_block := '';
  c := FChildren;
  while c <> 0 do
    begin
      child := FOwner.Exported[c - 1].UTObject;
      try
        child.ReadObject;
        if child is TUTObjectClassField then
          begin
            if child is TUTObjectClassState then
              func_block := TUTObjectClassState(child).Decompile(beautify) + #13#10 + func_block
            else if child is TUTObjectClassFunction then
              begin
                func_block := TUTObjectClassFunction(child).Decompile(beautify) + #13#10 + func_block;
                if ((TUTObjectClassFunction(child).functionflags and FUNC_Net) <> 0) and
                  (TUTObjectClassFunction(child).ReplicationOffset < $FFFF) then
                  if (TUTObjectClassFunction(child).functionflags and FUNC_NetReliable) <> 0 then
                    replication_list.addobject(TUTObjectClassFunction(child).UTobjectname, pointer($C0000000 or longword(TUTObjectClassFunction(child).ReplicationOffset)))
                  else
                    replication_list.addobject(TUTObjectClassFunction(child).UTobjectname, pointer($80000000 or longword(TUTObjectClassFunction(child).ReplicationOffset)));
              end;
            c := TUTObjectClassField(child).Next;
          end
        else
          c := 0;
      finally
        child.ReleaseObject;
      end;
    end;

  // read variable declarations
  var_block := '';
  c := FChildren;
  while c <> 0 do
    begin
      child := FOwner.Exported[c - 1].UTObject;
      try
        child.ReadObject;
        if child is TUTObjectClassField then
          begin
            if child is TUTObjectClassProperty then
              begin                     // it is a global variable
                var_block := var_block + TUTObjectClassProperty(child).GetDeclaration('var', UTobjectname) + ';'#13#10;
                if (TUTObjectClassProperty(child).propertyflags and CPF_Net) <> 0 then
                  begin
                    // TODO : get actual reliable flag
                    // try to find whether it is reliable based on functions with same RepOffset
                    is_reliable := $00000000; // default is unknown
                    for a := 0 to replication_list.count - 1 do
                      if (integer(replication_list.objects[a]) and $3FFFFFFF) = TUTObjectClassProperty(child).ReplicationOffset then
                        begin
                          is_reliable := integer(replication_list.objects[a]) and $C0000000;
                          break;
                        end;
                    replication_list.addobject(TUTObjectClassProperty(child).UTobjectname, pointer(is_Reliable or TUTObjectClassProperty(child).ReplicationOffset));
                  end;
              end
            else if child is TUTObjectClassConst then
              begin                     // it is a const
                var_block := var_block + TUTObjectClassConst(child).GetDeclaration + ';'#13#10;
              end
            else if child is TUTObjectClassEnum then
              begin                     // it is an enum
                var_block := var_block + TUTObjectClassEnum(child).GetDeclaration + ';'#13#10#13#10;
              end
            else if (child is TUTObjectClassStruct) and
              not ((child is TUTObjectClassFunction) or (child is TUTObjectClassState)) then
              begin                     // it is an struct
                var_block := var_block + TUTObjectClassStruct(child).GetDeclaration + ';'#13#10#13#10;
              end;
            c := TUTObjectClassField(child).Next;
          end
        else
          c := 0;
      finally
        child.ReleaseObject;
      end;
    end;

  if var_block <> '' then var_block := var_block + #13#10;

  // construct the replication block
  replication_block := '';
  while replication_list.count > 0 do
    begin
      o := integer(replication_list.objects[0]);
      offs := o and $3FFFFFFF;
      buffer.seek(FScriptStart, soFromBeginning);
      position_icode := 0;
      while position_icode <= offs do
        rep_condition := ReadToken;
      case (o and $C0000000) of
        $C0000000: rep_type := 'reliable';
        $80000000: rep_type := 'unreliable';
      else
        rep_type := 'un?reliable';
      end;
      n := '';
      b := replication_list.count - 1;
      while b > 0 do
        begin
          if replication_list.objects[b] = pointer(o) then
            begin
              n := replication_list[b] + ',' + n;
              replication_list.delete(b);
            end;
          dec(b);
        end;
      if n <> '' then n := ',' + copy(n, 1, length(n) - 1);
      n := replication_list[0] + n;
      replication_list.delete(0);
      replication_block := replication_block + #9 + rep_type + ' if ( ' + rep_condition + ' )'#13#10 + #9#9 + n + ';'#13#10;
    end;
  if replication_block <> '' then replication_block := 'replication'#13#10'{'#13#10 + replication_block + '}'#13#10#13#10;

  flags := #13#10;
  if (UTFlags and RF_Native) <> 0 then flags := flags + #9'native'#13#10;
  if (ClassFlags and CLASS_NoExport) <> 0 then flags := flags + #9'noexport'#13#10;
  if (ClassFlags and CLASS_NativeReplication) <> 0 then flags := flags + #9'nativereplication'#13#10;
  if (ClassFlags and CLASS_PerObjectConfig) <> 0 then flags := flags + #9'perobjectconfig'#13#10;
  if (ClassFlags and CLASS_NoUserCreate) <> 0 then flags := flags + #9'nousercreate'#13#10;
  if (ClassFlags and CLASS_Abstract) <> 0 then flags := flags + #9'abstract'#13#10;
  g.guid := FClassGuid;
  if g.n <> 0 then flags := flags + format(#9'guid(%d,%d,%d,%d)'#13#10, [g.A, g.B, g.C, g.D]);
  if (ClassFlags and CLASS_Transient) <> 0 then flags := flags + #9'transient'#13#10;
  //if (ClassFlags and CLASS_Localized)<>0 then flags:=flags+' localized'; // no longer required
  if ((ClassFlags and CLASS_Config) <> 0) and
    (lowercase(FOwner.Names[FClassConfigName]) <> 'system') then
    flags := flags + format(#9'config(%s)'#13#10, [FOwner.Names[FClassConfigName]]);
  if (ClassFlags and CLASS_SafeReplace) <> 0 then flags := flags + #9'safereplace'#13#10;
  if flags <> '' then delete(flags, length(flags) - 1, 2);

  within := '';
  if (FClassWithin <> 0) and (lowercase(FOwner.GetObjectPath(1, FClassWithin)) <> 'object') then
    within := ' within ' + FOwner.GetObjectPath(1, FClassWithin);

  setlength(bars, 80);
  fillchar(bars[1], 80, ord('='));
  result := '//' + bars + #13#10 + '// ' + UTObjectName + '.'#13#10 + '//' + bars + #13#10;
  result := result + 'class ' + UTObjectName;
  if SuperField <> 0 then result := result + ' expands ' + FOwner.GetObjectPath(1, SuperField);
  result := result + within + flags + ';'#13#10#13#10;
  result := result + var_block + replication_block + func_block;

  replication_list.free;
end;

function TUTObjectClassClass.DependencyCount: integer;
begin
  check_initialized;
  result := length(FDependencies);
end;

function TUTObjectClassClass.GetClassConfigName: integer;
begin
  check_initialized;
  result := FClassConfigName;
end;

function TUTObjectClassClass.GetClassFlags: longword;
begin
  check_initialized;
  result := FClassFlags;
end;

function TUTObjectClassClass.GetClassGuid: TGuid;
begin
  check_initialized;
  result := FClassGuid;
end;

function TUTObjectClassClass.GetClassWithin: integer;
begin
  check_initialized;
  result := FClassWithin;
end;

function TUTObjectClassClass.GetDependencies(i: integer): TUTDependency;
begin
  check_initialized;
  result := FDependencies[i];
end;

function TUTObjectClassClass.GetPackageImports(i: integer): integer;
begin
  check_initialized;
  result := FPackageImports[i];
end;

procedure TUTObjectClassClass.InitializeObject;
begin
  inherited;
  FClassFlags := 0;
  fillchar(FClassGuid, sizeof(TGuid), 0);
  setlength(FDependencies, 0);
  setlength(FPackageImports, 0);
  FClassWithin := 0;
  FClassConfigName := 0;
end;

procedure TUTObjectClassClass.InterpretObject;
var
  a: integer;
begin
  inherited;
  if FOwner.Version <= 61 then FOwner.read_int(buffer); // OldClassRecordSize
  FClassFlags := FOwner.read_int(buffer);
  FClassGuid := FOwner.read_guid(buffer);
  setlength(FDependencies, FOwner.read_idx(buffer));
  for a := 0 to high(FDependencies) do
    begin
      FDependencies[a]._Class := FOwner.read_idx(buffer);
      FDependencies[a].Deep := FOwner.read_int(buffer);
      FDependencies[a].ScriptTextCRC := FOwner.read_int(buffer);
    end;
  setlength(FPackageImports, FOwner.read_idx(buffer));
  for a := 0 to high(FPackageImports) do
    FPackageImports[a] := FOwner.read_idx(buffer);
  if FOwner.Version >= 62 then
    begin
      FClassWithin := FOwner.read_idx(buffer); // object
      FClassConfigName := FOwner.read_idx(buffer); // name
    end;
  ReadProperties;
end;

function TUTObjectClassClass.PackageImportsCount: integer;
begin
  check_initialized;
  result := length(FPackageImports);
end;

procedure TUTObjectClassClass.ReleaseObject;
begin
  FClassFlags := 0;
  fillchar(FClassGuid, sizeof(TGuid), 0);
  setlength(FDependencies, 0);
  setlength(FPackageImports, 0);
  FClassWithin := 0;
  FClassConfigName := 0;
  inherited;
end;

{ TUTObjectClassSkeletalMesh }

function TUTObjectClassSkeletalMesh.GetBoneWeight(
  i: integer): TUT_Struct_BoneInfluence;
begin
  check_initialized;
  result := FBoneWeights[i];
end;

function TUTObjectClassSkeletalMesh.GetBoneWeightCount: integer;
begin
  check_initialized;
  result := length(FBoneWeights);
end;

function TUTObjectClassSkeletalMesh.GetBoneWeightIdx(
  i: integer): TUT_Struct_BoneInfIndex;
begin
  check_initialized;
  result := FBoneWeightIdx[i];
end;

function TUTObjectClassSkeletalMesh.GetBoneWeightIdxCount: integer;
begin
  check_initialized;
  result := length(FBoneWeightIdx);
end;

function TUTObjectClassSkeletalMesh.GetExtWedge(
  i: integer): TUT_Struct_MeshExtWedge;
begin
  check_initialized;
  result := FExtWedges[i];
end;

function TUTObjectClassSkeletalMesh.GetExtWedgeCount: integer;
begin
  check_initialized;
  result := length(FExtWedges);
end;

function TUTObjectClassSkeletalMesh.GetLocalPoint(
  i: integer): TUT_Struct_Vector;
begin
  check_initialized;
  result := FLocalPoints[i];
end;

function TUTObjectClassSkeletalMesh.GetLocalPointCount: integer;
begin
  check_initialized;
  result := length(FLocalPoints);
end;

function TUTObjectClassSkeletalMesh.GetPoint(
  i: integer): TUT_Struct_Vector;
begin
  check_initialized;
  result := FPoints[i];
end;

function TUTObjectClassSkeletalMesh.GetPointCount: integer;
begin
  check_initialized;
  result := length(FPoints);
end;

function TUTObjectClassSkeletalMesh.GetRefSkeleton(
  i: integer): TUT_Struct_MeshBone;
begin
  check_initialized;
  result := FRefSkeleton[i];
end;

function TUTObjectClassSkeletalMesh.GetRefSkeletonCount: integer;
begin
  check_initialized;
  result := length(FRefSkeleton);
end;

procedure TUTObjectClassSkeletalMesh.InitializeObject;
begin
  inherited;
  setlength(FExtWedges, 0);
  setlength(FPoints, 0);
  setlength(FRefSkeleton, 0);
  setlength(FBoneWeightIdx, 0);
  setlength(FBoneWeights, 0);
  setlength(FLocalPoints, 0);
  FSkeletalDepth := 0;
  FDefaultAnimation := 0;
  FWeaponBoneIndex := 0;
  fillchar(FWeaponAdjust, 0, sizeof(TUT_Struct_Coords));
end;

procedure TUTObjectClassSkeletalMesh.InterpretObject;
var
  size, a: integer;
begin
  inherited;
  size := FOwner.read_idx(buffer);
  setlength(FExtWedges, size);
  for a := 0 to size - 1 do
    begin
      FExtWedges[a].iVertex := FOwner.read_word(buffer);
      FExtWedges[a].Flags := FOwner.read_word(buffer);
      FExtWedges[a].TexUV.U := FOwner.read_float(buffer);
      FExtWedges[a].TexUV.V := FOwner.read_float(buffer);
    end;
  size := FOwner.read_idx(buffer);
  setlength(FPoints, size);
  for a := 0 to size - 1 do
    begin
      FPoints[a].x := FOwner.read_float(buffer);
      FPoints[a].y := FOwner.read_float(buffer);
      FPoints[a].z := FOwner.read_float(buffer);
    end;
  size := FOwner.read_idx(buffer);
  setlength(FRefSkeleton, size);
  for a := 0 to size - 1 do
    begin
      FRefSkeleton[a].Name := FOwner.read_idx(buffer);
      FRefSkeleton[a].Flags := FOwner.read_int(buffer);
      FRefSkeleton[a].BonePos.Orientation.X := FOwner.read_float(buffer);
      FRefSkeleton[a].BonePos.Orientation.Y := FOwner.read_float(buffer);
      FRefSkeleton[a].BonePos.Orientation.Z := FOwner.read_float(buffer);
      FRefSkeleton[a].BonePos.Orientation.W := FOwner.read_float(buffer);
      FRefSkeleton[a].BonePos.Position.X := FOwner.read_float(buffer);
      FRefSkeleton[a].BonePos.Position.Y := FOwner.read_float(buffer);
      FRefSkeleton[a].BonePos.Position.Z := FOwner.read_float(buffer);
      FRefSkeleton[a].BonePos.Length := FOwner.read_float(buffer);
      FRefSkeleton[a].BonePos.XSize := FOwner.read_float(buffer);
      FRefSkeleton[a].BonePos.YSize := FOwner.read_float(buffer); // bug in UT? Y=X ?
      FRefSkeleton[a].BonePos.ZSize := FOwner.read_float(buffer);
      FRefSkeleton[a].NumChildren := FOwner.read_int(buffer);
      FRefSkeleton[a].ParentIndex := FOwner.read_int(buffer);
    end;
  size := FOwner.read_idx(buffer);
  setlength(FBoneWeightIdx, size);
  for a := 0 to size - 1 do
    begin
      FBoneWeightIdx[a].WeightIndex := FOwner.read_word(buffer);
      FBoneWeightIdx[a].Number := FOwner.read_word(buffer);
      FBoneWeightIdx[a].DetailA := FOwner.read_word(buffer);
      FBoneWeightIdx[a].DetailB := FOwner.read_word(buffer);
    end;
  size := FOwner.read_idx(buffer);
  setlength(FBoneWeights, size);
  for a := 0 to size - 1 do
    begin
      FBoneWeights[a].PointIndex := FOwner.read_word(buffer);
      FBoneWeights[a].BoneWeight := FOwner.read_word(buffer);
    end;
  size := FOwner.read_idx(buffer);
  setlength(FLocalPoints, size);
  for a := 0 to size - 1 do
    begin
      FLocalPoints[a].X := FOwner.read_float(buffer);
      FLocalPoints[a].Y := FOwner.read_float(buffer);
      FLocalPoints[a].Z := FOwner.read_float(buffer);
    end;
  FSkeletalDepth := FOwner.read_int(buffer);
  FDefaultAnimation := FOwner.read_idx(buffer);
  FWeaponBoneIndex := FOwner.read_int(buffer);
  FWeaponAdjust.Origin.X := FOwner.read_float(buffer);
  FWeaponAdjust.Origin.Y := FOwner.read_float(buffer);
  FWeaponAdjust.Origin.Z := FOwner.read_float(buffer);
  FWeaponAdjust.XAxis.X := FOwner.read_float(buffer);
  FWeaponAdjust.XAxis.Y := FOwner.read_float(buffer);
  FWeaponAdjust.XAxis.Z := FOwner.read_float(buffer);
  FWeaponAdjust.YAxis.X := FOwner.read_float(buffer);
  FWeaponAdjust.YAxis.Y := FOwner.read_float(buffer);
  FWeaponAdjust.YAxis.Z := FOwner.read_float(buffer);
  FWeaponAdjust.ZAxis.X := FOwner.read_float(buffer);
  FWeaponAdjust.ZAxis.Y := FOwner.read_float(buffer);
  FWeaponAdjust.ZAxis.Z := FOwner.read_float(buffer);
end;

procedure TUTObjectClassSkeletalMesh.ReleaseObject;
begin
  setlength(FExtWedges, 0);
  setlength(FPoints, 0);
  setlength(FRefSkeleton, 0);
  setlength(FBoneWeightIdx, 0);
  setlength(FBoneWeights, 0);
  setlength(FLocalPoints, 0);
  FSkeletalDepth := 0;
  FDefaultAnimation := 0;
  FWeaponBoneIndex := 0;
  fillchar(FWeaponAdjust, 0, sizeof(TUT_Struct_Coords));
  inherited;
end;

procedure TUTObjectClassSkeletalMesh.PrepareExporter(exporter: TUT_MeshExporter; frames: TIntegerArray);
const
  material_colors: array[0..5] of TColor = (clBlue, clRed, clLime, clYellow, clAqua, clSilver);
var
  matname: string;
  m: integer;
begin
  check_initialized;
  if length(frames) = 0 then
    begin
      setlength(frames, 1);
      frames[0] := 0;
    end;
  setlength(exporter.Materials, length(FMaterials));
  for m := 0 to high(FMaterials) do
    begin
      matname := 'SKIN' + inttostr(FMaterials[m].textureindex);
      if (FMaterials[m].flags and (PF_TwoSided or PF_Modulated)) = (PF_TwoSided or PF_Modulated) then
        matname := matname + '.MODU'    //LATED'
      else if (FMaterials[m].flags and (PF_TwoSided or PF_Translucent)) = (PF_TwoSided or PF_Translucent) then
        matname := matname + '.TRAN'    //SLUCENT'
      else if (FMaterials[m].flags and (PF_TwoSided or PF_Masked)) = (PF_TwoSided or PF_Masked) then
        matname := matname + '.MASK'    //ED'
      else if (FMaterials[m].flags and PF_TwoSided) = PF_TwoSided then
        matname := matname + '.TWOS'    //IDED'
      else if (FMaterials[m].flags and PF_NotSolid) = PF_NotSolid then
        matname := 'WEAPON';
      matname := copy(matname, 1, 10);
      exporter.Materials[m].name := matname;
      if m <= high(material_colors) then
        begin
          exporter.Materials[m].diffusecolor[0] := material_colors[m] and $FF;
          exporter.Materials[m].diffusecolor[1] := (material_colors[m] shr 8) and $FF;
          exporter.Materials[m].diffusecolor[2] := (material_colors[m] shr 16) and $FF;
        end
      else
        begin
          exporter.Materials[m].diffusecolor[0] := random(256);
          exporter.Materials[m].diffusecolor[1] := random(256);
          exporter.Materials[m].diffusecolor[2] := random(256);
        end;
    end;
  setlength(exporter.Vertices, length(FWedges) * length(frames));
  exporter.AnimationFrames := 1;
  for m := 0 to high(FWedges) do
    begin
      exporter.Vertices[m].x := FPoints[FSpecialVerts + FWedges[m].VertexIndex].X;
      exporter.Vertices[m].y := FPoints[FSpecialVerts + FWedges[m].VertexIndex].Y;
      exporter.Vertices[m].z := FPoints[FSpecialVerts + FWedges[m].VertexIndex].Z;
      exporter.Vertices[m].U := FWedges[m].U;
      exporter.Vertices[m].V := FWedges[m].V;
    end;
  setlength(exporter.Faces, length(FFaces));
  for m := 0 to high(FFaces) do
    begin
      exporter.Faces[m].VertexIndex1 := FFaces[m].WedgeIndex2; // corrected winding
      exporter.Faces[m].VertexIndex2 := FFaces[m].WedgeIndex1;
      exporter.Faces[m].VertexIndex3 := FFaces[m].WedgeIndex3;
      exporter.Faces[m].MaterialIndex := FFaces[m].MatIndex;
      exporter.Faces[m].Flags := FMaterials[FFaces[m].MatIndex].Flags;
    end;
end;

procedure TUTObjectClassSkeletalMesh.Save_3DS(filename: string;
  frame: integer; smoothing: TUT_3DStudioExporter_Smoothing; MirrorX: boolean);
var
  exporter: TUT_3DStudioExporter;
  frames: TIntegerArray;
begin
  check_initialized;
  exporter := TUT_3DStudioExporter.create;
  setlength(frames, 1);
  frames[0] := 0;
  PrepareExporter(exporter, frames);
  exporter.mirrorx := mirrorx;
  exporter.smoothing := smoothing;
  exporter.Save(filename);
  exporter.free;
end;

procedure TUTObjectClassSkeletalMesh.Save_Unreal3D(filename: string);
var
  exporter: TUT_Unreal3DExporter;
  frames: TIntegerArray;
begin
  check_initialized;
  exporter := TUT_Unreal3DExporter.create;
  setlength(frames, 1);
  frames[0] := 0;
  PrepareExporter(exporter, frames);
  exporter.CoordsDivisor := 4;
  exporter.Save(filename);
  exporter.free;
end;

procedure TUTObjectClassSkeletalMesh.Save_UnrealUC(filename: string);
var
  parent_class, basename, uc: string;
  k: char;
  a, script_idx: integer;
  ed2: TUTExportTableObjectData;
  id: TUTImportTableObjectData;
  str_uc: tfilestream;
begin
  check_initialized;
  a := FOwner.FindObject(utolExports, [utfwName, utfwClass, utfwPackage], '', UTobjectname, '');
  if a <> -1 then
    begin
      ed2 := FOwner.Exported[a];
      ed2.UTObject.ReadObject;
      parent_class := FOwner.GetObjectPath(1, TUTObjectClassClass(ed2.UTObject).SuperField);
      script_idx := TUTObjectClassClass(ed2.UTObject).ScriptText - 1;
      ed2.UTObject.ReleaseObject;
    end
  else
    begin
      parent_class := 'TournamentPlayer'; // do not localize these strings
      script_idx := FOwner.FindObject(utolExports, [utfwName, utfwClass, utfwPackage],
        UTobjectname, 'ScriptText', 'TextBuffer');
    end;
  if script_idx <> -1 then
    begin
      ed2 := FOwner.Exported[script_idx];
      ed2.UTObject.ReadObject;
      TUTObjectClassTextBuffer(ed2.UTObject).SaveToFile(filename);
      ed2.UTObject.ReleaseObject;
    end
  else
    begin
      k := DecimalSeparator;
      decimalSeparator := '.';
      basename := UTobjectname;
      // do not localize
      uc := '//============================================================================='#13#10;
      uc := uc + format('// %s.'#13#10, [basename]);
      uc := uc + '//============================================================================='#13#10;
      uc := uc + format('class %s extends %s;'#13#10#13#10, [basename, parent_class]);
      if DefaultAnimation > 0 then
        begin
          uc := uc + format('#exec ANIM IMPORT ANIM=%s ANIMFILE=MODELS\%s.PSA' {+' COMPRESS=x IMPORTSEQS=x MAXKEYS=x'} + #13#10, [FOwner.GetObjectPath(1, FDefaultAnimation), FOwner.GetObjectPath(1, FDefaultAnimation)]);
          uc := uc + #13#10;
        end;
      uc := uc + format('#exec MESH MODELIMPORT MESH=%s MODELFILE=MODELS\%s.PSK' {+' LODSTYLE=x X=x Y=x Z=x'} + #13#10, [basename, basename]);
      {uc := uc + format('#exec MESH LODPARAMS MESH=%s HYSTERESIS=%f STRENGTH=%f MINVERTS=%f MORPH=%f ZDISP=%f'#13#10,
        [basename, FLODHysteresis, FLODStrength, FLODMinVerts, FLODMorph, FLODZDisplace]);}
      uc := uc + format('#exec MESH ORIGIN MESH=%s X=%f Y=%f Z=%f YAW=%f ROLL=%f PITCH=%f'#13#10#13#10,
        [basename, FOrigin.x, FOrigin.y, FOrigin.z, FRotorigin.Yaw / 256, FRotOrigin.Roll / 256, FRotOrigin.Pitch / 256]);
      uc := uc + format('#exec MESH WEAPONATTACH MESH=%s BONE="%s"'#13#10, [basename, FOwner.Names[RefSkeleton[WeaponBoneIndex].name]]);
      uc := uc + format('#exec MESH WEAPONPOSITION MESH=%s YAW=%f PITCH=%f ROLL=%f X=%f Y=%f Z=%f'#13#10, [basename,
        0.0, 0.0, 0.0,                  // TODO : get this from weaponadjust.Xaxis, Yaxis and Zaxis (but with some calculation)
        weaponadjust.origin.x, weaponadjust.origin.y, weaponadjust.origin.z]);
      uc := uc + #13#10;
      for a := 0 to high(FTextures) do
        if FTextures[a].value > 0 then
          begin
            ed2 := FOwner.Exported[FTextures[a].value - 1];
            uc := uc + format('#exec TEXTURE IMPORT NAME=%s FILE=%s.PCX GROUP=%s'#13#10,
              [ed2.UTobjectname, ed2.UTobjectname, ed2.UTpackagename]);
            // TODO : TUTObjectClassSkeletalMesh.Save_UnrealUC : FLAGS=%d should put correct flags...
          end
        else if FTextures[a].value < 0 then
          begin
            id := FOwner.Imported[-FTextures[a].value - 1];
            uc := uc + format('#exec OBJ LOAD FILE=%s.utx PACKAGE=%s'#13#10, [id.UTobjectname, id.UTpackagename]);
          end;
      uc := uc + #13#10;
      for a := 0 to high(FTextures) do
        if FTextures[a].value > 0 then
          begin
            ed2 := FOwner.Exported[FTextures[a].value - 1];
            uc := uc + format('#exec MESHMAP SETTEXTURE MESHMAP=%s NUM=%d TEXTURE=%s'#13#10, [basename, a, ed2.UTobjectname]);
          end
        else if FTextures[a].value < 0 then
          begin
            id := FOwner.Imported[-FTextures[a].value - 1];
            uc := uc + format('#exec MESHMAP SETTEXTURE MESHMAP=%s NUM=%d TEXTURE=%s'#13#10, [basename, a, id.UTpackagename + '.' + id.UTobjectname]);
          end;
      uc := uc + #13#10;
      //uc:=uc+format('#exec MESHMAP NEW   MESHMAP=%s MESH=%s'#13#10,[basename,basename]);
      if (FScale.x <> 0.1) or (FScale.y <> 0.1) or (FScale.z <> 0.2) then
        uc := uc + format('#exec MESHMAP SCALE MESHMAP=%s X=%f Y=%f Z=%f'#13#10#13#10, [basename, FScale.x, FScale.y, FScale.z]);
      // FScale is incorrect, sometimes is 0 and it shouldnt be.

      uc := uc + format('#exec MESH DEFAULTANIM MESH=%s ANIM=%s'#13#10, [basename, FOwner.GetObjectPath(1, FDefaultAnimation)]);
      uc := uc + #13#10;
      if DefaultAnimation > 0 then
        begin
          uc := uc + '// Animation sequences not generated.'#13#10;
          //#exec ANIM SEQUENCE ANIM=SkeletalAnimation1 SEQ=All STARTFRAME=0 NUMFRAMES=25 RATE=0.2 GROUP=Default
          uc := uc + #13#10;
          uc := uc + format('#exec ANIM DIGEST ANIM=%s VERBOSE'#13#10, [FOwner.GetObjectPath(1, FDefaultAnimation)]);
          uc := uc + #13#10;
          uc := uc + '// Animation notifications not generated.'#13#10;
          //#exec ANIM NOTIFY ANIM=RiotAnim SEQ=RunLG TIME=0.25 FUNCTION=PlayFootStep
          uc := uc + #13#10;
        end;
      (*uc:=uc+#13#10+
      'defaultproperties'#13#10+
      '{'#13#10+
      '    DrawType=DT_Mesh'#13#10+
      format('    Mesh=%s'#13#10,[basename])+
      '}'#13#10;*)
      DecimalSeparator := k;
      str_uc := tfilestream.create(filename, fmCreate);
      try
        str_uc.write(uc[1], length(uc));
      finally
        str_uc.free;
      end;
    end;
end;

{ TUT_3DStudioExporter }

procedure TUT_3DStudioExporter.Save(filename: string);
type
  P3DSChunk = ^T3DSChunk;
  T3DSChunk = packed record
    tag: word;
    size: cardinal;
    datasize: cardinal;
    data: pointer;
    subchunks: tlist;
  end;
var
  str: tfilestream;
  m, m2, n: integer;
  dummy: byte;
  matname, meshname: string;
  p: pointer;
  facelist: array of word;
  rootchunk, mdatachunk, matentrychunk, namedobjectchunk, ntriobjectchunk,
    faceschunk, diffusechunk, {kfdatachunk, objectnodetagchunk, } x: P3DSChunk;
  function new_chunk(tag: word; datasize: cardinal; hasvalue: boolean; var value): P3DSChunk;
  begin
    new(result);
    result^.tag := tag;
    result^.size := 0;
    result^.datasize := datasize;
    if datasize > 0 then
      begin
        getmem(result^.data, datasize);
        if hasvalue then move(value, result^.data^, datasize);
      end
    else
      result^.data := nil;
    result^.subchunks := tlist.create;
  end;
  procedure calculate_chunk_size(chunk: P3DSChunk);
  var
    a: integer;
  begin
    chunk.size := sizeof(word) + sizeof(cardinal) + chunk.datasize;
    for a := 0 to chunk.subchunks.count - 1 do
      begin
        calculate_chunk_size(chunk.subchunks[a]);
        inc(chunk.size, P3DSChunk(chunk.subchunks[a]).size);
      end;
  end;
  procedure save_chunk(str: tstream; chunk: P3DSChunk);
  var
    a: integer;
  begin
    str.Write(chunk.tag, sizeof(word));
    str.write(chunk.size, sizeof(cardinal));
    if chunk.data <> nil then str.write(chunk.data^, chunk.datasize);
    for a := 0 to chunk.subchunks.count - 1 do
      save_chunk(str, chunk.subchunks[a]);
  end;
  procedure free_chunk(chunk: P3DSChunk);
  var
    a: integer;
  begin
    if chunk.data <> nil then freemem(chunk.data, chunk.datasize);
    for a := chunk.subchunks.count - 1 downto 0 do
      free_chunk(chunk.subchunks[a]);
    chunk.subchunks.free;
    dispose(chunk);
  end;
  function put_integer(p: pointer; v: integer): pointer;
  begin
    move(v, p^, sizeof(integer));
    result := pointer(integer(p) + sizeof(integer));
  end;
  function put_single(p: pointer; v: single): pointer;
  begin
    move(v, p^, sizeof(single));
    result := pointer(integer(p) + sizeof(single));
  end;
  function put_smallint(p: pointer; v: smallint): pointer;
  begin
    move(v, p^, sizeof(smallint));
    result := pointer(integer(p) + sizeof(smallint));
  end;
  function put_word(p: pointer; v: word): pointer;
  begin
    move(v, p^, sizeof(word));
    result := pointer(integer(p) + sizeof(word));
  end;
  function put_string(p: pointer; v: string): pointer;
  begin
    move(v[1], p^, length(v) + 1);
    result := pointer(integer(p) + length(v) + 1);
  end;
begin
  // Generate chunk tree
  // main chunk
  rootchunk := new_chunk(M3DMAGIC, 0, false, dummy);
  try
    x := new_chunk(M3D_VERSION, sizeof(cardinal), false, dummy);
    put_integer(x.data, 3);
    rootchunk.subchunks.add(x);
    // mesh data chunk
    x := new_chunk(MDATA, 0, false, dummy);
    rootchunk.subchunks.add(x);
    mdatachunk := x;
    // mesh version chunk
    x := new_chunk(MESH_VERSION, sizeof(integer), false, dummy);
    put_integer(x.data, 3);
    mdatachunk.subchunks.add(x);

    for m := 0 to high(Materials) do
      begin
        // material chunk
        x := new_chunk(MAT_ENTRY, 0, false, dummy);
        mdatachunk.subchunks.add(x);
        matentrychunk := x;
        matname := copy(Materials[m].name, 1, 10);
        // material name chunk
        x := new_chunk(MAT_NAME, length(matname) + 1, true, matname[1]);
        matentrychunk.subchunks.add(x);
        // TODO : must add support for material flags in parameters
        // material diffuse color chunk
        x := new_chunk(MAT_DIFFUSE, 0, false, dummy);
        matentrychunk.subchunks.add(x);
        diffusechunk := x;
        x := new_chunk(COLOR_24, 3, true, Materials[m].DiffuseColor[0]);
        diffusechunk.subchunks.add(x);
      end;

    // mesh chunk
    meshname := extractfilename(filename);
    meshname := copy(meshname, 1, 10);
    x := new_chunk(NAMED_OBJECT, length(meshname) + 1, true, meshname[1]);
    mdatachunk.subchunks.add(x);
    namedobjectchunk := x;
    // triangles chunk
    x := new_chunk(N_TRI_OBJECT, 0, false, dummy);
    namedobjectchunk.subchunks.add(x);
    ntriobjectchunk := x;
    // vertices chunk
    x := new_chunk(POINT_ARRAY, sizeof(word) + length(Vertices) * sizeof(single) * 3, false, dummy);
    p := put_word(x.data, length(Vertices));
    for m := 0 to high(Vertices) do
      begin
        if not mirrorx then
          p := put_single(p, Vertices[m].X)
        else
          p := put_single(p, -Vertices[m].X);
        p := put_single(p, Vertices[m].Y);
        p := put_single(p, Vertices[m].Z);
      end;
    ntriobjectchunk.subchunks.add(x);
    // texture coordinates chunk
    x := new_chunk(TEX_VERTS, sizeof(word) + length(Vertices) * sizeof(single) * 2, false, dummy);
    p := put_word(x.data, length(Vertices));
    for m := 0 to high(Vertices) do
      begin
        p := put_single(p, Vertices[m].U / 255);
        p := put_single(p, 1 - (Vertices[m].V / 255)); // reversed
      end;
    ntriobjectchunk.subchunks.add(x);
    // TODO : MESH_MATRIX chunk (not needed!?)
    // faces chunk
    x := new_chunk(FACE_ARRAY, sizeof(word) + length(Faces) * sizeof(word) * 4, false, dummy);
    p := put_word(x.data, length(Faces));
    for m := 0 to high(Faces) do
      begin
        if not mirrorX then
          begin
            p := put_word(p, Faces[m].VertexIndex1);
            p := put_word(p, Faces[m].VertexIndex2);
            p := put_word(p, Faces[m].VertexIndex3);
          end
        else
          begin                         // change vertex winding
            p := put_word(p, Faces[m].VertexIndex3);
            p := put_word(p, Faces[m].VertexIndex2);
            p := put_word(p, Faces[m].VertexIndex1);
          end;
        p := put_word(p, FaceCAVisable3DS or FaceBCVisable3DS or FaceABVisable3DS);
      end;
    ntriobjectchunk.subchunks.add(x);
    faceschunk := x;

    for m := 0 to high(Materials) do
      begin
        // face<->material chunk
        setlength(facelist, length(Faces));
        n := 0;
        for m2 := 0 to high(Faces) do
          begin
            if Faces[m2].MaterialIndex = m then
              begin
                facelist[n] := m2;
                inc(n);
              end;
          end;
        setlength(facelist, n);
        if n > 0 then
          begin
            matname := copy(Materials[m].name, 1, 10);
            x := new_chunk(MSH_MAT_GROUP, length(matname) + 1 + sizeof(word) + length(facelist) * sizeof(word), false, dummy);
            p := put_string(x.data, matname);
            p := put_word(p, length(facelist));
            for m2 := 0 to high(facelist) do
              p := put_word(p, facelist[m2]);
            faceschunk.subchunks.add(x);
          end;
      end;
    case smoothing of
      exp3ds_smooth_None: ;             // no smoothing groups
      exp3ds_smooth_One:
        begin                           // one smoothing group for all faces
          // Smooth Group chunk
          x := new_chunk(SMOOTH_GROUP, length(Faces) * sizeof(cardinal), false, dummy);
          p := x.data;
          for m := 0 to high(Faces) do
            p := put_integer(p, 1);
          faceschunk.subchunks.add(x);
        end;
      exp3ds_smooth_exp3ds_smooth_ByMaterial:
        begin                           // one smoothing group for each material
          // Smooth Group chunk
          x := new_chunk(SMOOTH_GROUP, length(Faces) * sizeof(cardinal), false, dummy);
          p := x.data;
          for m := 0 to high(Faces) do
            p := put_integer(p, 1 shl Faces[m].MaterialIndex); // one bit for each material
          faceschunk.subchunks.add(x);
        end;
    end;

    // keyframes chunk
      (*x := new_chunk(KFDATA, 0, false, dummy);
      rootchunk.subchunks.add(x);
      kfdatachunk := x;
      // keyframes header chunk
      name := 'MAXSCENE';
      x := new_chunk(KFHDR, sizeof(smallint) + length(name) + 1 + sizeof(integer), false, dummy);
      p := put_smallint(x.data, 5);
      p := put_string(p, name);
      {if allframes then
         put_integer(p, FAnimFrames)
      else}
      put_integer(p, 1);
      kfdatachunk.subchunks.add(x);
      // keyframes segments chunk
      x := new_chunk(KFSEG, sizeof(integer) + sizeof(integer), false, dummy);
      p := put_integer(x.data, 0);
      {if allframes then
         put_integer(p, FAnimFrames)
      else}
      put_integer(p, 1);
      kfdatachunk.subchunks.add(x);

      for frame := 0 to FAnimFrames - 1 do
        begin
          // keyframes current time chunk
          x := new_chunk(KFCURTIME, sizeof(integer), false, dummy);
          put_integer(x.data, frame);
          kfdatachunk.subchunks.add(x);
          // object node chunk
          x := new_chunk(OBJECT_NODE_TAG, 0, false, dummy);
          kfdatachunk.subchunks.add(x);
          objectnodetagchunk := x;
          // node id chunk
          x := new_chunk(NODE_ID, sizeof(smallint), false, dummy);
          put_smallint(x.data, frame);
          objectnodetagchunk.subchunks.add(x);
          // node header chunk
          meshname := extractfilename(filename);
          meshname:=copy(meshname,1,7);
          if allframes then meshname := meshname + format('%-3.3d', [frame]);
          x := new_chunk(NODE_HDR, length(meshname) + 1 + sizeof(word) + sizeof(word) + sizeof(smallint), false, dummy);
          p := put_string(x.data, meshname); // Mesh object
          p := put_word(p, $4000); // flags1=PRIMARY_NODE
          p := put_word(p, 0);     // flags2
          put_smallint(p, -1);     // Parent=No Parent
          objectnodetagchunk.subchunks.add(x);

          {
          // PIVOT chunk
          x := new_chunk(PIVOT, 3 * sizeof(single), false, dummy);
          p := put_single(x.data, 0);
          p := put_single(p, 0);
          put_single(p, 0);
          objectnodetagchunk.subchunks.add(x);
          // POS_TRACK_TAG chunk
          x := new_chunk(POS_TRACK_TAG, sizeof(word) + 3 * sizeof(integer) + 1 * (sizeof(integer) + sizeof(word) + 3 * sizeof(single)), false, dummy);
          p := put_word(x.data, 0);         // flags
          p := put_integer(p, 0);           // nu1
          p := put_integer(p, 0);           // nu2
          p := put_integer(p, 1);           // keycount
          // key 0
          p := put_integer(p, 0);           // time
          p := put_word(p, 0);              // rflags
          p := put_single(p, 0);            // point.X
          p := put_single(p, 0);            // point.Y
          put_single(p, 0);                 // point.Z
          objectnodetagchunk.subchunks.add(x);
          // ROT_TRACK_TAG chunk
          x := new_chunk(ROT_TRACK_TAG, sizeof(word) + 3 * sizeof(integer) + 1 * (sizeof(integer) + sizeof(word) + 4 * sizeof(single)), false, dummy);
          p := put_word(x.data, 0);         // flags
          p := put_integer(p, 0);           // nu1
          p := put_integer(p, 0);           // nu2
          p := put_integer(p, 1);           // keycount
          // key 0
          p := put_integer(p, 0);           // time
          p := put_word(p, 0);              // rflags
          p := put_single(p, 0);            // Angle
          p := put_single(p, 0);            // Axis.X
          p := put_single(p, 1);            // Axis.Y
          put_single(p, 0);                 // Axis.Z
          objectnodetagchunk.subchunks.add(x);
          // SCL_TRACK_TAG chunk
          x := new_chunk(SCL_TRACK_TAG, sizeof(word) + 3 * sizeof(integer) + 1 * (sizeof(integer) + sizeof(word) + 3 * sizeof(single)), false, dummy);
          p := put_word(x.data, 0);         // flags
          p := put_integer(p, 0);           // nu1
          p := put_integer(p, 0);           // nu2
          p := put_integer(p, 1);           // keycount
          // key 0
          p := put_integer(p, 0);           // time
          p := put_word(p, 0);              // rflags
          p := put_single(p, 1);            // scale.X
          p := put_single(p, 1);            // scale.Y
          put_single(p, 1);                 // scale.Z
          objectnodetagchunk.subchunks.add(x);
          }
          if not allframes then break;
        end;
      *)

      // calculate chunk sizes
    calculate_chunk_size(rootchunk);
    // Save chunk tree
    str := tfilestream.create(filename + '.3ds', fmCreate);
    try
      save_chunk(str, rootchunk);
    finally
      str.free;
    end;
  finally
    // free chunks
    free_chunk(rootchunk);
  end;
end;

{ TUT_Unreal3DExporter }

procedure TUT_Unreal3DExporter.Save(filename: string);
var
  a, xyz: integer;
  xx, yy, zz: single;
  data: word;
  str_d, str_a: tfilestream;
  // From 3DS2UNR
  _3D_dHeader: packed record
    NumPolygons: word;
    NumVertices: word;
    BogusRot: word;
    BogusFrame: word;
    BogusNormX: cardinal;
    BogusNormY: cardinal;
    BogusNormZ: cardinal;
    FixScale: cardinal;
    Unused: array[0..2] of cardinal;
    Unknown: array[0..11] of char;
  end;
  _3D_dPolygon: packed record
    mVertex: array[0..2] of word;       // Vertex indices.
    mType: byte;                        // James' mesh type.
    mColor: byte;                       // Color for flat and Gouraud shaded.
    mTex: array[0..2, 0..1] of byte;    // Texture UV coordinates.
    mTextureNum: byte;                  // Source texture offset.
    mFlags: byte;                       // Unreal mesh flags (currently unused).
  end;
begin
  if coordsdivisor < 1 then coordsdivisor := 1;
  filename := changefileext(filename, '');
  // Create file _d.3d
  fillchar(_3D_dHeader, sizeof(_3D_dHeader), #0);
  with _3D_dHeader do
    begin
      NumPolygons := length(Faces);
      NumVertices := length(Vertices) div AnimationFrames;
    end;
  str_d := tfilestream.create(filename + '_d.3d', fmCreate);
  try
    str_d.Write(_3D_dHeader, sizeof(_3D_dHeader));
    for a := 0 to high(Faces) do
      begin
        fillchar(_3D_dPolygon, sizeof(_3D_dPolygon), #0);
        _3D_dPolygon.mVertex[0] := Faces[a].vertexindex1;
        _3D_dPolygon.mVertex[1] := Faces[a].vertexindex2;
        _3D_dPolygon.mVertex[2] := Faces[a].vertexindex3;
        if (Faces[a].flags and (PF_TwoSided or PF_Modulated)) = (PF_TwoSided or PF_Modulated) then
          _3D_dPolygon.mType := 4       // MODULATED 2-Sided
        else if (Faces[a].flags and (PF_TwoSided or PF_Translucent)) = (PF_TwoSided or PF_Translucent) then
          _3D_dPolygon.mType := 3       // TRANSLUCENT 2-Sided
        else if (Faces[a].flags and (PF_TwoSided or PF_Masked)) = (PF_TwoSided or PF_Masked) then
          _3D_dPolygon.mType := 2       // MASKED 2-Sided
        else if (Faces[a].flags and PF_TwoSided) <> 0 then
          _3D_dPolygon.mType := 1       // 2-Sided
        else
          _3D_dPolygon.mType := 0;
        if (Faces[a].flags and PF_Unlit) <> 0 then _3D_dPolygon.mType := _3D_dPolygon.mType or $10;
        if (Faces[a].flags and PF_Flat) <> 0 then _3D_dPolygon.mType := _3D_dPolygon.mType or $20;
        if (Faces[a].flags and PF_Environment) <> 0 then _3D_dPolygon.mType := _3D_dPolygon.mType or $40;
        if (Faces[a].flags and PF_NoSmooth) <> 0 then _3D_dPolygon.mType := _3D_dPolygon.mType or $80;
        _3D_dPolygon.mTextureNum := Faces[a].MaterialIndex;
        _3D_dPolygon.mTex[0, 0] := Vertices[Faces[a].vertexindex1].U;
        _3D_dPolygon.mTex[0, 1] := Vertices[Faces[a].vertexindex1].V;
        _3D_dPolygon.mTex[1, 0] := Vertices[Faces[a].vertexindex2].U;
        _3D_dPolygon.mTex[1, 1] := Vertices[Faces[a].vertexindex2].V;
        _3D_dPolygon.mTex[2, 0] := Vertices[Faces[a].vertexindex3].U;
        _3D_dPolygon.mTex[2, 1] := Vertices[Faces[a].vertexindex3].V;
        str_d.write(_3D_dPolygon, sizeof(_3D_dPolygon));
      end;
  finally
    str_d.free;
  end;

  // Create file _a.3d
  str_a := tfilestream.create(filename + '_a.3d', fmCreate);
  try
    data := AnimationFrames;
    str_a.Write(data, sizeof(data));
    data := length(Vertices) * 4 div AnimationFrames;
    str_a.write(data, sizeof(data));
    for a := 0 to high(Vertices) do
      begin
        xx := Vertices[a].x / CoordsDivisor;
        yy := Vertices[a].y / CoordsDivisor;
        zz := Vertices[a].z / CoordsDivisor;
        xyz := (trunc(xx * 8) and $7FF) or
          ((trunc(yy * 8) and $7FF) shl 11) or
          ((trunc(zz * 4) and $3FF) shl 22);
        str_a.Write(xyz, 4);
      end;
  finally
    str_a.free;
  end;
end;

{ TUTImportTableObjectData }

procedure TUTImportTableObjectData.SetClassIndex(const Value: integer);
begin
  FClassIndex := Value;
end;

procedure TUTImportTableObjectData.SetClassName(const Value: string);
begin
  FClassName := Value;
end;

procedure TUTImportTableObjectData.SetClassPackageIndex(
  const Value: integer);
begin
  FClassPackageIndex := Value;
end;

procedure TUTImportTableObjectData.SetClassPackageName(
  const Value: string);
begin
  FClassPackageName := Value;
end;

procedure TUTImportTableObjectData.SetObjectIndex(const Value: integer);
begin
  FObjectIndex := Value;
end;

procedure TUTImportTableObjectData.SetObjectName(const Value: string);
begin
  FObjectName := Value;
end;

procedure TUTImportTableObjectData.SetPackageIndex(const Value: integer);
begin
  FPackageIndex := Value;
end;

procedure TUTImportTableObjectData.SetPackageName(const Value: string);
begin
  FPackageName := Value;
end;

{ TUTExportTableObjectData }

procedure TUTExportTableObjectData.SetClassIndex(const Value: integer);
begin
  FClassIndex := Value;
end;

procedure TUTExportTableObjectData.SetClassName(const Value: string);
begin
  FClassName := Value;
end;

procedure TUTExportTableObjectData.SetFlags(const Value: integer);
begin
  FFlags := Value;
end;

procedure TUTExportTableObjectData.SetObjectIndex(const Value: integer);
begin
  FObjectIndex := Value;
end;

procedure TUTExportTableObjectData.SetObjectName(const Value: string);
begin
  FObjectName := Value;
end;

procedure TUTExportTableObjectData.SetPackageIndex(const Value: integer);
begin
  FPackageIndex := Value;
end;

procedure TUTExportTableObjectData.SetPackageName(const Value: string);
begin
  FPackageName := Value;
end;

procedure TUTExportTableObjectData.SetSerialOffset(const Value: integer);
begin
  FSerialOffset := Value;
end;

procedure TUTExportTableObjectData.SetSerialSize(const Value: integer);
begin
  FSerialSize := Value;
end;

procedure TUTExportTableObjectData.SetSuperIndex(const Value: integer);
begin
  FSuperIndex := Value;
end;

procedure TUTExportTableObjectData.SetSuperName(const Value: string);
begin
  FSuperName := Value;
end;

procedure TUTExportTableObjectData.SetUTObject(const Value: TUTObject);
begin
  FUTObject := Value;
end;

var i:integer;

initialization

  // do not localize class names

  ClearUTClassEquivalences;
  setlength(RegisteredUTClasses, 0);

  // Classes with specific object class
  // (commented lines are for abstract or unimplemented classes)
  RegisterUTObjectClass('', TUTObjectClassClass); // special case, the class definitions do not have a class name
  RegisterUTObjectClass('Palette', TUTObjectClassPalette);
  RegisterUTObjectClass('Sound', TUTObjectClassSound);
  RegisterUTObjectClass('Music', TUTObjectClassMusic);
  RegisterUTObjectClass('TextBuffer', TUTObjectClassTextBuffer);
  RegisterUTObjectClass('Font', TUTObjectClassFont);
  RegisterUTObjectClass('Polys', TUTObjectClassPolys);
  //RegisterUTObjectClass ('Bitmap',TUTObjectClassBitmap);
  {****} RegisterUTObjectClass('Texture', TUTObjectClassTexture);
  {****}{****}// RegisterUTObjectClass('FractalTexture', TUTObjectClassTexture);
  {****}{****}{****} RegisterUTObjectClass('FireTexture', TUTObjectClassFireTexture);
  {****}{****}{****} RegisterUTObjectClass('IceTexture', TUTObjectClassTexture);
  {****}{****}{****} RegisterUTObjectClass('WaterTexture', TUTObjectClassTexture);
  {****}{****}{****}{****} RegisterUTObjectClass('WaveTexture', TUTObjectClassTexture);
  {****}{****}{****}{****} RegisterUTObjectClass('WetTexture', TUTObjectClassTexture);
  {****}{****} RegisterUTObjectClass('ScriptedTexture', TUTObjectClassTexture);
  RegisterUTObjectClass('Primitive', TUTObjectClassPrimitive);
  {****}// RegisterUTObjectClass ('Model',TUTObject);
  {****} RegisterUTObjectClass('Mesh', TUTObjectClassMesh);
  {****}{****} RegisterUTObjectClass('LodMesh', TUTObjectClassLodMesh);
  {****}{****}{****} RegisterUTObjectClass('SkeletalMesh', TUTObjectClassSkeletalMesh);
  RegisterUTObjectClass('Field', TUTObjectClassField);
  {****} RegisterUTObjectClass('Const', TUTObjectClassConst);
  {****} RegisterUTObjectClass('Enum', TUTObjectClassEnum);
  {****} RegisterUTObjectClass('Property', TUTObjectClassProperty);
  {****}{****} RegisterUTObjectClass('ByteProperty', TUTObjectClassByteProperty);
  {****}{****} RegisterUTObjectClass('IntProperty', TUTObjectClassIntProperty);
  {****}{****} RegisterUTObjectClass('BoolProperty', TUTObjectClassBoolProperty);
  {****}{****} RegisterUTObjectClass('FloatProperty', TUTObjectClassFloatProperty);
  {****}{****} RegisterUTObjectClass('ObjectProperty', TUTObjectClassObjectProperty);
  {****}{****}{****} RegisterUTObjectClass('ClassProperty', TUTObjectClassClassProperty);
  {****}{****} RegisterUTObjectClass('NameProperty', TUTObjectClassNameProperty);
  {****}{****} RegisterUTObjectClass('StructProperty', TUTObjectClassStructProperty);
  {****}{****} RegisterUTObjectClass('StrProperty', TUTObjectClassStrProperty);
  {****}{****} RegisterUTObjectClass('ArrayProperty', TUTObjectClassArrayProperty);
  {****}{****} RegisterUTObjectClass('FixedArrayProperty', TUTObjectClassFixedArrayProperty);
  {****}{****} RegisterUTObjectClass('MapProperty', TUTObjectClassMapProperty);
  {****}{****} RegisterUTObjectClass('StringProperty', TUTObjectClassStringProperty);
  {****} RegisterUTObjectClass('Struct', TUTObjectClassStruct);
  {****}{****} RegisterUTObjectClass('Function', TUTObjectClassFunction);
  {****}{****} RegisterUTObjectClass('State', TUTObjectClassState);
  {****}{****}{****} RegisterUTObjectClass('Class', TUTObjectClassClass);
  RegisterUTObjectClass('Linker', TUTObject_NoHeader);
  {****} RegisterUTObjectClass('LinkerLoad', TUTObject_NoHeader);
  {****} RegisterUTObjectClass('LinkerSave', TUTObject_NoHeader);
  RegisterUTObjectClass('Package', TUTObject_NoHeader);
  RegisterUTObjectClass('PackageMap', TUTObject_NoHeader);
  //RegisterUTObjectClass ('SubSystem',TUTObject_NoHeader);
  {****} RegisterUTObjectClass('System', TUTObject_NoHeader);
  RegisterUTObjectClass('RenderDevice', TUTObject_NoHeader);
  //RegisterUTObjectClass ('LevelBase',TUTObject_NoHeader);
  {****} RegisterUTObjectClass('Level', TUTObject_NoHeader);
  RegisterUTObjectClass('Console', TUTObject_NoHeader);

  setlength(NativeFunctions,length(NativeFunctions_UT));
  for i:=0 to high(NativeFunctions_UT) do
      NativeFunctions[i]:=NativeFunctions_UT[i];

end.

